
    #include <map>
    #include <list>
    #include <deque>
    #include <string>
    #include <set>
    
    #include <memory>
    
    #include "inttypes.h"
    
    #include "buffer.h"
    #include "containers.h"
    #include "core.h"
    #include "graphics.h"
    
    #include "texttypes.h"

#ifndef DUNGTYPES_H

    #define DUNGTYPES_H

namespace Thorn
{
    using std::list;
    using std::map;
    
    using std::string;
    
    using std::shared_ptr;
    using std::weak_ptr;
    
    using Containers::Matrix;
    using Containers::Ticket;
    
    using Graphics::TileEntry;
    
// =============================================================================

    typedef Matrix<TileEntry> Block;

// =============================================================================

    //! \brief      Can be used to create a Thorn::Block, but contains
    //!             no actual TileEntry objects itself.
    //! 
    //! \since      0.11
    class IndirectBlock : public Matrix<int>
    {
        
    public:
        
        //! \brief      Offset within a target rom where the tiles reside.
        uint32_t m_offset;
        
    public:
        
        IndirectBlock(int p_offset,
                      int p_rows,
                      int p_columns,
                      list<int> p_indices)
        
        : m_offset(p_offset),
          Matrix<int>(p_rows,
                      p_columns,
                      p_indices)
        
        {
            
        }
        
        virtual ~IndirectBlock()
        {
            
        }
        
    };

// =============================================================================

    class BlockRef
    {
        
    protected:
        
        bool m_hflip;
        bool m_vflip;
        
        Ticket<Block> m_ref;
        
    public:
        
        BlockRef(Ticket<Block> p_ref   = Ticket<Block>(),
                 bool          p_hflip = false,
                 bool          p_vflip = false)
        
        : m_ref(p_ref),
          m_hflip(p_hflip),
          m_vflip(p_vflip)
        
        {
            
        }
    };

// =============================================================================

    class Entity
    {
        
    public:
        
        // abstract base class
        // virtual void foo() = 0;
        
        virtual ~Entity() { }
    };

// =============================================================================

    enum Operator
    {
        PLUS,
        SUBTRACT,
        MULTIPLY,
        DIVIDE,
        PERIOD,
        LOG_EQ,
        LOG_NEQ,
        IN
    };

// =============================================================================
    
    enum Node_Type
    {
        assign,
        obj_call,
        conditional_block,
        expr,
        loop_block,
        var_decl
    };
    
// =============================================================================

    class Node
    {
        
    public:
        
        Node_Type m_type;
        
        Entity *m_obj;
        
    public:
        
        Node(Entity *p_obj, Node_Type p_type)
        
        : m_obj(p_obj),
          m_type(p_type)
        {
            
        }
        
        ~Node()
        {
            delete m_obj;
        }
    };

// =============================================================================

    class Assign : public Entity
    {
    };

// =============================================================================

    class Obj_Call : public Entity
    {
    };

// =============================================================================

    class Conditional_Block : public Entity
    {
        Node m_condition;
        
        list<Node> m_nodes;
    };

// =============================================================================

    class Expr : public Entity
    {
        Operator m_op;
        
        Node m_lhs;
        Node m_rhs;
    };

// =============================================================================

    class Loop_Block : public Entity
    {
        Node m_repeat;
        
        list<Node> m_nodes;
    };
    
// =============================================================================

    enum Param_Type
    {
        block_ref,
        repeat,
        offset,
    };

// =============================================================================

    class Obj_Param : public Entity
    {
        string m_name;
        
        Param_Type m_type;
        
    public:
        
        Obj_Param(string     p_name,
                  Param_Type p_type)
        
        : m_name(p_name),
          m_type(p_type)
        {
        }
    };

// =============================================================================

    class Var_Decl : public Entity
    {
        string m_name;
        
        Node m_lhs;
        Node m_rhs;
    };

// =============================================================================

    class Obj_Def
    {
        list<Obj_Param> m_params;
        
        list<Node> m_nodes;
        
    public:
        
        Obj_Def(list<Obj_Param> p_params,
                list<Node>      p_nodes)
        
        : m_params(p_params),
          m_nodes(p_nodes)
        {
            
        }
    };

// =============================================================================

    class Obj_Invoke
    {
        uint8_t m_x;
        uint8_t m_y;
        
        Obj_Call m_call;
    };

// =============================================================================

    void TestSharedPtr();
    void Durp();
    
    extern map<string, IndirectBlock> defaultBlocks;
    
}

namespace Dungeon
{

// =========================================================
    
    // Loading symbols from other moduless.
    
    using std::list;
    using std::deque;
    using std::string;
    
    using namespace Containers;
    using namespace Memory;
    
    using Graphics::TileEntry;
    using Graphics::TileAttr;
    
// =============================================================================

    //! \brief      In dungeons, all position objects are either on BG2, the
    //!             main background layer, or BG1, the secondary background
    //!             layer. I imagine to newcomers this seems somewhat
    //!             unintuitive, but that's just how the game's graphical layout
    //!             is set up.
    enum Floor
    {
        BG1,
        BG2
    };
    
    template<uint16_t grid = 1>
    class Position
    {
        
    protected:
        
        uint16_t m_x;
        uint16_t m_y;
        
        Floor m_floor;
        
    public:
        
        // _____________________________________________________________________
        Position(uint16_t p_x = 0, uint16_t p_y = 0, Floor p_floor = BG2)
        
        : m_floor(p_floor)
        {
            Set(p_x, p_y);
        }
        
        //! \brief      Virtual destructor added so that if this class is 
        //!             derived from, it can be overridden.
        virtual ~Position()
        {
            
        }
        
        // _____________________________________________________________________
        //! \brief      Helper member function for initialization of the
        //!             object's coordites. Snaps inputted coordinates to the
        //!             grid specified at compile time.
        virtual void Set(uint16_t p_x, uint16_t p_y)
        {
            m_x = ( p_x - (p_x % grid) );
            m_y = ( p_y - (p_y % grid) );
        }
        
        // _____________________________________________________________________
        //! \brief      Allows setting X position independently. Still snaps to
        //!             grid.
        virtual void SetX(uint16_t p_x)
        {
            Set(p_x, m_y);
        }
        
        // _____________________________________________________________________
        //! \brief      Allows setting Y position independently. Still snaps to
        //!             grid.
        virtual void SetY(uint16_t p_y)
        {
            Set(m_x, p_y);
        }
        
        // _____________________________________________________________________
        //! \brief      Allows changing the position object's floor setting.
        virtual void SetFloor(Floor p_f)
        {
            m_floor = p_f;
        }
        
        // _____________________________________________________________________
        //! \brief      Returns the position's X coordinate.
        virtual uint16_t GetX()
        {
            return m_x;
        }
        
        // _____________________________________________________________________
        //! \brief      Returns the position's Y coordinate.
        virtual uint16_t GetY()
        {
            return m_y;
        }
        
        // _____________________________________________________________________
        //! \brief      Returns the position's floor setting.
        virtual Floor GetFloor()
        {
            return m_floor;
        }
    };

// =============================================================================

    class Header
    {
    public:
        
        // The phrasing Byte X refers to where these values reside
        // in the acutal packed header in the rom.
        
        uint8_t m_bg0_props;          // {
        uint8_t m_collision;
        uint8_t m_lightsOut;          // } Byte 0
        
        uint8_t m_paletteIndex;       // Byte 1
        uint8_t m_bgGfx;              // Byte 2
        uint8_t m_spriteGfx;          // Byte 3
        uint8_t m_effect;             // Byte 4
        uint8_t m_tag_1;              // Byte 5
        uint8_t m_tag_2;              // Byte 6
        
        uint8_t m_warpPlane;          // {
        uint8_t m_staircasePlane1;
        uint8_t m_staircasePlane2;
        uint8_t m_staircasePlane3;    // } Byte 7
        
        uint8_t m_staircasePlane4;    // Byte 8
        
        uint8_t m_warpRoom;           // Byte 9
        
    protected:
        
        uint8_t m_staircaseRooms[4];   // Bytes 10 through 13
        
    public:
        
        Header();
        Header(Buffer *data);
        
        uint8_t GetStaircaseRoom(int i) const;
        
        void Initialize(Buffer *data);
    };
    
// =============================================================================

    class Sprite
    {
        
    public:
    
        enum DropType { none, small_key, big_key, custom};
        
    protected:
        
        uint8_t m_subType;
        
        uint16_t m_type;
        
        // indicates whether the sprite drops a key or not
        DropType m_drop;
        
        Position<16> m_pos;
        
    public:
        
        // _____________________________________________________________________
        Sprite(Floor    p_f = BG2,
               uint16_t p_x = 0,
               uint16_t p_y = 0,
               uint8_t  p_type = 0,
               uint8_t  p_subType = 0,
               DropType p_drop = none)
            
        : m_pos(p_x, p_y, p_f),
          m_drop(p_drop),
          m_type(p_type),
          m_subType(p_subType) 
        {
            return;
        }
        
        // _____________________________________________________________________
        Sprite(Sprite const &s)
        
        : m_pos(s.m_pos),
          m_drop(s.m_drop),
          m_type(s.m_type),
          m_subType(s.m_subType)          
        {
            
        }
                
        // _____________________________________________________________________
        void SetDrop(DropType p_drop)
        {
            m_drop = p_drop;
        }
    };

// =============================================================================

    class Overlord
    {
    protected:
        
        uint16_t m_type;
        
        Position<16> m_pos;
        
    public:
        
        // _____________________________________________________________________
        Overlord(Floor    p_floor = BG2,
                 uint16_t p_x = 0,
                 uint16_t p_y = 0,
                 uint8_t  p_type = 0)
        
        : m_pos(p_x, p_y, p_floor),
          m_type(p_type)
        {
            
        }
        
        // _____________________________________________________________________
        Overlord(Overlord const &p_o)
        
        : m_pos(p_o.m_pos),
          m_type(p_o.m_type)
        {
            
        }
    };

// =========================================================

    class SpriteList
    {
        
    public:
        
        bool sortSprites;
        
        list<Sprite>   sprites;
        list<Overlord> overlords;
        
        SpriteList()
        
        /// not sure if the default should be true or false
        /// but does it really matter?
        : sortSprites(false)
        {
        }
        
        SpriteList(SpriteList const &sl)
        
        : sortSprites(sl.sortSprites),
          sprites(sl.sprites),
          overlords(sl.overlords)
        {
            
        } 
    };

// =========================================================

    class Object
    {
    public:
        
        // subtypes of objects
        enum objType { type1_1, type1_2, type1_3};
        
        // This may be needed if an object triggers an error state
        // during load and has to be disabled
        bool m_disabled;
        
        objType m_type;
        
        // together with the type variable this uniquely identifies the object
        uint8_t m_routine;
        
        // tilemap address of the object (relative to $7E2000 and / or $7E4000)
        uint16_t m_addr;
        
        // correspond to $B2 and $B4
        uint32_t m_arg1;
        uint32_t m_arg2;
        
        /// the rest of these are of dubious utility, however they may be needed
        /// for writing the gui front end later. it will probably be necessary
        /// to toss a few of them out though.
        uint8_t m_width;
        uint8_t m_height;
        uint8_t m_layer;
        
        uint32_t m_x;
        uint32_t m_y;
        uint32_t m_mem_08;
        uint32_t m_mapAddr;
        uint32_t m_tileNum;
        uint32_t m_mapNum;
        uint32_t m_doMin;
        uint32_t m_doMax;
        
    public:
        
        Object()
        {
            m_type    = type1_1;
            m_routine = 0;
            m_arg1    = 0;
            m_arg2    = 0;
            m_width   = 0;
            m_height  = 0;
            m_x       = 0;
            m_y       = 0;
            m_addr    = 0;
        }
        
        Object(objType p_type, uint32_t p_data);
        Object(Object const &p_o);
    };
    
    typedef list<Object> ObjectList;

// =========================================================
    
    class Door
    {
        
    public:
        
        enum Direction
        {
            up    = 0,
            down  = 1,
            left  = 2,
            right = 3
        };
        
    protected:
        
        Direction m_direction;
        uint8_t m_position;
        uint8_t m_type;
        
        // game doesn't use this data
        uint8_t m_extra;
        
    public:
        
        class Problem : public exception
        {
            
        };
        
        Door(uint16_t data = 0)
        {
            m_extra = static_cast<uint8_t> ( (data >> 2) & 0x03 );
            
            SetDirection( static_cast<uint8_t>( (data >> 0) & 0x03 ) );
            
            SetPosition( (data >> 4) & 0x0F );
            
            SetType( (data >> 9) & 0x7F );
        }
        
        Door(Door const &d)
        {
            m_extra = d.m_extra;
            
            SetDirection(d.m_direction);
            
            SetPosition(d.m_position);
            
            SetType(d.m_type);
        }
        
        // _____________________________________________________________________
        void SetDirection(uint8_t arg)
        {
            switch(arg)
            {
                default:
                    throw Problem(); break;
                case 0:
                    m_direction = up; break;
                case 1:
                    m_direction = down; break;
                case 2:
                    m_direction = left; break;
                case 3:
                    m_direction = right; break;
            }
        }
        
        // _____________________________________________________________________
        void SetDirection(Direction dir)
        {
            m_direction = dir;   
        }
        
        // _____________________________________________________________________
        Direction GetDirection()
        {
            return m_direction;
        }
        
        // _____________________________________________________________________
        void SetPosition(uint8_t arg)
        {
            if(arg > 11)
            {
                throw Problem();
            }
            
            m_position = arg;
        }
        
        // _____________________________________________________________________
        uint8_t GetPosition()
        {
            return m_position;
        }
        
        // _____________________________________________________________________
        void SetType(uint8_t arg)
        {
            if(arg > 41)
            {
                throw Problem();
            }
            
            m_type = arg;
        }
        
        // _____________________________________________________________________
        uint8_t GetType()
        {
            return m_type;
        }
        
        // _____________________________________________________________________
        uint8_t GetExtra()
        {
            return m_extra;
        }
    };
    
// =========================================================

    typedef list<Door> DoorList;
    
// =========================================================

    class ObjectData
    {
        
    public:
        
        ObjectList type1;
        DoorList doors;
        
    public:
        
        // todo: constructor and destructor
    };

// =========================================================

    // class for push blocks in dungeons
    
    class Block
    {
        
    protected:
        
        // if set, the block can trigger an event by being pushed
        bool triggersEvent;
        
        Position<8> pos;
        
    public:
        
        Block(bool     p_triggersEvent = false,
              uint16_t p_x = 0,
              uint16_t p_y = 0,
              Floor    p_floor = BG2)
        
        : triggersEvent(p_triggersEvent),
          pos(p_x, p_y, p_floor)
        {
            
        }
        
        Block(Block const &b)
        
        : triggersEvent(b.triggersEvent),
          pos(b.pos)
        {
        
        }
    };
    
    typedef list<Block> BlockList;

// =========================================================

    class Torch
    {
    protected:
        
        bool m_lit;
        
        Position<8> m_pos;
        
    public:
        
        Torch(bool     p_lit = false,
              uint16_t p_x = 0,
              uint16_t p_y = 0,
              Floor    p_floor = BG2)
        
        : m_pos(p_x, p_y, p_floor),
          m_lit(p_lit)
        {
            return;
        }
        
        Torch(Torch const &p_t)
        
        : m_pos(p_t.m_pos),
          m_lit(p_t.m_lit)
        {
        
        }
    };
    
    typedef list<Torch> TorchList;

// =========================================================

    class Item
    {
    
    public:

        enum ItemType
        {
            none         = 0x00,
            rupee        = 0x01,
            heart        = 0x02,
            small_magic  = 0x03,
            large_magic  = 0x04,
            chicken      = 0x05,
            key          = 0x06,
            hole         = 0x80,
            trigger      = 0x88 // (floor switch)
        }; // need to add more

    protected:
        
        Position<8> m_pos;
        
        ItemType m_type;
        
    public:
        
    // >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
        
        Item(uint16_t p_x = 0,
             uint16_t p_y = 0,
             Floor    p_floor = BG2,
             ItemType p_type = none)
        
        : m_pos(p_x, p_y, p_floor),
          m_type(p_type)
        {
            return;
        }
        
    // >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
        
        Item(Item const &p_item)
        
        : m_pos(p_item.m_pos),
          m_type(p_item.m_type)
        {
        
        }
        
    // >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
        
        void SetType(uint8_t p_type)
        {
            switch(p_type)
            {
            case none:
            case rupee:
            
                m_type = (ItemType) p_type;
                
                break;
                
            default:
            
                break;
            }
        }
    };
    
    typedef list<Item> ItemList;
    
// =============================================================================
    
    class Chest
    {
        
    public:
    
        Position<8> m_pos;
        
        uint16_t m_contents;
        
    public:
        
        // _____________________________________________________________________
        
        Chest(uint16_t p_x = 0,
              uint16_t p_y = 0,
              Floor    p_floor = BG2,
              uint16_t p_contents = 0)
        
        : m_pos(p_x, p_y, p_floor),
          m_contents(p_contents)
        {
            
        }
    };
    
    typedef list<Chest> ChestList;

// =============================================================================

    class Lock
    {
        
    public:
        
        /// add a type of key?, otherwise this might as well just be
        /// a position object
        uint16_t m_event;
        
        Position<8> m_pos;
        
    public:
        
        // _____________________________________________________________________
        Lock(uint16_t p_x = 0,
             uint16_t p_y = 0,
             Floor    p_floor = BG2)
        : m_pos(p_x, p_y, p_floor),
          m_event(0)
        {
            
        }
        
        // _____________________________________________________________________
        Lock(Lock const &lock)
        
        : m_pos(lock.m_pos),
          m_event(lock.m_event)
        {
        
        }
    };
    
    typedef list<Lock> LockList;
    typedef list< Position<8> > StarList;

// =============================================================================

    //! \brief      Object representing telepathic tiles (and quite frankly, any
    //!             other readable tile object we can come up with, but the only
    //!             existing readable objects in the dungeons are telepathic
    //!             tiles.
    class Sign : public Position<8>
    {
        
    protected:
        
        // Name shortening type definitions.
        typedef Position<8>           pos_t;
        
        typedef Ticket<Text::Message> ticket_t;
        
        typedef Thorn::BlockRef       block_ref;
        
    protected:
        
        //! \brief      Reference to the text of the sign.
        ticket_t m_msg;
        
        //! \brief      Block used for the appearance of the sign.
        block_ref m_block;
        
        //! \brief      Height of the sign (number of times to repeat m_block 
        //!             vertically).
        uint8_t m_rows;
        
        //! \brief      Width of the sign (number of times to repeat m_block 
        //!             horizontally). 
        uint8_t m_columns;
        
    public:
        
        // _____________________________________________________________________
        Sign(pos_t  const    & p_pos,
             ticket_t const  & p_msg,
             block_ref const & p_block,
             uint8_t           p_rows    = 1,
             uint8_t           p_columns = 1)
        
        : pos_t(p_pos),
          m_msg(p_msg),
          m_block(p_block)
        {
            SetSize(p_rows, p_columns);
        }
        
        // _____________________________________________________________________
        //! \brief      Property to access the sign's text message handle.
        virtual ticket_t & Msg()
        {
            return m_msg;
        }
        
        // _____________________________________________________________________
        virtual block_ref & Block()
        {
            return m_block;
        }
        
        // _____________________________________________________________________
        virtual void SetSize(uint8_t p_rows, uint8_t p_columns)
        {
            if( !p_rows )
            {
                p_rows += 1;
            }
            
            if( !p_columns )
            {
                p_columns += 1;
            }
            
            m_rows    = p_rows;
            m_columns = p_columns;
        }
        
    };
    
    typedef list<Sign> SignList;

// =============================================================================

    class Staircase
    {
    public:
    
        Position<8> m_pos;
        
        uint16_t m_targetRoom;
        
        int m_type;
        
        // _____________________________________________________________________
        Staircase(uint16_t x = 0,
                  uint16_t y = 0,
                  Floor floor = BG2,
                  uint16_t targetRoom = 0,
                  int type = 0)
        
        : m_pos(x, y, floor),
          m_targetRoom(targetRoom),
          m_type(type)
        {
        
        }
        
        // _____________________________________________________________________
        Staircase(Staircase const &sc)
        
        : m_pos(sc.m_pos),
          m_targetRoom(sc.m_targetRoom),
          m_type(sc.m_type)
        {
        
        }
    };
    
    typedef list<Staircase> StairList;

// =============================================================================

    class Room
    {
    public:
        
        MetaData metaData;
        
        Thorn::BlockRef floor1;
        Thorn::BlockRef floor2;
        
        Ticket<ObjectList> m_layout;
        
        Header header;
        
        // repositories for all our object data
        ObjectData objects[3];
        
        SpriteList sprites;
        
        ItemList items;
        
        BlockList blocks;
        
        TorchList torches;
        
        ChestList chests;
        ChestList bigChests;
        
        SignList m_signs;
        
        StairList stairs;
        
        StarList stars;
        
        LockList locks;
        
    public:
        
        Room();
    };
    
    typedef list<Room*> RoomList;
    typedef deque< Show<Room> > RoomDeque;

// =========================================================
    
    class Entrance
    {
        
    public:
        
        MetaData metaData;
        
        bool m_startingPoint;
        
        uint8_t mainGraphics;
        
        uint8_t song;
        uint8_t startingFloor;
        uint8_t startingBg;
        uint8_t palace;
        uint8_t quadrant1;
        uint8_t quadrant2;
        
        // only used for normal entrances
        uint8_t doorOrientation;
        
        uint8_t relativeCoords[8];
        
        uint16_t doorSettings; // make own type?
        
        // only used if it's a starting point
        uint16_t associatedEntrance;
        
        Position<1> scroll;
        Position<1> player;
        Position<1> camera;
        
        Ticket<Room> room;
        
    public:
        
        // consider doing some template magic with this to exclude different
        // functionality... (magic of the Alexandrescu variety)
        Entrance(bool startingPoint = false)
        {
            m_startingPoint = startingPoint;
        }
    };
    
    typedef deque< Show<Entrance> > EntranceDeque;

// =============================================================================

    typedef TileAttr AttrSet[0x200];
    typedef TileAttr DefaultAttr[0x180];
    typedef TileAttr CustomAttr[0x80];

    typedef list<CustomAttr> CustomAttrList;

// =============================================================================

    struct Tile
    {
        TileAttr m_attr;
        
        TileEntry m_tile;
        
        Tile()
        
        : m_attr(0),
          m_tile()
        {
            
        }
    };
    
    /// test object
    class Tilemap
    {
        
    public:
    
        Matrix<Tile> m_bg1;
        Matrix<Tile> m_bg2;
        
    public:
    
        Tilemap()
        
        : m_bg1(0x40, 0x40),
          m_bg2(0x40, 0x40)
        {
        
        
        }
        
        static uint16_t ToX(uint16_t value)
        {
            return ((value & 0x7E) << 3);
        }
        
        static uint16_t ToY(uint16_t value)
        {
            return ((value & 0x1F80) >> 4);
        }
            
    };

// =============================================================================

    class State
    {
    protected:
    
        bool darkWorld;
        
        string name;
        
        AttrSet attrTable;
        
        Tilemap maps[2];
        
    public:
        
        State(Entrance const &entr,
              Room const     &room,
              ObjectList     &layout);
        
        void ParseObject(Object &obj);
    };

// =============================================================================

}

#endif
