
    #include <fstream>
    #include <string.h>
    
    #include "core.h"
    
    #include "expandinfo.h"
    
// ===============================================================

    int FileSize(std::ifstream &f)
    {
        /// f.open(sFileName, std::ios_base::binary | std::ios_base::in);
        
        if(!f.good() || f.eof() || !f.is_open())
        {
            return 0;
        }
        
        f.seekg(0, std::ios_base::beg);
        
        std::ifstream::pos_type begin_pos = f.tellg();
        
        f.seekg(0, std::ios_base::end);
        
        std::ifstream::pos_type end_pos = f.tellg();
        
        f.seekg(0, std::ios_base::beg);
        
        return static_cast<int>(end_pos - begin_pos);
    }
    
// ===============================================================
    
    const uint32_t ExpandInfo::expand_version = 1;
    
// ===============================================================
    
    ExpandInfo::ExpandInfo(std::ifstream &zeldaFile)
    {
        int fileLength = FileSize(zeldaFile);
        
        // ---------------------------------
        
        m_data = new Buffer(0x200);
        
        m_headerBuf = new Buffer(0x200);
        
        if(fileLength & 0x200)
        {
            m_hasHeader = true;
            
            zeldaFile.read(reinterpret_cast<char*>(m_headerBuf->contents),
                           0x200);
            
            fileLength -= 0x200;
        }
        
        m_rom = new Buffer(fileLength);
        
        zeldaFile.read(reinterpret_cast<char*>(m_rom->contents),
                       fileLength);
        
        Initialize();
    }
    
// =============================================================================
    
    ExpandInfo::~ExpandInfo()
    {
        delete m_rom;
        
        if(m_hasHeader)
        {
            delete m_headerBuf;
        }
        
        delete m_data;
    }
    
// =============================================================================
    
    void ExpandInfo::Initialize()
    {
        // Initialize the Expansion Information header with default values from
        // a fresh rom
        
        //! double check to make sure the null terminator is added during 
        //! debugging 
        strcpy(m_designation, "EXPAND_INFO");
        
        m_header.version         = expand_version;
        
        m_dungeon.headerOffset   = m_rom->Get3Bytes(asm_header_ref);
        m_dungeon.headerBank     = m_rom->GetByte(asm_header_bank);
        m_dungeon.objOffset      = m_rom->Get3Bytes(asm_object_ref);
        m_dungeon.layoutOffset   = m_rom->Get3Bytes(asm_layout_ref);
        
        //! (why not loading anything from rom?)
        m_dungeon.entranceOffset = asm_entrance_ref;
        
        m_dungeon.chestOffset    = m_rom->Get3Bytes(asm_chest_ref);
        m_dungeon.numChests      = m_rom->Get2Bytes(asm_num_chests) / 3;
        m_dungeon.spriteOffset   = m_rom->Get2Bytes(asm_sprite_ref);
        
        m_overworld.map32To16UL  = m_rom->Get3Bytes(asm_map32To16UL);
        m_overworld.map32To16UR  = m_rom->Get3Bytes(asm_map32To16UR);
        m_overworld.map32To16LL  = m_rom->Get3Bytes(asm_map32To16LL);
        m_overworld.map32To16LR  = m_rom->Get3Bytes(asm_map32To16LR);
        
        // Check to see if there is a hooked JML to extended regions
        // If there is a hook, this will be the offset in-rom for the bank table
        if(m_rom->GetByte(asm_sprite_hook) == 0x5C)
        {
            uint32_t hook = m_rom->Get3Bytes(asm_sprite_hook + 1);
            
            m_dungeon.spriteBanks = CpuToRomAddr(hook) + 0x0E;
        }
        else
        {
            // In this case there isn't. And in that case the sprite data is all
            // in bank 09
            m_dungeon.spriteBanks = 0x090000;
        }
        
        m_overworld.holeOffset    = 0x1BB800;
        m_overworld.numHoles      = 0x13;
        
        m_overworld.entrOffset    = 0x1BB96F;
        m_overworld.numEntr       = 0x81;
        
        m_overworld.exitOffset    = 0x02DD8A;
        m_overworld.numExits      = 0x4F;
        
        m_overworld.sprOffset0    = 0x9C881;
        m_overworld.sprOffset2    = 0x9C901;
        m_overworld.sprOffset3    = 0x9CA21;
        
        // default location of this data.
        m_overworld.itemOffset    = 0x1BC2F9;
        
        // initialize default information about the structure of the graphics data
        m_graphics.sprOffset      = m_rom->Get2Bytes(asm_spr_gfx_ptr);
        m_graphics.bgOffset       = m_rom->Get2Bytes(asm_bg_gfx_ptr);
    }

// ===============================================================

    bool ExpandInfo::CheckForExpansion()
    {
        // This is the alleged header offset.
        uint32_t offset = CpuToRomAddr( m_rom->Get4Bytes(expand_info_loc) );
        
        // ------------------------------
        
        if( (offset < 0x100000) || (offset > m_rom->length) )
            return false;
        
        return Deserialize(offset);
    }
    
// ===============================================================

    void ExpandInfo::Serialize()
    {
        int i = 0;
        
        // ----------------------------
        
        // initialize pointer for writing into binary version of the header
        m_data->wOffset = 0;
        
        // Copy the magic string to the buffer
        for(i = 0; i < 12; ++i)
        {
            m_data->PutByte(i, m_designation[i]);
        }
        
        m_data->Put4Bytes(m_header.version);
        m_data->Put4Bytes(m_header.size);
        
        m_data->Put4Bytes(m_dungeon.headerOffset);
        m_data->Put4Bytes(m_dungeon.headerBank);
        m_data->Put4Bytes(m_dungeon.objOffset);
        m_data->Put4Bytes(m_dungeon.entranceOffset);
        m_data->Put4Bytes(m_dungeon.chestOffset);
        m_data->Put4Bytes(m_dungeon.numChests);
        m_data->Put4Bytes(m_dungeon.spriteOffset);
        m_data->Put4Bytes(m_dungeon.spriteBanks);
        m_data->Put4Bytes(m_dungeon.layoutOffset);
        
        m_data->Put4Bytes(m_text.offset);
        m_data->Put4Bytes(m_text.numMsgs);
        
        m_data->Put4Bytes(m_overworld.numParts);
        m_data->Put4Bytes(m_overworld.holeOffset);
        m_data->Put4Bytes(m_overworld.numHoles);
        m_data->Put4Bytes(m_overworld.entrOffset);
        m_data->Put4Bytes(m_overworld.numEntr);
        m_data->Put4Bytes(m_overworld.itemOffset);
        m_data->Put4Bytes(m_overworld.sprOffset0);
        m_data->Put4Bytes(m_overworld.sprOffset2);
        m_data->Put4Bytes(m_overworld.sprOffset3);
        m_data->Put4Bytes(m_overworld.map32To16UL);
        m_data->Put4Bytes(m_overworld.map32To16UR);
        m_data->Put4Bytes(m_overworld.map32To16LL);
        m_data->Put4Bytes(m_overworld.map32To16LR);
        
        m_data->Put4Bytes(m_graphics.sprOffset);
        m_data->Put4Bytes(m_graphics.bgOffset);
    }
    
// ===============================================================

    bool ExpandInfo::Deserialize(uint32_t offset)
    {
        uint8_t tempByte = 0;
        
        uint8_t validDesignation[12] = "EXPAND_INFO";
        
        int i = 0;
        
        // ----------------------------
        
        // Copy the alleged expansion info data from the rom from the given location
        // to our local buffer
        Buffer::Copy(m_data, m_rom, 0, offset, 0x200);
        
        // reset read pointer so we can do sequential reads from the data buffer
        m_data->rOffset = 0;
        
        // Copy the magic string to the buffer
        for(i = 0; i < 12; ++i)
        {
            tempByte = (uint8_t) m_data->GetByte();
            
            if(tempByte == validDesignation[i])
            {
                m_designation[i] = (uint8_t) m_data->GetByte();
            }
            else
            {
                // it clearly isn't valid expansion info if the magic string doesn't match
                return false;
            }
        }
        
        m_header.version             = m_data->Get4Bytes();
        m_header.size                = m_data->Get4Bytes();
        
        m_dungeon.headerOffset       = m_data->Get4Bytes();
        m_dungeon.headerBank         = m_data->Get4Bytes();
        m_dungeon.objOffset          = m_data->Get4Bytes();
        m_dungeon.entranceOffset     = m_data->Get4Bytes();
        m_dungeon.chestOffset        = m_data->Get4Bytes();
        m_dungeon.numChests          = m_data->Get4Bytes();
        m_dungeon.spriteOffset       = m_data->Get4Bytes();
        m_dungeon.spriteBanks        = m_data->Get4Bytes();
        m_dungeon.layoutOffset       = m_data->Get4Bytes();
        
        m_text.offset                = m_data->Get4Bytes();
        m_text.numMsgs               = m_data->Get4Bytes();
        
        m_overworld.numParts         = m_data->Get4Bytes();
        
        m_overworld.holeOffset        = m_data->Get4Bytes();
        m_overworld.numHoles          = m_data->Get4Bytes();
        
        m_overworld.entrOffset        = m_data->Get4Bytes();
        m_overworld.numEntr           = m_data->Get4Bytes();
        
        m_overworld.itemOffset        = m_data->Get4Bytes();
        
        m_overworld.sprOffset0        = m_data->Get4Bytes();
        m_overworld.sprOffset2        = m_data->Get4Bytes();
        m_overworld.sprOffset3        = m_data->Get4Bytes();
        
        m_overworld.map32To16UL       = m_data->Get4Bytes();
        m_overworld.map32To16UR       = m_data->Get4Bytes();
        m_overworld.map32To16LL       = m_data->Get4Bytes();
        m_overworld.map32To16LR       = m_data->Get4Bytes();
        
        m_graphics.sprOffset       = m_data->Get4Bytes();
        m_graphics.bgOffset        = m_data->Get4Bytes();
        
        return true;
    }

// ===============================================================

