
    #include <deque>
    
    #include "inttypes.h"
    
    #include "containers.h"
    #include "core.h"

#ifndef GRAPHICS_H

    #define GRAPHICS_H

namespace Graphics
{
    using std::deque;
    using std::exception;
    
    using namespace Containers;
    using namespace Memory;
    
// =============================================================================

    enum PalRef
    {
        pal0 = 0,
        pal1,
        pal2,
        pal3,
        pal4,
        pal5,
        pal6,
        pal7        
    };
    
// =============================================================================

    typedef uint8_t TileAttr;
    
// =============================================================================

    class TileEntry
    {
        
    protected:
        
        uint16_t m_chr;
        
        
        void Configure(uint16_t raw)
        {
            m_vFlip     = (raw & 0x8000) ? true : false;
            m_hFlip     = (raw & 0x4000) ? true : false;
            m_priority  = (raw & 0x2000) ? true : false;
            
            m_pal = (PalRef) ((raw & 0x1c00) >> 10);
            
            SetChr(raw);
        }
        
    public:
        
        bool m_vFlip;
        bool m_hFlip;
        bool m_priority;
        
        PalRef m_pal;
        
        explicit TileEntry(bool     p_hFlip    = false,
                           bool     p_vFlip    = false,
                           bool     p_priority = false,
                           PalRef   p_pal      = pal0,
                           uint16_t p_chr      = 0)
        
        : m_vFlip(p_vFlip),
          m_hFlip(p_hFlip),
          m_priority(p_priority),
          m_pal(p_pal)
        {
            SetChr(p_chr);
        }
        
        TileEntry(uint16_t raw)
        {
            Configure(raw);
        }
        
        TileEntry & operator = (uint16_t raw)
        {
            Configure(raw);
            
            return (*this);
        }
        
        uint16_t GetChr()
        {
            return m_chr;
        }
        
        void SetChr(uint16_t chr)
        {
            m_chr = (chr % 0x400);
        }
    };    
    
// =============================================================================

    enum PixelDepth
    {
        //! SNES native 2bpp bit depth.
        Pixel2bpp  = 2,
        
        //! ALTTP's softare based 3bpp bit depth.
        Pixel3bpp  = 3,
        
        //! SNES native 4bpp bit depth.
        Pixel4bpp  = 4,
        
        //! Solely for use with Palette::Color
        Pixel5bpp  = 5,
        
        //! SNES native 8bpp bit depth.
        Pixel8bpp  = 8,
        
        //! Mode 7 8bpp bit depth (for template specialization).
        PixelMode7 = 8
    };
    
// =============================================================================
    
    /**
        \brief      Template based pixel class with a compile time determined
                    pixel depth.
                    
        \since      0.11
    */
    template<PixelDepth bitDepth>
    class Pixel
    {
            
    protected:
        
        uint8_t m_value;
        
        class Problem : exception
        {
            
        };
        
    public:
        
        // _____________________________________________________________________
        void Set(uint8_t value)
        {
            if( value > (1 << (int) bitDepth) )
            {
                throw Problem();
            }
            else
            {
                m_value = value;
            }
        }
        
        // _____________________________________________________________________
        Pixel(uint8_t value = 0)
        {
            Set(value);
        }
        
        // _____________________________________________________________________
        Pixel(Pixel const &p)
        {
            Set(p.Get());
        }
        
        // _____________________________________________________________________
        uint8_t Get() const
        {
            return m_value;
        }
        
        // _____________________________________________________________________
        uint8_t operator = (uint8_t value)
        {
            Set(value);
            
            return m_value;
        }
        
        // _____________________________________________________________________
        void operator |= (uint8_t value)
        {
            Set(m_value | value);
        }
        
        // _____________________________________________________________________
        operator uint8_t()
        {
            return m_value;
        }
    };

// =============================================================================

    template<PixelDepth bitDepth>
    class Tile
    {
        
    protected:
        
        Matrix< Pixel<bitDepth> > m_colors;
        
    public:
        
        Tile() : m_colors(8, 8)
        {
            
        }
        
        // _____________________________________________________________________
        // for accessing individual pixels of the tile
        Pixel<bitDepth> & operator () (uint32_t row,
                                       uint32_t column,
                                       bool     hFlip = false,
                                       bool     vFlip = false)
        {
            if(hFlip) { column = (m_colors.m_columns - 1) - column; }
            if(vFlip) { row    = (m_colors.m_rows    - 1) - row;   }
            
            return m_colors(row, column);
        }
    };
    
// =============================================================================

    /**
        \brief      Template class that serves as a container for a fixed number
                    of tiles of a given bit depth.
        
        \author     MathOnNapkins (5/8/2012)
        
        \since      0.11
     */
    template<PixelDepth bitDepth, size_t numTiles>
    class TileSet
    {
        
    protected:
        
        Tile<bitDepth> m_tiles[numTiles];
        
        bool m_compressed;
        
    public:
        
        MetaData metaData;
        
        class Problem : exception
        {
            
        };
        
        TileSet(bool compressed = false) : m_compressed(compressed)
        {
        
        }
        
        uint32_t NumTiles()
        {
            return numTiles;
        }
        
        Tile<bitDepth> & operator () (uint32_t i)
        {
            if(i >= numTiles)
            {
                throw Problem();
            }
            
            return m_tiles[i];
        }
    };
    
    // The most common tile set types
    typedef TileSet<Pixel3bpp, 0x40> TileSet3bpp;
    typedef TileSet<Pixel2bpp, 0x80> TileSet2bpp;
    
// =============================================================================
    
    struct TileData
    {
        // Specific graphics sets that don't fall into the two above categories
        TileSet<Pixel4bpp, 0x380> linkGfx;
        
        TileSet<Pixel2bpp, 0x100> fontGfx;
        
        TileSet<Pixel8bpp, 0x100> mode7Gfx;
        
        // All the background graphics packs
        deque< Show<TileSet3bpp> > bg;
        
        // All the sprite graphics packs
        deque< Show<TileSet3bpp> > spr;
        
        // 2bpp graphics packs for hud and legend
        deque< Show<TileSet2bpp> > legend;
        
        deque< Show<TileSet3bpp> > expansion;
        
        TileData() : linkGfx(false), fontGfx(false), mode7Gfx(false)
        {
            
        }
    };

// =============================================================================

    /** 
        \brief      An object that contains references to a fixed number of
                    tilesets.
       
    */
    template<class TileSetType, size_t numSets>
    class TileTheme
    {
        
    protected:
        
        Ticket<TileSetType> m_sets[numSets];
        
    public:
        
        MetaData metaData;
        
        // _____________________________________________________________________
        class Problem : exception
        {
            
        };        
        
        // _____________________________________________________________________
        int NumSets()
        {
            return numSets;
        }
        
        // _____________________________________________________________________
        TileTheme()
        {
            
        }
        
        // _____________________________________________________________________
        Ticket<TileSetType> & operator [] (int i)
        {
            if(i >= numSets)
            {
                throw Problem();
            }
            
            return m_sets[i];
        }    
    };

// =============================================================================

    //! \brief      A set of 8 groups of 64 3bpp tiles.
    typedef TileTheme<TileSet3bpp, 8> BgTileMain;
    
    //! \brief      A set of 4 groups of 64 3bpp tiles.
    typedef TileTheme<TileSet3bpp, 4> BgTileAux;
    
    //! \brief      A set of 4 groups of 64 3bpp tiles.
    typedef TileTheme<TileSet3bpp, 4> SprTileMain;
    
    //! \brief      A set of 2 groups of 64 3bpp tiles.
    typedef TileTheme<TileSet3bpp, 2> SprTileAux;
    
    struct TileThemes
    {
        deque< Show<BgTileMain>  > bgMain;
        deque< Show<BgTileAux>   > bgAux;
        deque< Show<SprTileMain> > sprMain;
        deque< Show<SprTileAux>  > sprAux;
    };

// =============================================================================

    struct TileConfig
    {
        BgTileMain bgMain;
        BgTileAux bgAux;
        
        SprTileMain sprMain;
        SprTileAux sprAux;
    };    

// =============================================================================

    struct Color
    {
        Pixel<Pixel8bpp> m_alpha;
        Pixel<Pixel5bpp> m_red;
        Pixel<Pixel5bpp> m_green;
        Pixel<Pixel5bpp> m_blue;
        
        Color(uint16_t rawColor = 0,
              uint8_t  alpha    = 0xFF)
        {
            m_alpha = alpha;
            m_red   = (rawColor >>  0) & 0x1F;
            m_green = (rawColor >>  5) & 0x1F;
            m_blue  = (rawColor >> 10) & 0x1F;
        }
        
        operator uint16_t()
        {
            uint16_t r = ((uint16_t) m_red)   <<  0;
            uint16_t g = ((uint16_t) m_green) <<  5;
            uint16_t b = ((uint16_t) m_blue)  << 10;
            
            return (r | g | b);
        }
        
        operator uint32_t()
        {
            uint32_t a = ((uint32_t) m_alpha) << (24 + 0);
            uint32_t r = ((uint32_t) m_red)   << (16 + 3);
            uint32_t g = ((uint32_t) m_green) << (8  + 3);
            uint32_t b = ((uint32_t) m_blue)  << (0  + 3);
            
            return (a | r | g | b);
        }
    };

// =============================================================================

    // palette offsets. corresponds to positions in CGRAM (palette RAM)
    // BP = background palette
    // SP = sprite palette
    enum
    {
        BP0 = 0x00,
        BP1 = 0x20,
        BP2 = 0x40,
        BP3 = 0x60,
        BP4 = 0x80,
        BP5 = 0xA0,
        BP6 = 0xC0,
        BP7 = 0xE0,
        SP0 = 0x100,
        SP1 = 0x120,
        SP2 = 0x140,
        SP3 = 0x160,
        SP4 = 0x180,
        SP5 = 0x1A0,
        SP6 = 0x1C0,
        SP7 = 0x1E0
    };

// =============================================================================

    template<size_t N>
    class Palette
    {
    protected:
    
        Color colors[N];
    
    public:
        
        class PaletteException : exception
        {
        
        };
        
        uint32_t Length()
        {
            return N;
        }
        
        uint32_t Size()
        {
            return (N * 2);
        }
        
        Color & operator [] (const uint32_t index)
        {
            if(index >= N)
            {
                throw PaletteException();
            }
            
            return colors[index];
        }
    };

// =============================================================================

    //! \brief 
    //!  
    template<size_t palSize, size_t palCount>
    struct Group
    {
        //! \brief      Type definition
        typedef deque< Show<Group> > ShowDeque;
       
        // _____________________________________________________________________
        class GroupException : exception { };
        
        // _____________________________________________________________________
        // Members.
        
        Palette<palSize> p[palCount];
        
        MetaData metaData;
        
        // _____________________________________________________________________
        // Member functions.
        
        // _____________________________________________________________________
        Palette<palSize> & operator [] (const int index)
        {
            if( ( index >= 0) && (index < palCount) )
            {
                return p[index];
            }
            else
            {
                throw GroupException();
            }
        }
        
        // _____________________________________________________________________
        //! \brief      The number of palettes in the group.
        uint32_t PaletteCount()
        {
            return palCount;
        }
        
        // _____________________________________________________________________
        //! \brief      The number of colors in each palette.
        uint32_t PaletteSize()
        {
            return palSize;
        }
        
        // _____________________________________________________________________
        //! \brief      Total size of the group expressed in number of colors.
        uint32_t Length()
        {
            return (palSize * palCount);
        }
        
        // _____________________________________________________________________
        //! \brief      Total size of the group expressed in bytes.
        uint32_t Size()
        {
            return Length() * 2;
        }
    };

// =============================================================================

    struct PaletteData
    {
        Group<4, 8>::ShowDeque hud;
        
        Group<3, 1>::ShowDeque  sword;
        Group<4, 1>::ShowDeque  shield;
        Group<15, 1>::ShowDeque armor;
        
        struct tag_overworld
        {
            Group<7, 5>::ShowDeque primary;
            Group<7, 3>::ShowDeque secondary;
            Group<7, 1>::ShowDeque tertiary;
        }
        overworld;
        
        struct tag_dungeon
        {
            Group<15, 6>::ShowDeque primary;
        }
        dungeon;
        
        struct tag_sprite
        {
            Group<15, 4>::ShowDeque primary;
            Group<7, 1>::ShowDeque secondary;
            Group<7, 1>::ShowDeque tertiary;
            Group<7, 1>::ShowDeque misc;
        }
        sprite;
        
        struct tag_palaceMap
        {
            Group<16, 6>::ShowDeque bg;
            Group<7, 3>::ShowDeque  sprite;
        }
        palaceMap;
    };

// =============================================================================

    struct DungTheme
    {
    
        Ticket< Group<15, 6> > bgPrimary;
        Ticket< Group<7, 1> >  sprSecondary1;
        Ticket< Group<7, 1> >  sprSecondary2;
        Ticket< Group<7, 1> >  sprTertiary;
    };

// =============================================================================

    struct OverBgTheme
    {
        Ticket< Group<7, 5> > primary;
        Ticket< Group<7, 3> > secondary1;
        Ticket< Group<7, 3> > secondary2;
        Ticket< Group<7, 1> > tertiary;
    };

// =============================================================================

    struct OverSprTheme
    {
        Ticket< Group<7, 1> > secondary1;
        Ticket< Group<7, 1> > secondary2;
    };

// =============================================================================

    struct PaletteThemes
    {
        deque< Show<DungTheme>    > dungeon;
        deque< Show<OverBgTheme>  > overworldBg;
        deque< Show<OverSprTheme> > overworldSpr;
    };

// =============================================================================

    /**
        \brief      Whereas the aforementioned ordinary Tile class is but a
                    matrix of pixel indices, this class represents a fully
                    realized, colored 8x8 SNES tile. It consists of a matrix of
                    5 bits per channel (red, green, and blue), along with 8 bits
                    of transparency information. It is intended as a helper
                    class for interfacing with a graphical level application.
                    
    */
    class ColorTile
    {
        
    protected:
        
        //! \brief      The internal container for the 8x8 grid of colored pixel
        //!             data.
        Matrix<Color> m_colors;
        
    public:
        
        // _____________________________________________________________________
        //! \brief      This default constructor generates a completely black
        //!             tile with full alpha channel (fully opaque).
        ColorTile() : m_colors(8, 8)
        {
            
        }
        
        // _____________________________________________________________________
        //! \brief      This constructor accepts a colorless tile and a palette
        //!             with which to intantiate an colored tile. 
        //!             
        //! \since      0.11
        template<PixelDepth depth, size_t palSize>
        ColorTile(Tile<depth> &t, Palette<palSize> &p, int startIndex)
        
        : m_colors(8, 8)
        {
            if(palSize > (1 << depth) )
            {
                throw exception();
            }
            
            for(int row = 0; row < 8; ++row)
            {
                for(int column = 0; column < 8; ++column)
                {
                    // quick note, startIndex is for cases like Palette<1,7>,
                    // where the first palette index is omitted b/c it refers to
                    // transparency
                    int palIndex = t(row, column);
                    
                    if( palIndex == 0)
                    {
                        // make the color completely transparent
                        m_colors(row, column).m_alpha = 0;
                    }
                    else
                    {
                        m_colors(row, column) = p[palIndex - startIndex];
                    }
                }
            }
        }
        
        // _____________________________________________________________________
        //! \brief      Overloaded operator for accessing individual pixels of
        //!             the tile.
        Color & operator () (uint32_t row,
                             uint32_t column,
                             bool hFlip = false,
                             bool vFlip = false)
        {
            if(hFlip) { column = (m_colors.m_columns - 1) - column; }
            if(vFlip) { row    = (m_colors.m_rows    - 1) - row;   }
            
            return m_colors(row, column);
        }
    };

// =============================================================================

}

#endif