
    #include <memory>
    #include <sstream>
    
    #include "loaders.h"
    
namespace Loaders
{
    
// ===============================================================

    void RomLoader::LoadGame()
    {
        LoadThornBlocks();
        
        LoadDialogue();
        
        LoadTileSets();
        LoadTileThemes();
        
        LoadPalettes();
        LoadPaletteThemes();
        
        LoadLayouts();

        LoadDungeonRooms();
        
        LoadDungeonEntrances();
        
        LoadOverworldAreas();
    }

// ===============================================================
    
    // Should only be used for 2bpp, 4bpp, and 8bpp native formats
    template<PixelDepth depth>
    void RomLoader::LoadTile(Tile<depth> &tile,
                             Buffer      *b,
                             uint32_t     position)
    {
        int offsets[8] = { 0x00, 0x01, 0x10, 0x11, 0x20, 0x21, 0x30, 0x31 };
        
        // -----------------------------------------
        
        for(int row = 0; row < 8; ++row)
        {
            uint8_t mask = 0x80;
            
            for
            (
                int column = 0;
                column < 8;
                column += 1, mask >>= 1
            )
            {
                Pixel<depth> p;
                
                for(int bitPlane = 0; bitPlane < depth; ++bitPlane)
                {
                    if(b->GetByte(position + offsets[bitPlane]) & mask)
                    {
                        p |= static_cast<uint8_t> (1 << bitPlane);
                    }
                }
            }
            
            // readjust offsets after every row (horizontal line) of the tile
            for(int i = 0; i < depth; ++i)
            {
                offsets[i] += 2;
            }
        }
    }

// ===============================================================
    
    // Should only be used for 3bpp abstract format (2bpp x 1bpp composite in Zelda. It's not native
    // to the SNES)
    template<>
    void RomLoader::LoadTile<Pixel3bpp>(Tile<Pixel3bpp> &tile,
                                        Buffer          *b,
                                        uint32_t         position)
    {
        int offsets[3] = { 0x00, 0x01, 0x10 };
        
        // -----------------------------------------
        
        for(int row = 0; row < 8; ++row)
        {
            uint8_t mask = 0x80;
            
            for
            (
                int column = 0;
                column < 8;
                column += 1, mask >>= 1
            )
            {
                Pixel<Pixel3bpp> p;
                
                for(int bitPlane = 0; bitPlane < Pixel3bpp; ++bitPlane)
                {
                    if(b->GetByte(position + offsets[bitPlane]) & mask)
                    {
                        p |= static_cast<uint8_t> (1 << bitPlane);
                    }
                }
                
                tile(row, column) = p;
            }
            
            offsets[0] += 2;
            offsets[1] += 2;
            offsets[2] += 1;
        }
    }

// ===============================================================
    
    // Should be noted that this is for Mode 7 format tiles, not the bitplaned
    // 8bpp tiles. ( not used in ALTTP)
    template<>
    void RomLoader::LoadTile(Tile<PixelMode7> &tile,
                             Buffer           *b,
                             uint32_t          position)
    {
        for(int row = 0; row < 8; ++row)
        {
            for(int column = 0; column < 8; ++column)
            {
                Pixel<PixelMode7> p = b->GetByte(position);
                
                tile(row, column) = p;
                
                position += 1;
            }
        }
    }

// ===============================================================
    
    template<PixelDepth depth, size_t numTiles>
    void RomLoader::LoadTileSet(TileSet<depth, numTiles> &ts,
                                Buffer                   *b,
                                uint32_t                  base)
    {
        // (8 pixels * 8 pixels * depth bits * 1 byte)
        //                        ----------   ------
        //                        1 pixels   * 8 bits)
        uint32_t bytesPerTile = (8 * 8 * depth) / 8;
        
        uint32_t tileCapacity = (b->length - base) / bytesPerTile;
        
        // ----------------------------------------------
        
        // if not an exact match, we get mad!
        if(tileCapacity < numTiles)
        {
            throw typename TileSet<depth, numTiles>::Problem();
        }
        
        for(int i = 0; i < numTiles; ++i)
        {
            LoadTile<depth>(ts(i), b, base + (bytesPerTile * i) );
        }
    }

// ===============================================================
    
    template<PixelDepth depth, size_t numTiles>
    void RomLoader::LoadTileSet(deque< Show< TileSet<depth, numTiles> > > &d,
                                Buffer *b,
                                bool compressed,
                                uint32_t base)
    {
        // (8 pixels * 8 pixels * depth bits * 1 byte)
        //                        ----------   ------
        //                        1 pixels   * 8 bits)
        uint32_t bytesPerTile = (8 * 8 * depth) / 8;
        
        uint32_t tileCapacity = (b->length - base) / bytesPerTile;
        
        typedef TileSet<depth, numTiles> TileSetType;
        
        TileSetType tileSet(compressed);
        
        // ----------------------------------------------
        
        // if not an exact match, we get mad!
        if(tileCapacity < numTiles)
        {
            throw typename TileSetType::Problem();
        }
        
        for(int i = 0; i < numTiles; ++i)
        {
            LoadTile<depth>(tileSet(i), b, base + (bytesPerTile * i) );
        }
        
        d.push_back( Show<TileSetType>(tileSet) );
    }

// ===============================================================
    
    int RomLoader::GetSprGfxPtr(int index)
    {
        uint32_t base = CpuToRomAddr(default_spr_gfx_ptr);
        uint32_t source  = 0;
        
        // -------------------------------------
        
        // broke ass old method in the original game
        source  = m_rom->GetByte(base               + index) << 0x10;
        source |= m_rom->GetByte(base + 0xDF        + index) << 0x08;
        source |= m_rom->GetByte(base + (0x0DF * 2) + index);
        
        source = CpuToRomAddr(source);
        
        return source;
    }

// ===============================================================

    int RomLoader::GetBgGfxPtr(int index)
    {
        uint32_t base = CpuToRomAddr(default_bg_gfx_ptr);
        uint32_t source  = 0;
        
        // -------------------------------------
        
        // broke ass old method in the original game
        source  = m_rom->GetByte(base              + index) << 0x10;
        source |= m_rom->GetByte(base + 0xDF       + index) << 0x08;
        source |= m_rom->GetByte(base + (0xDF * 2) + index);
        
        source = CpuToRomAddr(source);
        
        return source;
    }

// ===============================================================
    
    void RomLoader::LoadTileSets()
    {
        uint32_t i           = 0;
        uint32_t offset      = 0;
        
        TileData &td = m_zg.tileData;
        
        // ----------------------------------
        
        // decompress all sprite graphics
        for(i = 0; i < 0x6C; ++i)
        {
            offset = GetSprGfxPtr(i);
            
            if(i < 0x0C)
            {
                // The first 12 graphics sets are uncompressed in the rom, hence
                // the differing treatment.
                
                LoadTileSet<Pixel3bpp, 0x40>(td.spr, m_rom, false, offset);
            }
            else
            {
                // The rest are compressed...
                
                Buffer *temp = DecompressStandard(m_rom, offset);
                
                switch(i)
                {
                    case 0x67:
                    case 0x68:
                    case 0x69:
                    case 0x6A:
                    case 0x6B:
                        
                        LoadTileSet<Pixel2bpp, 0x80>(td.legend, temp, true);
                        
                        break;
                        
                    default:
                        
                        LoadTileSet<Pixel3bpp, 0x40>(td.spr, temp, true);
                        
                        break;
                }
                
                delete temp;
            }
        }
        
        // decompress all background graphics
        for(i = 0; i < 0x73; ++i)
        {
            offset = GetBgGfxPtr(i);
            
            Buffer *temp = DecompressStandard(m_rom, offset);
            
            switch(i)
            {
                case 0x71:
                case 0x72:
                    
                    LoadTileSet<Pixel2bpp, 0x80>(td.legend, temp, true);
                    
                    break;
                    
                default:
                    
                    LoadTileSet<Pixel3bpp, 0x40>(td.bg, temp, true);
                    
                    break;
            }
            
            delete temp;
        }
        
        // Load the miscellaneous graphics types (2bpp font, Link's 4bpp gfx,
        // and the mode7 gfx) Note that these graphics were all left
        // uncompressed in the ROM.
        LoadTileSet<Pixel2bpp, 0x100>(td.fontGfx, m_rom, 0x70000);
        LoadTileSet<Pixel4bpp, 0x380>(td.linkGfx, m_rom, 0x80000);
        LoadTileSet<PixelMode7, 0x100>(td.mode7Gfx, m_rom, 0xC4000);
        
        return;
    }

// ===============================================================
    
    void RomLoader::LoadTileThemes()
    {
        using namespace Graphics;
        using namespace Containers;
        
        // Load main bg tile themes
        for(int i = 0; i < 0x25; ++i)
        {
            uint32_t addr = 0x6073 + (i * 8);
            
            BgTileMain theme;
            
            // -----------------------------------------
            
            for(int ts = 0; ts < theme.NumSets(); ++ts)
            {
                uint16_t index = m_rom->GetByte(addr + ts);
                
                // ----------------------------------------------
                
                if(index < m_zg.tileData.bg.size())
                {
                    theme[ts] = m_zg.tileData.bg[index].Issue();
                }
                else
                {
                    index = 0;
                }
            }
            
            m_zg.tileThemes.bgMain.push_back( Show<BgTileMain>(theme) );
        }
        
        // Load auxiliary bg tile themes
        for(int i = 0; i < 0x52; ++i)
        {
            uint32_t addr = 0x5D97 + (i * 4);
            
            BgTileAux theme;
            
            // ------------------------------------
            
            for(int ts = 0; ts < theme.NumSets(); ++ts)
            {
                uint16_t index = m_rom->GetByte(addr + ts);
                
                // ----------------------------------------------
                
                if(index < m_zg.tileData.bg.size())
                {
                    if(index != 0)
                    {
                        theme[ts] = m_zg.tileData.bg[index].Issue();
                    }
                }
                else
                {
                    index = 0;
                }
            }
            
            m_zg.tileThemes.bgAux.push_back( Show<BgTileAux>(theme) );
        }
        
        // load main sprite tile themes
        for(int i = 0; i < 0x90; ++i)
        {
            uint32_t addr = 0x5B57 + (i * 4);
            
            SprTileMain theme;
            
            // -------------------------------------
            
            for(int ts = 0; ts < theme.NumSets(); ++ts)
            {
                uint16_t index = m_rom->GetByte(addr + ts);
                
                // ------------------------------------------
                
                if( index < m_zg.tileData.spr.size() )
                {
                    theme[ts] = m_zg.tileData.spr[index].Issue();
                }
                else
                {
                    index = 0;
                }                
            }
            
            m_zg.tileThemes.sprMain.push_back( Show<SprTileMain>(theme) );
        }
        
        for(int i = 0; i < 0x03; ++i)
        {
            const char *names[3] =  {
                                        "Overworld - Light World",
                                        "Dungeon",
                                        "Overworld - Dark World"
                                    };
            
            uint16_t indices[3] = { 0x01, 0x0A, 0x0B };
            
            SprTileAux theme;
            
            // -----------------------------------
            
            theme.metaData.name = names[i];
            
            theme[0] = m_zg.tileData.spr[indices[i]].Issue();
            theme[1] = m_zg.tileData.spr[0x06].Issue();
            
            m_zg.tileThemes.sprAux.push_back( Show<SprTileAux>(theme) );
        }
    }

// ===============================================================

    template<size_t palSize,
             size_t palCount>
    void RomLoader::LoadGroup(uint32_t numGroups,
                              uint32_t addr,
                              typename Group<palSize, palCount>::ShowDeque &d)
    {
        using namespace Graphics;
        using namespace Containers;
        
        uint32_t i  = 0;
        
        uint32_t romAddr = CpuToRomAddr(addr);
        
        // ------------------------------------
        
        for(i = 0; i < numGroups; ++i)
        {
            Group<palSize, palCount> group;
            
            // ------------------------------------
            
            for(uint32_t g = 0; g < group.PaletteCount(); ++g)
            {
                Palette<palSize> &pal = group[g];
                
                for(uint32_t c = 0; c < pal.Length(); c += 1, romAddr += 2)
                {
                    pal[c] = m_rom->Get2Bytes(romAddr);
                }
            }
            
            d.push_back( Show< Group<palSize, palCount> >(group) );
        }
    }

// ===============================================================

    void RomLoader::LoadPalettes()
    {
        using namespace Graphics;
        using namespace Containers;
        
        uint32_t i      = 0;
        uint32_t addr   = 0;
        
        PaletteData &pd = m_zg.palData;
        
        // ------------------------------------
        
        // corresponds to $DEC9E
        // writes SP1 through SP4 (full)
        LoadGroup<15, 4>(2, 0x1BD218, pd.sprite.primary);
         
        // corresponds to $DED6E
        LoadGroup<7, 1>(11, 0x1BD446, pd.sprite.misc);
        
        // corresponds to $DECE4
        // Set SP6 (first half) 
        LoadGroup<7, 1>(24, 0x1BD4E0, pd.sprite.secondary);
        
        LoadGroup<7, 1>(12, 0x1BD39E, pd.sprite.tertiary);
        
        // Load Equipment palettes
        
        // corresponds to $DED03 (loads sword palette)
        // Set SP5 (second half), 1st three colors
        LoadGroup<3, 1>(5, 0x1BD360, pd.sword);
        
        // corresponds to $DED29 (loads shield palette)
        // Set SP5 (second half), last 4 colors
        LoadGroup<4, 1>(4, 0x1BD648, pd.shield);
        
        // corresponds to $DEDF9 (loads armor palette)
        LoadGroup<15, 1>(3, 0x1BD308, pd.armor);
        
        // Load Heads Up Display palettes
        LoadGroup<4, 8>(2, 0x1BD660, pd.hud);
        
        // ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
        // Load Overworld palettes
        
        // corresponds to $DEEC7, Sets BP2 through BP6 (first halves)
        // depends on $0AB3 on the overworld (which depends upon the area number
        LoadGroup<7, 5>(6, 0x1BE6C8, pd.overworld.primary);
        
        // corresponds to $DEEE8
        // load the aux 1/2 overworld palette pool
        LoadGroup<7, 3>(20, 0x1BE86C, pd.overworld.secondary);
        
        // corresponds to ....?
        // load the aux 3 overworld palette pool
        LoadGroup<7, 1>(14, 0x1BE604, pd.overworld.tertiary);
        
        // ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
        
        // Load Dungeon palettes
        LoadGroup<15, 6>(20, 0x1BD734, pd.dungeon.primary);
        
        // ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
        
        // Load Palace Map palettes
        LoadGroup<16, 6>(1, 0x1BE544, pd.palaceMap.bg);
        LoadGroup<7, 3> (1, 0x1BD70A, pd.palaceMap.sprite);
    }

// ===============================================================

    void RomLoader::LoadPaletteThemes()
    {
        using namespace Graphics;
        using namespace Containers;
        
        uint32_t addr = 0;
        
        PaletteData &pd = m_zg.palData;
        
        PaletteThemes &pt = m_zg.palThemes;
        
        // --------------------------------------------
        
        for(uint32_t i = 0; i < 0x29; ++i)
        {
            DungTheme t;
            
            // --------------------------------
            
            addr = 0x75460 + (i * 4);
            
            /// Clunky, I know... clean this up later maybe?
            uint32_t temp1 = m_rom->GetByte(addr + 0) / 2;
            uint32_t temp2 = m_rom->GetByte(addr + 1);
            uint32_t temp3 = m_rom->GetByte(addr + 2);
            uint32_t temp4 = m_rom->GetByte(addr + 3);
            
            if(temp1 < 0x80)
            {
                t.bgPrimary = pd.dungeon.primary[temp1].Issue();
            }
            
            if(temp2 < 0x80)
            {
                t.sprTertiary = pd.sprite.tertiary[temp2].Issue();
            }
            
            if(temp3 < 0x80)
            {
                t.sprSecondary1 = pd.sprite.secondary[temp3].Issue();
            }
            
            if(temp4 < 0x80)
            {
                t.sprSecondary2 = pd.sprite.secondary[temp4].Issue();
            }
            
            pt.dungeon.push_back( Show<DungTheme>(t) );
        }
        
        for(uint32_t i = 0; i < 0x1F; ++i)
        {
            uint16_t primTheme = 0;
            
            PaletteData::tag_overworld &ot = pd.overworld;
            
            OverBgTheme t;
            
            // ------------------------------
            
            addr = 0x75504 + (i * 4);
            
            uint32_t temp1 = m_rom->GetByte(addr);    
            uint32_t temp2 = m_rom->GetByte(addr + 1);
            uint32_t temp3 = m_rom->GetByte(addr + 2);
            
            // excuse the messy loop, but it's required b/c the original game
            // doesn't associate the primary overworld bg palette theme with the
            // other components of the theme. Thus, this lookup algorithm was
            // implemented, allowing a finer degree of control over an area's
            // theme
            for(uint16_t j = 0; j <= 0x88; ++j)
            {
                uint16_t bgTheme = m_rom->GetByte(0x7D1C + j);
                
                if(bgTheme == i)
                {
                    switch(j)
                    {
                        case 0x03: case 0x05: case 0x07:
                        {   
                            primTheme = 0x02;
                            
                            break;
                        }
                        case 0x43: case 0x45: case 0x47:
                        {   
                            primTheme = 0x03;
                            
                            break;
                        }
                        default:
                        {
                            if(j < 0x40)
                            {
                                primTheme = 0x00;
                            }
                            else if(j < 0x80)
                            {
                                primTheme = 0x01;
                            }
                        }
                    }
                    
                    break;
                }
            }
            
            t.primary    = ot.primary[primTheme].Issue();
            
            if(temp1 < 0x80)
            {
                t.secondary1 = ot.secondary[temp1].Issue();
            }
            
            if(temp2 < 0x80)
            {
                t.secondary2 = ot.secondary[temp2].Issue();
            }
            
            if(temp3 < 0x80)
            {
                t.tertiary   = ot.tertiary [temp3].Issue();
            }
            
            /// (the fourth byte of each entry is not used, but we could use it later)
            /// t.unused = .....
            
            pt.overworldBg.push_back( Show<OverBgTheme>(t) );
        }
        
        for(uint32_t i = 0; i < 0x14; ++i)
        {
            OverSprTheme t;
            
            // -------------------------------
            
            addr = 0x75580 + (i * 2);
            
            uint32_t temp1 = m_rom->GetByte(addr + 0);
            uint32_t temp2 = m_rom->GetByte(addr + 1);
            
            if(temp1 < 0x80)
            {
                t.secondary1 = pd.sprite.secondary[temp1].Issue();
            }
            
            if(temp2 < 0x80)
            {
                t.secondary2 = pd.sprite.secondary[temp2].Issue();
            }
            
            pt.overworldSpr.push_back( Show<OverSprTheme>(t) );
        }
    }

// ===============================================================
    
    void RomLoader::LoadDungeonHeader(uint16_t roomNum,
                                      uint32_t headerBase,
                                      uint32_t headerBank,
                                      Dungeon::Header &header)
    {
        uint32_t distance        = 0;
        uint32_t headerLength    = 0x0E;
        uint32_t i               = 0;
        uint32_t otherRoomOffset = 0;
        uint32_t roomOffset      = 0;
        
        Buffer *headerData       = new Buffer(14);
        
        // ------------------------------------
        
        roomOffset = m_rom->Get2Bytes(headerBase + (roomNum * 2));
        
        // determine the lengths of the original game's headers
        // (they are often not the full 14 bytes that a header can have)
        for(i = 0; i < 319; ++i)
        {
            otherRoomOffset = m_rom->Get2Bytes(headerBase + (i * 2));
            
            distance        = otherRoomOffset - roomOffset;
            
            if( (distance != 0) && (distance < 0x0E) && (distance < headerLength) )
            {
                headerLength = distance;
            }
        }
        
        roomOffset = CpuToRomAddr(roomOffset | headerBank);
        
        Buffer::Copy(headerData, m_rom, 0, roomOffset, headerLength);
        
        header.Initialize(headerData);
        
        delete headerData;
        
        return;
    }

// ===============================================================
    
    void RomLoader::LoadDungeonEntrances()
    {
        uint32_t i;
        uint32_t i2;
        uint32_t j;
        
        using Dungeon::Entrance;
        
        // ----------------------------------
        
        // default loading of entrance data
        for(i = 0; i < 85; ++i)
        {
            i2 = i << 1;
            
            Entrance e(false);
            
            e.room = m_zg.dungRooms[ m_rom->GetByte(0x14813 + i2) ].Issue();
            
            for(j = 0; j < 8; ++j)
            {
                e.relativeCoords[j] = m_rom->GetByte(0x14913 + (i * 8) + j);
            }
            
            e.scroll.Set( m_rom->Get2Bytes(0x14D45 + i2),
                          m_rom->Get2Bytes(0x14E4F + i2) );
            
            e.player.Set( m_rom->Get2Bytes(0x15063 + i2),
                          m_rom->Get2Bytes(0x14F59 + i2) );
            
            e.camera.Set( m_rom->Get2Bytes(0x15277 + i2),
                          m_rom->Get2Bytes(0x1516D + i2) );
            
            e.mainGraphics    = m_rom->GetByte(0x15381 + i);
            e.startingFloor   = m_rom->GetByte(0x15406 + i);
            e.palace          = m_rom->GetByte(0x1548B + i);
            e.doorOrientation = m_rom->GetByte(0x15510 + i);
            e.startingBg      = m_rom->GetByte(0x15595 + i);
            
            e.quadrant1       = m_rom->GetByte(0x1561A + i);
            e.quadrant2       = m_rom->GetByte(0x1569F + i);
            
            e.doorSettings    = m_rom->Get2Bytes(0x15724 + i2);
            
            e.song            = m_rom->GetByte(0x1582E + i);
            
            m_zg.entrances.push_back( Show<Entrance>(e) );
        }
        
        for(i = 0; i < 7; ++i)
        {
            i2 = i << 1;
            
            Entrance e(true);
            
            e.room = m_zg.dungRooms[ m_rom->GetByte(0x15B6E + i2) ].Issue();
            
            for(j = 0; j < 8; ++j)
            {
                e.relativeCoords[j] = m_rom->GetByte(0x15B7C + (i * 8) + j);
            }
            
            e.scroll.Set(m_rom->Get2Bytes(0x15BB4 + i2), m_rom->Get2Bytes(0x15BC2 + i2) );
            e.player.Set(m_rom->Get2Bytes(0x15BDE + i2), m_rom->Get2Bytes(0x15BD0 + i2) );
            e.camera.Set(m_rom->Get2Bytes(0x15BFA + i2), m_rom->Get2Bytes(0x15BEC + i2) );
            
            e.mainGraphics    = m_rom->GetByte(0x15C08 + i);
            e.startingFloor   = m_rom->GetByte(0x15C0F + i);
            e.palace          = m_rom->GetByte(0x15C16 + i);
            e.startingBg      = m_rom->GetByte(0x15C1D + i);
            
            e.quadrant1       = m_rom->GetByte(0x15C24 + i);
            e.quadrant2       = m_rom->GetByte(0x15C2B + i);
            
            e.doorSettings       = m_rom->Get2Bytes(0x15C32 + i2);
            e.associatedEntrance = m_rom->Get2Bytes(0x15C40 + i2);
            
            e.song            = m_rom->GetByte(0x15C4E + i);
            
            m_zg.startingPoints.push_back( Show<Entrance>(e) );
        }
    }

// ===============================================================
    
    void RomLoader::LoadDungeonRooms()
    {
        using Containers::Show;
        
        using Dungeon::Room;
        
        uint16_t i          = 0;
        
        uint32_t headerBase = 0;
        uint32_t headerBank = 0;
        uint32_t objectBase = 0;
        uint32_t layoutBase = 0;
        uint32_t itemBase   = 0;
        
        // This is the default entrance data location
        // The plan is to later maybe expand entrance data but for now there's
        // enough.
        uint32_t entranceBase = 0;
        
        /// Also remember to change not only $E9FB, but $EA0C, and $EC10
        /// Also, $EBF6 is the max number of chests * 3 from the game's perspective.
        /// uint32_t chestLocationBase = game->bm_Header.dngChestOffset;
        
        uint32_t currentPtr     = 0;
        uint32_t spritePtrBank  = 0;
        uint32_t spritePtrTbl   = GetBank(spritePtrBank) << 0x10;
        
        uint32_t room           = 0;
        uint32_t roomOffset     = 0;
        uint32_t size           = 0;
        uint32_t objectSize     = 0;
        uint32_t index          = 0;
        uint32_t index2         = 0;
        uint32_t sum            = 0;
        
        ExpandInfo &e = m_ei;
        
        // -----------------------------
        
        headerBase = CpuToRomAddr(e.m_dungeon.headerOffset);
        headerBank = e.m_dungeon.headerBank << 16;
        
        objectBase = CpuToRomAddr(e.m_dungeon.objOffset);
        
        entranceBase = e.m_dungeon.entranceOffset;
        
        /// hardcoded for now
        itemBase = 0xdb69;
        
        for(i = 0; i < 320 ; ++i)
        {
            Show<Room> r;
            
            Room &r_re = r.GetRef();
            
            // -------------------------------------------
            
            LoadDungeonHeader (i, headerBase, headerBank, r_re.header);
            LoadDungeonObjects(i, objectBase, r_re);
            LoadDungeonSprites(i, r_re.sprites);
            LoadDungeonItems  (i, itemBase, r_re.items);
            LoadDungeonBlocks (i, r_re.blocks);
            LoadDungeonTorches(i, r_re.torches);
            
            m_zg.dungRooms.push_back(r);
        }
    }

// ===============================================================

    void RomLoader::LoadLayouts()
    {
        using namespace Dungeon;
        
        uint32_t layoutBase = CpuToRomAddr(m_ei.m_dungeon.layoutOffset);
        
        // -----------------------------
        
        for(int i = 0; i < 8; ++i)
        {
            uint32_t layoutPtr = m_rom->Get3Bytes(layoutBase + (i * 3) );
            
            // provides a cpu address for this particular room's object data
            uint32_t layoutOffset = CpuToRomAddr(layoutPtr);
            
            uint32_t layoutStart = layoutOffset;
            
            uint32_t objectSize = 0;
            uint32_t size       = 0;
            
            ObjectList layout;
            
            // -----------------------------
            
            // variable to help us remember where the data started at.
            layoutStart = layoutOffset;
            
            // next determine the size (length) of this object data before we copy 
            // the data to a buffer.
            for(size = 0; ; size += 3)
            {
                // minor precaution to avoid crashes
                if(size >= 0x8000)
                {
                    throw exception();
                    
                    return;
                }
                
                // 0xFFFF is the terminator for each layout's object data.
                if( m_rom->Get2Bytes(layoutOffset + size) == 0xFFFF )
                {
                    size += 2;
                    
                    // And end the loop.
                    break;
                }
                
                if( m_rom->Get2Bytes(layoutOffset + size) == 0xFFF0 )
                {
                    // We shouldn't find any doors in dungeon object lists!
                    throw exception();
                    
                    return;
                }
                else
                {
                    uint32_t data = m_rom->Get3Bytes(layoutOffset + size);
                    
                    Object::objType t;
                    
                    Object obj;
                    
                    // ---------------------------------
                    
                    if( (data & 0xfc) == 0xfc)
                    {
                        t = Object::type1_2;
                    }
                    else if( (data >> 16) >= 0xf8)
                    {
                        t = Object::type1_3;
                    }
                    else
                    {
                        t = Object::type1_1;
                    }
                    
                    obj = Object(t, data);
                    
                    layout.push_back(obj);
                }
            }
            
            m_zg.m_layouts.push_back( Show<ObjectList>(layout) );
        }
        
        return;
    }

// =============================================================================

    string RomLoader::GetFloorBlock(int i)
    {
        stringstream ss;
        
        // -----------------------------
        
        ss << "obj_floor";
        
        if( (i < 0) || (i > 15) )
        {
            i = 0;
        }
        
        ss << i;
        
        return ss.str();
    }

// =============================================================================

    void RomLoader::LoadDungeonObjects(uint16_t       roomNum,
                                       uint32_t       objectBase,
                                       Dungeon::Room &room)
    {
        using namespace Dungeon;
        
        using Thorn::Block;
        using Thorn::BlockRef;
        
        using Containers::Ticket;
        
        uint32_t roomOffset = 0;
        uint32_t roomStart  = 0;
        uint32_t objectSize = 0;
        uint32_t size       = 0;
        uint32_t layer      = 0;
        
        // ---------------------------------
        
        // provides a cpu address for this particular room's object data
        roomOffset = CpuToRomAddr( m_rom->Get3Bytes(objectBase + (roomNum * 3)) );
        
        // variable to help us remember where the data started at.
        roomStart  = roomOffset;
        
        uint8_t floorVal = m_rom->GetByte(roomOffset);
        
        string f1_s = GetFloorBlock( ( floorVal & 0xf0 ) >> 4 );
        string f2_s = GetFloorBlock( ( floorVal & 0x0f ) << 0 );    
        
        Ticket<Block> f1_ticket = m_zg.m_thornBlocks["obj_floor1"].Issue();
        Ticket<Block> f2_ticket = m_zg.m_thornBlocks["obj_floor1"].Issue();
        
        room.floor1 = BlockRef(f1_ticket);
        room.floor2 = BlockRef(f2_ticket);
        
        uint32_t layout_num = ( m_rom->GetByte(roomOffset + 1) >> 2 );
        
        Ticket<ObjectList> layout_ticket = m_zg.m_layouts[layout_num].Issue();
        
        room.m_layout = layout_ticket;
        
        // start off with the size of the object data being 2 bytes long
        // This accounts for the two properties entries
        for(size = 2, layer = 0; layer < 3; ++layer)
        {
            // next determine the size (length) of this object data before we 
            // copy the data to a buffer.
            for(objectSize = 3;  ; size += objectSize)
            {
                // minor precaution to avoid crashes
                if(size >= 0x8000)
                {
                    throw exception();
                    
                    return;
                }
                
                // 0xFFFF is the terminator for each room's object data.
                if( m_rom->Get2Bytes(roomOffset + size) == 0xFFFF )
                {
                    // since roomOffset double functions as size, we increment
                    //  by 2.
                    size += 2;
                    
                    // and end the loop
                    break;
                }
                
                // check to see if we are going to switch to door objects
                // which are two byte objects rather than three byte
                // not all rooms have them >_>
                if( m_rom->Get2Bytes(roomOffset + size) == 0xFFF0 )
                {
                    // switch to type 2 objects, which are only 2 bytes each
                    if(objectSize == 3)
                    {
                        objectSize = 2;
                        
                        continue;
                    }
                }
                else
                {
                    uint32_t data = m_rom->Get3Bytes(roomOffset + size);
                    
                    Object::objType t;
                    
                    Object obj;
                    
                    // ---------------------------------
                    
                    if(objectSize == 3)
                    {
                        Floor f = (layer == 1) ? BG1 : BG2;
                        
                        if( (data & 0xFC) == 0xFC)
                        {
                            t = Object::type1_2;
                        }
                        else if( (data >> 16) >= 0xF8)
                        {
                            t = Object::type1_3;
                        }
                        else
                        {
                            t = Object::type1_1;
                        }
                        
                        obj = Object(t, data);
                        
                        if
                        (
                            !FilterParameterizedObjects(roomNum,
                                                        f,
                                                        obj,
                                                        room)
                        )
                        {
                            room.objects[layer].type1.push_back(obj);
                        }
                    }
                    else
                    {
                        Door d(m_rom->Get2Bytes(roomOffset + size));
                        
                        room.objects[layer].doors.push_back(d);
                    }
                }
            }
        }
        
        return;
    }

// ===============================================================

    bool RomLoader::FilterParameterizedObjects(uint16_t         roomNum,
                                               Dungeon::Floor   floor,
                                               Dungeon::Object &obj,
                                               Dungeon::Room   &room)
    {
        using namespace Dungeon;
        
        bool parameterized = true;
        
        uint32_t hits = 0;
        
        // ------------------------------------
        
        if(obj.m_type == Object::type1_1)
        {
            switch(obj.m_routine)
            {
                // Note: 0x00 has been added to keep the compiler from bitching
                // about a switch statement with only a default case
                case 0x00:
                default:
                {
                    parameterized = false;
                    
                    break;
                }
            }
        }
        else if(obj.m_type == Object::type1_2)
        {
            switch(obj.m_routine)
            {
                case 0x2d: // in-floor inter-room up-north staircase
                case 0x38: // in-wall inter-room up-north spiral staircase (BG2 only?)
                case 0x3a: // in-wall inter-room up-north spiral staircase (BG1 only?)
                case 0x2e: // in-floor inter-room down-south staircase
                case 0x2f: // in-floor inter-room down-south staircase (use with hidden wall)
                case 0x39: // in-wall inter-room down-north spiral staircase (BG1 only?)
                case 0x3b: // in-wall inter-room down-north spiral staircase (BG2 only?)
                {
                    FilterStaircase(roomNum, floor, obj, room);
                    
                    break;
                }
                default:
                {
                    parameterized = false;
                    
                    break;
                }
            }
        }
        else if(obj.m_type == Object::type1_3)
        {
            switch(obj.m_routine)
            {
                case 0x13: // Signs (telepathic tiles).
                {
                    FilterSign(roomNum, floor, obj, room);
                    
                    break;
                }
                case 0x19: // Chests
                case 0x31: // Big Chests
                {
                    uint16_t chestsFound = 0;
                    uint16_t chestItem   = 0; 
                    uint16_t chestRoom   = 0;
                    
                    // Retrieve the rom address of this data.
                    uint32_t chestBase  = 0;
                    uint32_t i          = 0;
                    
                    Chest ch;
                    
                    // -----------------------------
                    
                    chestBase = CpuToRomAddr(m_ei.m_dungeon.chestOffset);
                    
                    for(i = 0; i < m_ei.m_dungeon.numChests; ++i)
                    {
                        chestRoom = m_rom->Get2Bytes(chestBase + (i * 3) ) & 0x7FFF;
                        chestItem = m_rom->GetByte(chestBase + (i * 3) + 2);
                        
                        if(chestRoom == roomNum)
                        {
                            if(chestsFound < hits)
                            {
                                chestsFound += 1;
                                
                                chestItem = 0xFFFF;
                                
                                continue;
                            }
                            else
                            {
                                hits += 1;
                                
                                break;                                            
                            }
                        }                                        
                        else
                        {
                            chestItem = 0xFFFF;                                        
                        }
                    }
                    
                    if(chestItem == 0xFFFF)
                    {
                        throw exception();
                    }
                    
                    ch = Chest(obj.m_x, obj.m_y, floor, chestItem);
                    
                    if(obj.m_routine == 0x31)
                    {
                        room.bigChests.push_back(ch);
                    }
                    else if(obj.m_routine == 0x19)
                    {
                        room.chests.push_back(ch);
                    }
                    
                    break;
                }
                case 0x18: // big key locks
                {
                    Lock lock;
                    
                    // -----------------------------
                    
                    lock = Lock(obj.m_x, obj.m_y, floor);
                    
                    room.locks.push_back(lock);
                    
                    break;
                }
                case 0x1e: // in-wall inter-room up-north straight staircase (BG2 only)
                case 0x26: // in-wall inter-room up-north straight staircase (BG1 only)
                case 0x20: // in-wall inter-room up-south straight staircase (BG2 only)
                case 0x28: // in-wall inter-room up-south straight staircase (BG1 only)
                case 0x1f: // in-wall inter-room down-north straight staircase (BG2 only)
                case 0x27: // in-wall inter-room down-north straight staircase (BG1 only)
                case 0x21: // in-wall inter-room down-south straight staircase (BG2 only)
                case 0x29: // in-wall inter-room down-south straight staircase (BG1 only)
                {
                    FilterStaircase(roomNum, floor, obj, room);
                    
                    break;
                }
                case 0x4a:
                {
                    parameterized = false;
                    // FilterWarp(roomNum, room);
                }
                default:
                {
                    parameterized = false;
                    
                    break;
                }
            }
        }
        
        
        return parameterized;
    }

// =============================================================================

    void RomLoader::FilterStaircase(uint16_t         roomNum,
                                    Dungeon::Floor   floor,
                                    Dungeon::Object &obj,
                                    Dungeon::Room   &room)
    {
        using namespace Dungeon;
        
        // limit the index into the header's target rooms for staircases is always
        // between 0 and 3 inclusive.
        int index = (int) (room.stairs.size() % 4);
        
        uint16_t targetRoom = room.header.GetStaircaseRoom(index);
        
        // ----------------------------------
        
        // This is to deal with the fact that the original game only reserved
        // a byte in the header for each target room. Thus, you couldn't jump
        // from, say, room 0x0004 to 0x0145 because their high byte wasn't
        // the same.  You could only target a room such as 0x0045 from 0x0004.
        if(roomNum >= 0x100)
        {
            targetRoom |= 0x100;
        }
        
        // todo: index staircase rooms somehow
        room.stairs.push_back(Staircase(obj.m_x,
                                        obj.m_y,
                                        floor,
                                        targetRoom,
                                        0));
        
    }

// =============================================================================

    void RomLoader::FilterSign(uint16_t         p_roomNum,
                               Dungeon::Floor   p_floor,
                               Dungeon::Object &p_obj,
                               Dungeon::Room   &p_room)
    {
        using Text::Message;
        
        using namespace Thorn;
        using namespace Dungeon;
        using namespace Containers;
        
        uint16_t msgIndex = 0;
        
        // Base of the array of text messages for each room.
        const uint32_t base = 0x3f61d;
        
        // -----------------------------
        
        // Get the numerical index of the message the sign uses.
        msgIndex = m_rom->Get2Bytes( base + (p_roomNum * 2) );
        
        if(1)
        {
            // Now the messy business of generating all of the components that
            // go into the sign...
            
            Position<8> pos(p_obj.m_x, p_obj.m_y, p_floor);
            
            Ticket<Message> t = m_zg.m_messages[msgIndex].Issue();
            
            BlockRef br( m_zg.m_thornBlocks["obj_telepath_tile"].Issue() );
            
            p_room.m_signs.push_back( Sign(pos, t, br) );
        }
        
        return;
    }
    
// =============================================================================
    
    void RomLoader::LoadDungeonSprites(uint16_t roomNum,
                                       Dungeon::SpriteList &list)
    {
        using namespace Dungeon;
        
        uint32_t roomOffset = 0;
        uint32_t size       = 0;
        
        uint32_t spriteBase = 0;
        uint32_t spriteBank = 0;
        
        // --------------------------------
        
        spriteBase = m_ei.m_dungeon.spriteOffset;
        spriteBank = m_ei.m_dungeon.spriteBanks;        
        
        if(spriteBank == 0x090000)
        {
            spriteBase |= spriteBank;
            spriteBase  = CpuToRomAddr(spriteBase);
        }
        else
        {
            /*
            spriteBase = RomToCpuAddr(spritePtrBase);
            
            // This will provide the bank the data is stored in.
            // Do not confuse this with spritePtrTbl, which says which bank
            // the sprite pointer table is in. The latter never changes
            spriteBank   = GetByte(game->image, game->bm_Header.dngSpriteBanks + index);
            spriteBank <<= 0x10;
            */         
        }
        
        /// 1/22/2010 - taken out temporarily b/c this was probably just a
        /// line of code that was modified so that shards of might would open
        /// currentPtr = m_rom->Get2Bytes(asm_sprite_local + (index * 2) );
        
        /// replaced with the following line:
        // Get the rom address of the current room's sprites
        roomOffset  = m_rom->Get2Bytes(spriteBase + (roomNum * 2));
        roomOffset |= spriteBank;
        roomOffset  = CpuToRomAddr(roomOffset);
        
        // first byte determines whether sprites are sorted or not.
        list.sortSprites = m_rom->GetByte(roomOffset) ? true : false;
        
        // The size check is kind of arbirtary. Just to make sure
        // our data doesn't go crazy.
        for(size = 1 ; size < 0xFFFF ; size += 3)
        {
            uint8_t x;
            uint8_t y;
            uint8_t type;
            uint8_t subType;
            
            uint8_t raw0 = m_rom->GetByte(roomOffset + size + 0);
            uint8_t raw1 = m_rom->GetByte(roomOffset + size + 1);
            uint8_t raw2 = m_rom->GetByte(roomOffset + size + 2);
            
            Sprite::DropType drop = Sprite::none;
                    
            Floor floor;
            
            Sprite spr;
            Overlord ovl;
            
            // ------------------------------------------------
            
            // Check for the end of the sprite list for the room....
            if( m_rom->GetByte(roomOffset + size) == 0xFF )
            {
                break;
            }
            
            // determine common parameters to overlords and sprites
            x     = (raw1 & 0x1F) << 4;
            y     = (raw0 & 0x1F) << 4;
            
            floor = (raw0 & 0x80) ? BG1
                                  : BG2;
            
            type  = raw2;
            
            // if the top 3 bits are not set it's a normal sprite
            if(raw1 < 0xE0)
            {
                if(type > 0xF2)
                {
                    throw exception();
                }
                
                if(type == 0xE4)
                {
                    // Key sprites can be used to modify the previous sprite
                    // in order to make the previous sprite drop a key after it dies
                    // big keys can be made to drop as well.
                    
                    if(raw0 == 0xFE)
                    {
                        drop = Sprite::small_key;
                    }
                    else if(raw0 == 0xFD)
                    {
                        drop = Sprite::big_key;
                    }
                    
                    if(drop != Sprite::none)
                    {
                        if(list.sprites.size() > 0)
                        {
                            // this sprite only serves as a modifier to a
                            // previous sprite. Move on to the next sprite.
                            list.sprites.back().SetDrop(drop);
                            
                            continue;
                        }
                        else
                        {
                            // Can't modify the previous sprite when there
                            // isn't one!
                            throw exception();
                        }
                    }
                }
                
                // 0bb00000 -> 000bb000
                subType  = (raw0 & 0x60) >> 2;
                
                // aaa00000 -> 00000aaa
                subType |= (raw1 & 0xE0) >> 5;
                
                if(subType >= 0x01)
                {
                    subType = subType;
                }
                
                spr = Sprite(floor, x, y, raw2, subType);
                
                list.sprites.push_back(spr);
            }
            else
            {
                // Instead, the sprite is an overlord
                if(type > 0x1A)
                {
                    throw exception();
                }
                
                floor = (raw0 & 0x80) ? BG1
                                      : BG2;
                
                ovl = Overlord(floor, x, y, type);
                
                list.overlords.push_back(ovl);
            }
        }
        
        return;
    }

// ===============================================================

    void RomLoader::LoadDungeonItems(uint16_t roomNum,
                                     uint32_t itemBase,
                                     Dungeon::ItemList &list)
    {
        using namespace Dungeon;
        
        uint32_t i          = 0;
        uint32_t roomOffset = 0;
        
        // -----------------------------------------------
        
        // grab 2-byte pointer for room and OR in bank 0x01
        roomOffset  = m_rom->Get2Bytes(itemBase + (roomNum << 1) );
        roomOffset |= 0x010000;
        
        // convert to rom offset
        roomOffset = CpuToRomAddr(roomOffset);
        
        for(i = 0;  ; i += 3)
        {
            uint16_t addr  = 0;
            uint16_t x     = 0;
            uint16_t y     = 0;
            
            Floor floor;
            
            Item::ItemType type = Item::none;
            
            Item *item;
            
            Position<8> pos;
            
            // ------------------------
            
            addr = m_rom->Get2Bytes(roomOffset + i + 0);
            /// type = m_rom->GetByte  (roomOffset + i + 2);
            
            if(addr == 0xFFFF)
            {
                break;
            }
            else if(type == 0x00)
            {
                break;
            }
            
            x = Tilemap::ToX(addr);
            y = Tilemap::ToY(addr);
            
            floor = (addr & 0x2000) ? BG1
                                    : BG2;
            
            item = new Item(x, y, floor);
        }
        
        return;
    }

// ===============================================================

    void RomLoader::LoadDungeonBlocks(uint16_t roomNum,
                                      Dungeon::BlockList &list)
    {
        using namespace Dungeon;
        
        uint16_t room  = 0;
        uint16_t value = 0;
        
        uint32_t i     = 0;
        
        // ----------------------------
        
        for(i = 0; i < 0x18C; i += 4)
        {
            room  = m_rom->Get2Bytes(0x271DE + i + 0);
            value = m_rom->Get2Bytes(0x271DE + i + 2);
            
            if(room == roomNum)
            {
                bool triggersEvent = ( (value & 0x4000) != 0 ) ? true : false;
                
                uint16_t x = Tilemap::ToX(value);
                uint16_t y = Tilemap::ToY(value);
                
                Floor floor = ( (value & 0x2000) != 0 ) ? BG1 : BG2;
                
                Block *block;
                
                // -----------------------------------------------
                
                block = new Block(triggersEvent, x, y, floor);
                
                list.push_back(block);
            }
        }
    }

// ===============================================================

    void RomLoader::LoadDungeonTorches(uint16_t roomNum,
                                       Dungeon::TorchList &list)
    {
        using namespace Dungeon;
        
        uint16_t room  = 0;
        uint16_t value = 0;
        
        uint32_t i     = 0;
        
        // -----------------------------------------------------
        
        for(i = 0; i < 0x120; i += 2)
        {
            room = m_rom->Get2Bytes(0x2736A + i);
            
            for(i += 2; i < 0x120; i += 2)
            {
                value = m_rom->Get2Bytes(0x2736A + i);
                
                if(value == 0xFFFF)
                {
                    break;
                }
                
                if(room == roomNum)
                {
                    bool lit = (value & 0x8000) ? true : false;
                    
                    uint16_t x = Tilemap::ToX(value);
                    uint16_t y = Tilemap::ToY(value);
                    
                    Floor floor = (value & 0x2000) ? BG1
                                                   : BG2;
                    
                    Torch *t = new Torch(lit, x, y, floor);
                    
                    // -----------------------------------------------
                    
                    list.push_back(t);
                }
            }
        }
        
        return;
    }

// ===============================================================
    
    uint16_t RomLoader::ResolveArea(uint16_t area)
    {
        uint16_t resolvedArea = 0;
        
        // -----------------------
        
        // large areas group together 4 small areas
        if(resolvedArea < 0x80)
        {
            // Maps 512x512 area number to the larger 1024x1024 area number that it belongs to.
            resolvedArea = ( m_rom->GetByte(0x125EC + (area & 0x3F)) + (area & 0x40) );
        }
        else
        {
            resolvedArea = area;
        }
        
        return resolvedArea;
    }

// ===============================================================
    
    void RomLoader::LoadOverworldAreas()
    {
        LoadMap32To16();
        LoadMap16To8();
        LoadOverworldMap();
        LoadOverworldGfxSettings();
        
        for(uint16_t i = 0; i < m_zg.areas.size(); ++i)
        {
            Overworld::Area &area = m_zg.areas[i];
            
            // ----------------------------------
            
            LoadOverworldHoles(i, area.holes);
            LoadOverworldEntrances(i, area.entr);
            LoadOverworldExits(i, area.exits);
        }
    }

// ===============================================================
    
    void RomLoader::LoadOverworldGfxSettings()
    {
        for(uint16_t i = 0; i < m_zg.areas.size(); ++i)
        {
            uint16_t bgTheme;
            
            Overworld::Area &area = m_zg.areas[i];
            
            Graphics::TileThemes &tt = m_zg.tileThemes;
            
            Graphics::PaletteThemes &pt = m_zg.palThemes;
            
            // --------------------------------------
            
            // Tile themes
            if(i < 0x40)
            {
                area.bgTileThemeMain = tt.bgMain[0x20].Issue();
                
                area.bgTileThemeAux  = tt.bgAux[m_rom->GetByte(0x7C9C + i) ].Issue();
                
                auto &sm = tt.sprMain;
                
                area.sprTileThemeMain[0] = sm[m_rom->GetByte(0x7A41 + 0x00 + i)].Issue();
                area.sprTileThemeMain[1] = sm[m_rom->GetByte(0x7A41 + 0x40 + i)].Issue();
                area.sprTileThemeMain[2] = sm[m_rom->GetByte(0x7A41 + 0x80 + i)].Issue();
                
                area.sprTileThemeAux = tt.sprAux[0x00].Issue();
                
                auto &ows = pt.overworldSpr;
                
                area.sprPalTheme[0] = ows[m_rom->GetByte(0x7B41 + 0x00 + i)].Issue();
                area.sprPalTheme[1] = ows[m_rom->GetByte(0x7B41 + 0x40 + i)].Issue();
                area.sprPalTheme[2] = ows[m_rom->GetByte(0x7B41 + 0x80 + i)].Issue();
            }
            else if(i < 0x80)
            {
                uint32_t j = i - 0x40;
                
                // -----------------------------------
                
                area.bgTileThemeMain = tt.bgMain[0x21].Issue();
                
                area.bgTileThemeAux  = tt.bgAux[m_rom->GetByte(0x7C9C + i) ].Issue();
                
                area.sprTileThemeMain[0] = tt.sprMain[m_rom->GetByte(0x7A41 + 0xC0 + j)].Issue();
                area.sprTileThemeMain[1] = tt.sprMain[m_rom->GetByte(0x7A41 + 0xC0 + j)].Issue();
                area.sprTileThemeMain[2] = tt.sprMain[m_rom->GetByte(0x7A41 + 0xC0 + j)].Issue();
                
                area.sprTileThemeAux = tt.sprAux[0x01].Issue();
                
                area.sprPalTheme[0] = pt.overworldSpr[m_rom->GetByte(0x7B41 + 0xC0 + j)].Issue();
                area.sprPalTheme[1] = pt.overworldSpr[m_rom->GetByte(0x7B41 + 0xC0 + j)].Issue();
                area.sprPalTheme[2] = pt.overworldSpr[m_rom->GetByte(0x7B41 + 0xC0 + j)].Issue();
            }
            else
            {
                // hardcoded b/c higher up areas don't actually
                // have this data, they just use the area you came from.
                area.bgTileThemeMain = tt.bgMain[0x20].Issue();
                
                area.bgTileThemeAux  = tt.bgAux[0x21].Issue();
                
                area.sprTileThemeMain[0] = tt.sprMain[0x00].Issue();
                area.sprTileThemeMain[1] = tt.sprMain[0x00].Issue();
                area.sprTileThemeMain[2] = tt.sprMain[0x00].Issue();
                
                area.sprTileThemeAux = tt.sprAux[0x00].Issue();
                
                area.sprPalTheme[0] = pt.overworldSpr[0x00].Issue();
                area.sprPalTheme[1] = pt.overworldSpr[0x00].Issue();
                area.sprPalTheme[2] = pt.overworldSpr[0x00].Issue();
            }
            
            // Palette Themes
            if(i < 0x88)
            {
                bgTheme = m_rom->GetByte(0x7D1C + i);
            }
            else
            {
                bgTheme = 0;
            }
            
            area.bgPalTheme  = pt.overworldBg[bgTheme].Issue();
        }
    }

// ===============================================================
    
    void RomLoader::LoadOverworldMap()
    {
        for(uint16_t i = 0; i < 0xc0; ++i)
        {
            m_zg.areas.push_back( Overworld::Area() );
        }
        
        for(uint16_t i = 0; i < m_zg.areas.size(); ++i)
        {
            // Check whether the map is 1024x1024 (loads 4 areas) or 512x512 (loads 1 area)
            bool largeArea = false;
            
            bool darkWorld = ( (i & 0xffc0) == 0x0040) ? true : false;
            
            Overworld::Area &area = m_zg.areas[i];
            
            // ------------------------------------
            
            switch(i)
            {
                case 0x81:
                    
                    largeArea = true;
                    
                    break;
                
                case 0x80:
                case 0x88:
                case 0x93:
                case 0x94:
                    
                    largeArea = false;
                    
                    break;
                
                // presumably all overlays...?
                case 0x95:
                case 0x96:
                case 0x97: // looks very much the same as 0x9D
                case 0x9C:
                case 0x9D:
                case 0x9E:
                case 0x9F:
                {
                    largeArea = false;
                    
                    break;
                }
                
                default:
                {
                    if(i < 0x80)
                    {
                        if(m_rom->GetByte(0x12844 + (i & 0x3F) ) )
                        {
                            largeArea = true;
                        }
                        
                        if(ResolveArea(i) != i)
                        {
                            // area proper that wasn't used to its fullest in the original game.
                            area.areaMap = Overworld::Map16Buf(true);
                            
                            continue;
                        }
                    }
                    else
                    {
                        // area that wasn't used in the original game.
                        area.areaMap = Overworld::Map16Buf(true);
                        
                        continue;
                    }
                    
                    break;
                }
            }
            
            area.areaMap = Overworld::Map16Buf(largeArea);
            
            if(largeArea)
            {
                // Maps 512x512 area number to the larger 1024x1024 area number that it belongs to.
                uint16_t sourceArea   = m_rom->GetByte(0x125EC + (i & 0x3F)) + (darkWorld ? 0x40 : 0x00);
                
                Overworld::Area &resolvedArea = m_zg.areas[ ResolveArea(sourceArea) ];
                
                // -----------------------------------------------------------------
                
                // Loads all four of the 512x512 submaps that make up the large area.
                LoadMap16(resolvedArea, sourceArea + 0, upper_left);
                LoadMap16(resolvedArea, sourceArea + 1, upper_right);
                LoadMap16(resolvedArea, sourceArea + 8, lower_left);
                LoadMap16(resolvedArea, sourceArea + 9, lower_right);
            }
            else
            {
                LoadMap16(area, i, upper_left);
            }
        }
    }

// ===============================================================

    void RomLoader::LoadOverworldHoles(uint16_t areaNum, Overworld::EntranceList &list)
    {
        uint16_t x        = 0;
        uint16_t y        = 0;
        uint16_t area     = 0;
        uint16_t entrance = 0;
        
        uint32_t numHoles  = m_ei.m_overworld.numHoles;
        uint32_t mapOffset = 0, areaOffset = 0, entrOffset  = 0;
        uint32_t map16Val  = 0;
        
        // -------------------------------
        
        mapOffset   = CpuToRomAddr(m_ei.m_overworld.holeOffset);
        areaOffset  = mapOffset  + (numHoles * 2);
        entrOffset  = areaOffset + (numHoles * 2);
        
        for(uint32_t i = 0; i < numHoles; ++i)
        {
            // note the odd formula for the map16
            map16Val = m_rom->Get2Bytes(mapOffset + (i * 1) ) + 0x400;
            
            x        = Overworld::Map16Buf::GetX(map16Val);
            y        = Overworld::Map16Buf::GetY(map16Val);
            
            area     = m_rom->Get2Bytes(areaOffset + (i * 2) );
            entrance = m_rom->GetByte(entrOffset + i);
            
            if(area == areaNum)
            {
                list.push_back( Overworld::Entrance(x, y, entrance) );
            }
        }
    }

// ===============================================================

    void RomLoader::LoadOverworldEntrances(uint16_t areaNum,
                                           Overworld::EntranceList &list)
    {
        uint16_t area     = 0;
        uint16_t entrance = 0;
        uint16_t x        = 0;
        uint16_t y        = 0;
        
        uint32_t i         = 0;
        uint32_t numEntr   = 0;
        uint32_t mapOffset = 0, areaOffset = 0, entrOffset  = 0;
        uint32_t map16Val  = 0;
        
        // -------------------------------
        
        numEntr     = m_ei.m_overworld.numEntr;
        areaOffset  = CpuToRomAddr(m_ei.m_overworld.entrOffset);
        mapOffset   = areaOffset + (numEntr * 2);
        entrOffset  = mapOffset  + (numEntr * 2);
        
        for(i = 0; i < numEntr; ++i)
        {
            // note the normal formula for entrances versus holes. strange.
            map16Val = m_rom->Get2Bytes(mapOffset + (i * 2) );
            
            x        = Overworld::Map16Buf::GetX(map16Val);
            y        = Overworld::Map16Buf::GetY(map16Val);
            
            area     = m_rom->Get2Bytes(areaOffset + (i * 2) );
            entrance = m_rom->GetByte(entrOffset + i);
            
            if(area == areaNum)
            {
                list.push_back( Overworld::Entrance(x, y, entrance) );
            }
        }
    }

// ===============================================================
    
    void RomLoader::LoadOverworldExits(uint16_t areaNum,
                                       Overworld::ExitList &list)
    {
        using namespace Overworld;
        
        uint32_t numExits    = 0;
        uint32_t roomOffset  = 0;
        uint32_t areaOffset  = 0;
        uint32_t yOffset     = 0;
        uint32_t xOffset     = 0;
        uint32_t doorOffset  = 0;
        uint32_t doorOffset2 = 0;
        uint32_t map16Val    = 0;
        
        Exit::exit_type doorType = Exit::down;
        
        // -----------------------------------------
        
        numExits    = m_ei.m_overworld.numExits;
        roomOffset  = CpuToRomAddr(m_ei.m_overworld.exitOffset);
        areaOffset  = roomOffset + (numExits * 2);
        yOffset     = areaOffset + (numExits * 7);
        xOffset     = yOffset    + (numExits * 2);
        doorOffset  = xOffset    + (numExits * 8);
        doorOffset2 = doorOffset + (numExits * 2);
        
        for(uint32_t i = 0; i < numExits; ++i)
        {
            uint32_t room = m_rom->Get2Bytes(roomOffset + (i * 2) );
            uint32_t area = m_rom->GetByte(areaOffset + i);
            
            uint32_t xStart = (area & 0x07) << 9;
            uint32_t yStart = (area & 0x38) << 6;
            
            uint32_t x = m_rom->Get2Bytes(xOffset + (i * 2) ) - xStart;
            uint32_t y = m_rom->Get2Bytes(yOffset + (i * 2) ) - yStart;
            
            uint32_t doorData = m_rom->Get2Bytes(doorOffset + (i * 2) );
            
            // -------------------------------------
            
            switch(doorData)
            {
                case 0x0000:
                {
                    // replace the value of doorData with the next word in
                    // this case.
                    doorData = m_rom->Get2Bytes(doorOffset + (i * 2) );
                    
                    if(doorData & 0x8000)
                    {
                        doorType = Exit::palace;
                    }
                    else if(doorData != 0x0000)
                    {
                        doorType = Exit::sanctuary;
                    }
                    else
                    {
                        doorType = Exit::down;
                    }
                    
                    break;
                }
                case 0xFFFF:
                {
                    doorType = Exit::up;
                    
                    break;
                }
                default:
                {
                    doorType = (doorData & 0x8000) ? Exit::bombable
                                                   : Exit::wooden;
                    
                    break;
                }
            }
            
            
            
            // FIXME
            // refactored code, no construct for this currently
            // exitdoor->SetPos(GetMap16X(doorData), GetMap16Y(doorData));
            
            if(areaNum == area)
            {
                list.push_back( Exit(x, y, area, room, doorType) );
            }
        }
    }

// ===============================================================

    void RomLoader::LoadMap16To8()
    {
        using namespace Overworld;
        
        for(int i = 0; i < 0xEA8; ++i)
        {
            Map16Tile tile;
            
            for(int j = 0; j < 4; ++j)
            {
                uint32_t offset = CpuToRomAddr( 0xF8000 + (j * 2) + (i * 8) );
                
                tile[j] = m_rom->Get2Bytes(offset);
            }
            
            m_zg.map16To8.push_back(tile);
        }
    }

// ===============================================================

    void RomLoader::LoadMap32To16()
    {
        using namespace Overworld;
        
        uint32_t i        = 0;
        uint16_t value    = 0;
        uint16_t base     = 0;
        uint16_t index    = 0;
        
        // default offsets for the conversion arrays.
        uint32_t offsets[4] = { 0x18000, 0x1B400, 0x20000, 0x23400 };
        
        // corresponds to the upper left, upper right, lower left, lower right map16 values of the map32 block
        const uint8_t shift[4] = {0x04, 0x00, 0x0C, 0x08}; 
        
        // -------------------------------------------------
        
        for(i = 0, value = 0; value < 0x2AA7; i += 8, value++)
        {
            Map32Tile tile(0, 0, 0, 0);
            
            base = (value / 4) * 6;
            
            index = value & 0x03;
            
            tile[0] = ((m_rom->Get2Bytes(offsets[0] + 4 + base)) >> shift[index]);
            tile[0] = (tile[0] & 0x0F) << 8;
            tile[0] |= m_rom->GetByte(offsets[0] + index + base);
            
            tile[1] = ((m_rom->Get2Bytes(offsets[1] + 4 + base)) >> shift[index]);
            tile[1] = (tile[1] & 0x0F) << 8;
            tile[1] |= m_rom->GetByte(offsets[1] + index + base);
            
            tile[2] = ((m_rom->Get2Bytes(offsets[2] + 4 + base)) >> shift[index]);
            tile[2] = (tile[2] & 0x0F) << 8;
            tile[2] |= m_rom->GetByte(offsets[2] + index + base);
            
            tile[3] = ((m_rom->Get2Bytes(offsets[3] + 4 + base)) >> shift[index]);
            tile[3] = (tile[3] & 0x0F) << 8;
            tile[3] |= m_rom->GetByte(offsets[3] + index + base);
            
            map32To16.push_back(tile);
        }
        
        return;
    }

// ===============================================================

    //! \brief      Debug variable. We can remove it whenever.
    uint32_t totalMap32DataSize = 0;
    
    Buffer* RomLoader::LoadMap32(uint16_t mapNum)
    {
        uint8_t a = 0;
        uint8_t b = 0;
        
        uint32_t i         = 0;
        uint32_t tableLoc  = 0;
        uint32_t tableLoc2 = 0;
        
        uint32_t offset         = 0;
        uint32_t romOffset      = 0;
        uint32_t romOffsetStart = 0;
        
        Buffer *data1 = nullptr;
        Buffer *data2 = nullptr;
        
        // A map32 buffer should contain 0x200 bytes of data.
        Buffer *map32Buf = new Buffer(0x10, 0x10, 2);
        
        // ---------------------------------------------
        
        tableLoc  = CpuToRomAddr( m_rom->Get3Bytes(asm_overmap_ptrs) );
        tableLoc2 = CpuToRomAddr( m_rom->Get3Bytes(asm_overmap_ptrs2) );
        
        offset    = m_rom->Get3Bytes(tableLoc + (mapNum * 3));
        
        romOffsetStart = romOffset = CpuToRomAddr(offset);
        
        data1    = DecompressStandard(m_rom, romOffset, false);
        
        totalMap32DataSize += (romOffset - romOffsetStart);
        
        offset    = m_rom->Get3Bytes(tableLoc2 + (mapNum * 3));
        
        romOffsetStart = romOffset = CpuToRomAddr(offset);
        
        data2    = DecompressStandard(m_rom, romOffset, false );
        
        totalMap32DataSize += (romOffset - romOffsetStart);
        
        for(i = 0; i < 0x100; ++i)
        {
            // notice how this data is interlaced from two separate sources
            // strange storage method... maybe it compresses better that way
            a = data1->GetByte(i);
            b = data2->GetByte(i);
            
            map32Buf->PutByte( (i * 2) + 1, a);
            map32Buf->PutByte( (i * 2)    , b);
        }
        
        delete data1;
        delete data2;
        
        // signal that there's a problem if the map32 data is not of the correct size
        if(map32Buf->length != 0x200)
        {
            map32Buf->error = 1;
        }
        
        return map32Buf;
    }

// ===============================================================

    void RomLoader::LoadMap16(Overworld::Area &area, uint16_t subArea, map16Pos p)
    {
        uint32_t map32Val = 0;
        
        uint32_t i = 0;
        uint32_t j = 0;
        uint32_t k = 0;
        
        // first time using auto_ptr, hope it goes over well...
        auto_ptr<Buffer> map32Buf(LoadMap32(subArea & 0xFF));
        
        /// Buffer *map32Buf = LoadMap32(subArea & 0xFF);
        
        // -----------------------------------------------
        
        // pretty complex for loop eh?
        for
        (
            k = 0x10, i = 0, j = (uint32_t) p;
            i < 0x200;
            i += 2, j += 4, k--
        )
        {
            // this conditional statement when activated, positions j
            // so that it will start writing to the next "line" of the map16 buffer
            if(!k)
            {
                j += 0xC0;
                k  = 0x10;
            }
            
            map32Val = map32Buf->Get2Bytes(i);
            
            area.areaMap(j + 0x00) = map32To16[map32Val][0];
            area.areaMap(j + 0x02) = map32To16[map32Val][1];
            area.areaMap(j + 0x80) = map32To16[map32Val][2];
            area.areaMap(j + 0x82) = map32To16[map32Val][3];
        }
    }

// ===============================================================

    void RomLoader::LoadDialogue()
    {
        using namespace Text;
        
        bool endOfPointers = false;
        
        // -----------------------------------------
        
        for(uint32_t pos = 0x1c8000; !endOfPointers; )
        {
            uint32_t messageBase   = CpuToRomAddr(pos);
            uint32_t messageOffset = CpuToRomAddr(pos);
            
            Message m;
            
            // ----------------------------------------
            
            while(true)
            {
                uint8_t code = m_rom->GetByte(messageOffset);
                
                // -------------------------------------
                
                if
                (
                    ( (code > 0x80) && (code < 0x88) )
                 || ( (code > 0xe8) && (code < 0xff) )
                )
                {
                    // These are invalid character codes in any circumstance,
                    // at least, in the legacy loader. We might change the
                    // way text is loaded and saved in the modern loader(s).
                    throw Text::Problem();
                }
                else if(code < 0x67)
                {
                    // Normal characters. (not dictionary entries)
                    
                    Entity e(Character, code);
                    
                    m.push_back(e);
                    
                    messageOffset += 1;
                }
                else if( (code >= 0x88) && (code <= 0xe8) )
                {
                    // Dictionary characters that encode a commonly
                    // used string of multiple normal characters.
                    // 
                    // The actual condition in the rom is >= 0x80, but
                    // these conditions are safer in the event 
                    // of junk data
                    
                    uint16_t index = (code - 0x88) * 2;
                    uint16_t start = 0;
                    uint16_t upper = 0;
                    
                    // -------------------------------
                    
                    start = m_rom->Get2Bytes(0x74703 + index);
                    upper = m_rom->Get2Bytes(0x74705 + index);
                    
                    for( ; (start < upper); ++start)
                    {
                        uint8_t dict_char = 0;
                        
                        uint32_t char_offset = CpuToRomAddr(0x0e0000 + start);
                        
                        // --------------------------------
                        
                        dict_char = m_rom->GetByte(char_offset);
                        
                        Entity e(Character, dict_char);
                        
                        m.push_back(e);
                    }
                    
                    messageOffset += 1;
                }
                else if( (code >= 0x67) && (code <= 0x7f) )
                {
                    // commands (variable length - 1 or 2 bytes in all)
                    uint8_t codeLen = m_rom->GetByte(asm_code_lengths + code);
                    
                    // ----------------------------------
                    
                    Entity e((EntityType) code);
                    
                    if(codeLen > 1)
                    {
                        e = Entity((EntityType) code,
                                   m_rom->GetByte(messageOffset + 1));
                    }
                    
                    m.push_back(e);
                    
                    messageOffset += codeLen;
                }
                
                if(code == 0x80)
                {
                    if( !m.empty() )
                    {
                        throw Text::Problem();
                    }
                    
                    pos = 0x0edf40;
                    
                    break;
                }
                else if(code == 0x7f)
                {
                    // Note: even a message with just an [End] command is
                    // considered nonempty. This is done here to prevent
                    // the creation of messages for the 0x80 and 0xff bytes
                    // that don't have any real message content to offer.
                    m_zg.m_messages.push_back( Show<Message>(m) );
                    
                    pos = RomToCpuAddr(messageOffset);
                    
                    break;
                }
                else if(code == 0xff)
                {
                    if( !m.empty() )
                    {
                        throw Text::Problem();
                    }
                    
                    endOfPointers = true;
                    
                    break;
                }
            }
        }
    }

// =============================================================================

    Thorn::Block RomLoader::
    ThornBlockFromIndirect(Thorn::IndirectBlock p_iblock)
    const
    {
        using namespace Thorn;
        
        int offset = 0x1b52 + p_iblock.m_offset;
        
        Block b(p_iblock.m_rows,
                p_iblock.m_columns);
        
        // -----------------------------
        
        for(int i = 0; i < p_iblock.m_rows; ++i)
        {
            for(int j = 0; j < p_iblock.m_columns; ++j)
            {
                // (tme = tile map entry)
                // The indirect block contains indices relative from the
                // starting offset, which is
                uint16_t tme = m_rom->Get2Bytes( offset + p_iblock(i, j));
                
                b(i, j) = tme;
            }
        }
        
        return b;
    }

// =============================================================================

    void RomLoader::LoadThornBlocks()
    {
        using namespace Thorn;
        
        using Containers::Show;
        
        auto &dbMap = defaultBlocks;
        
        auto end = dbMap.end();
        
        // -----------------------------
        
        for(auto it = dbMap.begin(); it != end; ++it)
        {
            try
            {
                string key = it->first;
                
                IndirectBlock ib = it->second;
                
                Block b = ThornBlockFromIndirect(ib);
                
                // -----------------------------
                
                m_zg.m_thornBlocks[key] = Show<Block>(b);
            }
            catch(std::exception const & e)
            {
                printf("wtf maet");
            }
        }
    }

// =============================================================================
    
    uint32_t RomLoader::AdvancePointer(uint32_t *offset, uint32_t size)
    {
        uint8_t oldBank = 0;
        uint8_t newBank = 0;
        uint8_t maxBank = 0;
        
        uint32_t oldOffset = 0;
        
        // --------------------------------
        
        // invalid pointer
        if(!offset)
        {
            return 0;
        }
        
        oldOffset = (*offset);
        
        if(IsBankChange(oldOffset, size))
        {
            oldBank = GetBank(oldOffset);
            newBank = oldBank + 1;
            maxBank = GetBank(RomToCpuAddr(m_rom->length)) - 1;
            
            // No more room left in the rom!
            if(newBank > maxBank)
            {
                throw exception();
                
                return 0;
            }
            
            // Moves the address into the start of the next bank.
            oldOffset = (newBank << 0x10) | 0x8000;
        }
        
        (*offset) = oldOffset + size;
        
        if(!((*offset) & 0x8000))
        {
            (*offset) = (GetBank(*offset) << 0x10) | 0x8000;
        }
        
        return oldOffset;
    }

// ===============================================================

    bool RomLoader::SaveGraphics(uint32_t *offset)
    {
    #if 0
    
        char *gfxText   = NULL;
        
        uint32_t cpuOffset   = 0;
        uint32_t i           = 0;
        uint32_t size        = 0;
        
        PathName *p     = NULL;
        
        ExpandInfo *e   = this->expandInfo;
        
        bufPtr compBuf  = NULL;
        
        FILE *f         = NULL;
        
        // ---------------------
        
        gfxText = GetAsmTextResource(IDR_GRAPHICS_ASM);
        
        if(!gfxText)
        {
            return false;
        }
        
        (*offset) = RomToCpuAddr(*offset);
        
        p = new PathName( romName->GetFullPath() );
        
        p->SetName("graphics");
        p->SetExt("asm");
        
        f = fopen(p->GetFullPath(), "wt");
        
        compBuf = new Buffer(0x2000);
        
        if(f)
        {
            // leave enough room for the hooked in routine
            fprintf(f, "\norg $%06X", AdvancePointer(offset, 0x80));
            fprintf(f, "%s", gfxText);
            
            // compress all the sprite graphics packs and construct the
            // table of their pointers in the asm file.
            e->sprGfxOffset = AdvancePointer(offset, 0x300);
            
            fprintf(f, "\norg $%06X", e->sprGfxOffset);
            fprintf(f, "\n    SprGfxPtrTable:");
            
            for(i = 0; i < 0x100; ++i)
            {
                if(i < 0x0C)
                {
                    // These graphics don't get compressed
                    size = sprPacks[i]->length;
                    Buffer::CopyBuffer(compBuf, sprPacks[i], 0, 0, size);
                }
                else
                {
                    size = Compresch_LTTP::Compress(sprPacks[i]->contents, sprPacks[i]->length, compBuf->contents);
                }
                
                cpuOffset = AdvancePointer(offset, size);
                
                fprintf(f, "\n        dl $%06X", cpuOffset);
                
                Buffer::CopyBuffer(m_rom, compBuf, CpuToRomAddr(cpuOffset), 0, size);
            }
            
            // Same as before, but this time for background graphics
            e->bgGfxOffset = AdvancePointer(offset, 0x300);
            
            fprintf(f, "\norg $%06X", e->bgGfxOffset);
            fprintf(f, "\n    BgGfxPtrTable:");
            
            for(i = 0; i < 0x100; ++i)
            {
                size      = Compresch_LTTP::Compress(bgPacks[i]->contents, bgPacks[i]->length, compBuf->contents);
                cpuOffset = AdvancePointer(offset, size);
                
                fprintf(f, "\n        dl $%06X", cpuOffset);
                
                Buffer::CopyBuffer(m_rom, compBuf, CpuToRomAddr(cpuOffset), 0, size);
            }
            
            fclose(f);
        }
        
        // the locations of these graphics don't change (for now)
        Buffer::CopyBuffer(m_rom, fontGfx,  0x70000, 0, 0x1000);
        Buffer::CopyBuffer(m_rom, linkGfx,  0x80000, 0, 0x7000);
        Buffer::CopyBuffer(m_rom, mode7Gfx, 0xC4000, 0, 0x4000);
        
        (*offset) = CpuToRomAddr(*offset);
        
        // clean up ------------------------------------------
        
        delete p;
        
        delete compBuf;
        
    #endif
    
        return true;
    }

// =============================================================================
    
    /// may or may not get used in the end
    bool RomLoader::ExportGfx()
    {
    #if 0
        char path[0x200];
        
        uint32_t i = 0;
        
        // -----------------------------
        
        /// chdir( (const char*) romName->GetDirectory() );
        
        /// i = _mkdir("gfx");
        
        // dump all sprite graphics
        for(i = 0; i < 0x100; ++i)
        {                
            sprintf(path, "%sgfx\\spr_gfx_%x.bin", romName->GetDirectory(), i);
            
            this->sprPacks[i]->ToFile(path);
        }
        
        // dump all background graphics
        for(i = 0; i < 0x100; ++i)
        {
            sprintf(path, "%sgfx\\bg_gfx_%x.bin", romName->GetDirectory(), i);
            
            this->bgPacks[i]->ToFile(path); 
        }
        
        // Output the 2bpp font graphics for the menus / dialogue
        sprintf(path, "%sgfx\\font_gfx.bin", romName->GetDirectory());
        
        this->fontGfx->ToFile(path);
        
        // Output Link's 4bpp sprite graphics
        sprintf(path, "%sgfx\\link_gfx.bin", romName->GetDirectory());
        
        this->linkGfx->ToFile(path);
        
        // Output mode7 graphics (to-do listed)
        sprintf(path, "%sgfx\\mode7_gfx.bin", romName->GetDirectory());
        
        this->mode7Gfx->ToFile(path);
        
    #endif
        return true;
    }

// =============================================================================
    
    /// may or may not get used in the end
    bool RomLoader::ImportGfx()
    {
    #if 0
        char path[0x400];
        
        uint32_t i = 0;
        
        // ------------------------------
        
        // import all sprite packs
        for(i = 0; i < 0x100; ++i)
        {
            sprintf(path, "%sgfx\\spr_gfx_%x.bin", romName->GetDirectory(), i);
            
            this->sprPacks[i]->FromFile(path);
        }
        
        // import all background packs
        for(i = 0; i < 0x100; ++i)
        {
            sprintf(path, "%sgfx\\bg_gfx_%x.bin", romName->GetDirectory(), i);
            
            this->bgPacks[i]->FromFile(path);
        }
        
        this->fontGfx->FromFile("gfx\\font_gfx.bin");
        this->linkGfx->FromFile("gfx\\link_gfx.bin");
        this->mode7Gfx->FromFile("gfx\\mode7_gfx.bin");
        
    #endif
        return true;
    }

// =============================================================================

    int ModernLoader::GetBgGfxPtr(int index)
    {
        uint32_t base = CpuToRomAddr(m_ei.m_graphics.bgOffset);
        
        // -----------------------------------------------------
        
        // new school method
        return CpuToRomAddr( m_rom->Get3Bytes(base + (index * 3) ) );
    }

// =============================================================================

    int ModernLoader::GetSprGfxPtr(int index)
    {
        uint32_t base = CpuToRomAddr(m_ei.m_graphics.sprOffset);
        
        // -----------------------------------------------------
        
        // new school method
        return CpuToRomAddr( m_rom->Get3Bytes(base + (index * 3) ) );
    }

// =============================================================================

}
