    
    #include <deque>
    
    #include "expandinfo.h"
    #include "zeldagame.h"
    
#ifndef LOADERS_H

    #define LOADERS_H

namespace Loaders
{
    using std::deque;
    using std::exception;
    using std::stringstream;
    using std::auto_ptr;
    
    using namespace Graphics;

    class RomLoader
    {
        
    public:
        
        // Raw binary image of the ROM
        Buffer *m_rom;
        
        ExpandInfo &m_ei;
        
        ZeldaGame &m_zg;
        
    public:
        
        RomLoader(ZeldaGame &p_zg, ExpandInfo &ei)

        : m_zg(p_zg),
          m_rom(ei.m_rom),
          m_ei(ei)
        {
            
        }
        
        void LoadGame();
        
        virtual void LoadDialogue();
        
        virtual void LoadTileSets();
        virtual void LoadTileThemes();
        
        virtual void LoadPalettes();
        virtual void LoadPaletteThemes();
        
        virtual void LoadDungeonEntrances();
        virtual void LoadDungeonRooms();
        
        void LoadOverworldMap();
        
        virtual void LoadOverworldAreas();
        
        virtual void LoadOverworldGfxSettings();
        
        virtual void LoadMap16To8();
        
        virtual void LoadLayouts();
        
        void LoadDungeonHeader(uint16_t roomNum,
                               uint32_t headerBase,
                               uint32_t headerBank,
                               Dungeon::Header &header);
        
        void LoadDungeonObjects(uint16_t       roomNum,
                                uint32_t       objectBase,
                                Dungeon::Room &room);
        
        void LoadDungeonSprites(uint16_t roomNum, Dungeon::SpriteList &list);
        
        void LoadDungeonItems(uint16_t roomNum, uint32_t itemBase, Dungeon::ItemList &list);
        
        void LoadDungeonBlocks(uint16_t roomNum, Dungeon::BlockList &list);
        
        void LoadDungeonTorches(uint16_t roomNum, Dungeon::TorchList &list);
        
        void LoadOverworldHoles(uint16_t areaNum, Overworld::EntranceList &list);
        
        void LoadOverworldEntrances(uint16_t areaNum, Overworld::EntranceList &list);
        
        void LoadOverworldExits(uint16_t areaNum, Overworld::ExitList &list);
        
        virtual int GetSprGfxPtr(int index);
        virtual int GetBgGfxPtr(int index);
        
        bool SaveGraphics(uint32_t *offset);
        bool ExportGfx();
        bool ImportGfx();
        
        uint32_t AdvancePointer(uint32_t *offset, uint32_t size);
        
        template<PixelDepth depth>
        void LoadTile(Tile<depth> &tile, Buffer *b, uint32_t position);
        
        template<PixelDepth depth, size_t numTiles>
        void LoadTileSet(TileSet<depth, numTiles> &ts, Buffer *b, uint32_t base = 0);
        
        template<PixelDepth depth, size_t numTiles>
        void LoadTileSet(deque< Show< TileSet<depth, numTiles> > > &d,
                         Buffer *b,
                         bool compressed = false,
                         uint32_t base = 0);
        
        template<size_t N, size_t M>
        void LoadGroup(uint32_t numGroups,
                       uint32_t addr,
                       typename Graphics::Group<N, M>::ShowDeque &d);
        
    // these members and member functions will not be inherited in the modern loader
    private:
        
        // _____________________________________________________________________
        // Protected member types and variables
        
        // map16 offsets for use in interlacing the map32 maps into the map16 structure
        typedef enum
        {
            upper_left  = 0x0000,
            upper_right = 0x0040,
            lower_left  = 0x1000,
            lower_right = 0x1040
            
        } map16Pos;
        
        // buffer that tells us how to convert map32 tiles to map16 tiles (for first time load only)
        deque<Overworld::Map32Tile> map32To16;
        
    protected:
        
        // _____________________________________________________________________
        // Protected member functions
        
        bool FilterParameterizedObjects(uint16_t         roomNum,
                                        Dungeon::Floor   floor,
                                        Dungeon::Object &obj,
                                        Dungeon::Room   &room);
        
        void FilterStaircase(uint16_t         roomNum,
                             Dungeon::Floor   floor,
                             Dungeon::Object &obj,
                             Dungeon::Room   &room);
        
        void FilterSign(uint16_t         b_roomNum,
                        Dungeon::Floor   b_floor,
                        Dungeon::Object &b_obj,
                        Dungeon::Room   &b_room);                        
        
        void LoadMap32To16();
        
        uint16_t ResolveArea(uint16_t area);
        
        Buffer* LoadMap32(uint16_t mapNum);
        
        void LoadMap16(Overworld::Area &area,
                       uint16_t         subArea,
                       map16Pos         p);
        
        Thorn::Block
        ThornBlockFromIndirect(Thorn::IndirectBlock b)
        const;
        
        void LoadThornBlocks();
        
        string GetFloorBlock(int i);
    };
    
    template<>
    void RomLoader::LoadTile(Tile<Pixel3bpp> &tile,
                             Buffer *b,
                             uint32_t position);
    
    template<>
    void RomLoader::LoadTile(Tile<PixelMode7> &tile,
                             Buffer           *b,
                             uint32_t          position);
    
    class LegacyLoader : public RomLoader
    {
    public:
        

    };
    
    class ModernLoader : public RomLoader
    {
    public:
    
        ModernLoader(ZeldaGame &p_zg,
                     ExpandInfo &ei)
        
        : RomLoader(p_zg, ei)
        {
        
        }
        
        int GetSprGfxPtr(int index);
        int GetBgGfxPtr(int index);
        
        
        
    };
    
}
    
#endif
