#-------------------------------------------------------------------------------
# Name:        thorn_ply
# Purpose:
#
# Author:      MathOnNapkins
#
# Created:     01/03/2012
# Copyright:   (c) MathOnNapkins 2012
# Licence:     MIT
#-------------------------------------------------------------------------------
#!/usr/bin/env python

import os

from thorn_lex import thornLexer

from thorn_yacc import thornParser

from thorn_helpers import *

# ==============================================================================

convertible = {
                # Offset is the lowest common denominator of
                # byte width types
                'offset' : [ 'offset' ],
                'repeat' : [ 'offset', 'repeat' ],
                'number' : [ 'offset', 'repeat', 'chr'],
                
                # The only conversion that can happen for blocks is
                # const to nonconst block_refs, and when passing as a param
                # to an object, or assigning to a non constant block_ref
                # variable.
                'block_ref' : [ 'block_ref' ],
                'block_flip' : [ 'block_flip' ],
                'block_dimension' : [ 'block_dimension' ],
                'const block_ref' : [ 'block_ref', 'const block_ref' ],
                
                # Can't think of any logical situations where these are 
                # convertible to other types.
                'background' : [ 'background' ],
                'truth_value' : [ 'background', 'truth_value' ],
                
                # Same here...
                'obj_ref' : [ 'obj_ref' ],
                
                # Word width types
                'chr' : [ 'chr' ],
                
              }

class ast:
    
    # __________________________________________________________________________
    # static members of the class so we can check for jerks trying to redefine
    # default objects
    default_obj = [
                    obj_def('Draw', [ obj_param('block_ref', 'b0') ] ),
                    
                    obj_def('Prioritize',[ obj_param('truth_value', 'p0'),
                                           obj_param('offset', 'rows'),
                                           obj_param('offset', 'columns') ] ),
                    
                    obj_def('SetAttr', [ obj_param('offset', 'a0'),
                                         obj_param('offset', 'rows'),
                                         obj_param('offset', 'columns') ] )
                  ]
    
    default_obj_dict = { k.name : k for k in default_obj }
    
    # __________________________________________________________________________
    def __init__(self, module):
        
        self.context = OrderedDict()
        
        for node in module:
            
            mod_dict = self.context
            
            name = node.obj.name
            
            if not name in mod_dict:
                
                self.context[node.obj.name] = node
            
            else:
                
                # raise Exception("Symbol %s already in symbol table!" % (name))
                
                print "%s Already in symbol table!" % (name)
        
        d = { k.obj.name : k.obj for k in module if k.typ == 'block_def' }
        
        self.block_dict = d
        
        d = { k.obj.name : k.obj for k in module if k.typ == 'obj_def' }
        
        self.obj_dict = d
        
        for k in self.default_obj_dict:
            
            if k in self.obj_dict:
                
                raise Exception
        
    # __________________________________________________________________________
    def resolve_expr_type(self, e, local_context):
        
        before_node_type = e.typ
        node_type        = e.typ
        
        # ----------------------------------
        
        # Only expression nodes and identifiers should need to be resolved,
        # currently.
        if node_type == 'expr':
            
            arith_ops = [ '-', '+', '*', '/' ]
            
            logic_ops = [ '==', '!=' ]
            
            arith_types = [
                            'number',
                            'repeat',
                            'offset',
                          ]
            
            block_types = [
                            'block_ref',
                            'const block_ref',
                          ]
            
            num_types    = arith_types[0:1]
            repeat_types = arith_types[0:2]
            
            # ----------------------------------
            
            ex = e.obj
            
            lhs = ex.left
            op  = ex.op
            rhs = ex.right
            
            # recursively resolve the type of the left and right nodes.
            # These may be 'expr' type nodes themselves, hence the recursion.
            left_type  = self.resolve_expr_type(lhs, local_context)
            right_type = self.resolve_expr_type(rhs, local_context)
            
            if op in arith_ops:
                
                if left_type in arith_types:
                
                    if right_type in arith_types:
                        
                        # figure out the conversion process!!
                        valid  = right_type in convertible[left_type]
                        valid |= left_type in convertible[right_type]
                        
                        if not valid:
                            
                            # Neither type can be converted to the other,
                            # so we have serious problems!
                            raise Exception
                        
                        else:
                            
                            # The arithmetic types form a hierarchy going from
                            # numeric up to offset. Resolve to the most
                            # sophisticated type (offset at maximum).
                            left_index  = arith_types.index(left_type)
                            right_index = arith_types.index(right_type)
                            
                            resolved_index = max(left_index, right_index)
                            
                            node_type = arith_types[resolved_index]
                            
                else:
                    
                    # Can't use arithmetic operators with non arithmetic types.
                    raise Exception
                
            elif op == '.':
                
                if left_type not in block_types:
                    
                    # The accessor operator is only for block references.
                    raise Exception
                
                if right_type == 'block_dimension':
                    
                    # Obtaining width or height of a block
                    node_type = 'offset'
                
                elif right_type == 'block_flip':
                    
                    # The type is unchanged, it just has a different
                    # orientation.
                    node_type = left_type
                
                else:
                    
                    # Anything else would be uncivilized! (block references
                    # have no other properties, currently.
                    raise Exception                                            
            
            elif op in logic_ops:
                
                # The operands of logical operators have an either/or approach.
                # If one is convertible to the other, we have a valid
                # logical expression.
                if right_type in convertible[left_type]:
                    
                    node_type = 'truth_value'
                    
                elif left_type in convertible[right_type]:
                    
                    node_type = 'truth_value'
            
                else:
                    
                    raise Exception
                    
            elif op == 'in':
                
                # Ah yes, the 'in' expression type. What a lovely, asinine
                # construct. Why do I torture myself like this?
                
                if left_type == 'tuple' or right_type != 'tuple':
                    
                    # We're trying to find if an expression is inside a tuple.
                    # finding a tuple inside of a tuple might be valid in some
                    # strange circumstance I haven't thought of, but I feel
                    # it's better to kaboom the program if this happens.
                    # Also, the right hand expression needs to be a tuple,
                    # otherwise you're just trollin'.
                    raise Exception
                
                for n in rhs.obj:
                    
                    element_type = self.resolve_expr_type(n,
                                                          local_context)
                    
                    left_convertible = left_type    in convertible[element_type]
                    elem_convertible = element_type in convertible[left_type]
                    
                    if not left_convertible and not elem_convertible:
                        
                        raise Exception
                    
                node_type = 'truth_value'
                
            else:
                
                if op == '.':
                    
                    print "wtf"
                
                # Any one else want some? We've got no other valid operators
                # to work with, that we know of. KABOOM! For now.
                raise Exception
        
        elif node_type == 'identifier':
            
            # check if it's a reference to a block (global scope)
            if e.obj in self.block_dict:
                
                node_type = 'const block_ref'
            
            # is it a reference to an object script? (global scope)
            elif e.obj in self.obj_dict:
                
                node_type = 'const obj_ref'
                
            # is it a parameter of the object script or a local variable
            # thereof?
            elif e.obj in local_context:
                
                node_type = local_context[e.obj].typ
            
            else:
                
                # Wasn't found, so it's an undeclared identifier.
                raise Exception
        
        elif node_type == 'unary_minus':
            
            return self.resolve_expr_type(e.obj,
                                          local_context)
        
        return node_type
    
    # __________________________________________________________________________
    def print_obj(self, caller, block, local_context):
        
        # iterate over obj statements, get type and statement object
        for n in block:
            
            typ, obj = n.typ, n.obj
            
            if typ == 'obj_call':
                
                call_ref = obj
                
                # Get the corresponding object that the name corresponds to
                if call_ref.name in self.obj_dict:
                    
                    call_def = self.obj_dict[call_ref.name]
                    
                elif call_ref.name in self.default_obj_dict:
                    
                    call_def = self.default_obj_dict[call_ref.name]
                    
                else:
                    
                    raise Exception
                
                sig = call_def.get_signature()
                
                # Number of parameters need to match
                if len(sig) != len(call_ref.params):
                    
                    raise Exception()
                
                for (i, e) in enumerate(call_ref.params):
                    
                    node_type = self.resolve_expr_type(e,
                                                       local_context)
                    
                    # get rid of constness if you're just passing the
                    # identifier as a parameter
                    if node_type.startswith("const "):
                        node_type = node_type.replace("const ", "")
                        
                    if node_type != sig[i]:
                        
                        if sig[i] not in convertible[node_type]:
                        
                            raise Exception
                
            elif typ == 'var_decl':
                
                # aliased, just to avoid confusion
                v_decl = obj
                
                if not v_decl.name in local_context:
                    
                    resolved_type = self.resolve_expr_type(v_decl.rhs,
                                                           local_context)
                    
                    if v_decl.typ not in convertible[resolved_type]:
                        raise Exception
                    
                    # store a node that is a quick reference of the 
                    # variable declaration's type and name.
                    local_context[v_decl.name] = node(v_decl.typ, v_decl.name)
                    
                else:
                    
                    # can't declare the same variable name twice
                    # in the same object (even if it falls out of scope)
                    # clarification: objects don't have scoped variables
                    # other than at the object level. Different objects can
                    # define locals with the same name.
                    raise Exception
                
            elif typ == 'assign':
                
                v_assign = obj
                
                lhs_type = self.resolve_expr_type(v_assign.lhs,
                                                  local_context)
                
                rhs_type = self.resolve_expr_type(v_assign.rhs,
                                                  local_context)
                
                if lhs_type not in convertible[rhs_type]:
                    
                    raise Exception
                
            elif typ == 'loop_block':
                
                loop_expr = obj.loop_expr
                
                loop_expr_type = self.resolve_expr_type(loop_expr,
                                                        local_context)
                
                if 'repeat' not in convertible[loop_expr_type]:
                    
                    raise Exception('Incompatible loop expression type')
                    
                self.print_obj(caller, obj.block, local_context)
            
            elif typ == 'conditional_block':
                
                cond_expr = obj.cond_expr
                
                cond_expr_type = self.resolve_expr_type(cond_expr,
                                                        local_context)
                
                if 'truth_value' not in convertible[cond_expr_type]:
                    
                    raise Exception('Incompatible conditional expression type.')
                
                self.print_obj(caller, obj.block, local_context)
            
            else:
                
                raise Exception('Unhandled obj_statement node type')
                        
# ==============================================================================

def get_text(path, name):

    try:
        
        f = open(os.path.join(path, name), "rt");
        
        content = f.read();
        
        f.close()
        
    except:
        
        print "Couldn't find file!"
        return None
    
    return content

# ==============================================================================

def main():
    
    path = r"."
    
    # ------------------------------------
    
    objects = get_text(".", "objects_mockup.thorn")
    blocks  = get_text(".", "blocks.thorn")
    
    thlex = thornLexer()
    thpar = thornParser()
    
    objects_tree = thpar.parser.parse(input = objects,
                                      lexer = thlex.lexer,
                                      debug = 0)
    
    blocks_tree = thpar.parser.parse(input = blocks,
                                     lexer = thlex.lexer,
                                     debug = 0)
    
    # combine the two parse trees together
    tree = ast(objects_tree + blocks_tree)
    
    block_strings = []
    
    for block_name in sorted( tree.block_dict.keys() ):
        
        b = tree.block_dict[block_name]
        
        height = b.params[1].obj
        width  = b.params[2].obj
        
        if not b.indices:
            
            b.indices = range(0, height * width)
        
        elif len(b.indices) != (height * width):
        
            # The indices list supplied does not form a block of the 
            # size specified!
            raise Exception
        
        if len(b.params) == 3:
            
            indir_b = indir_block_def(b.name,
                                      b.params[0].obj,
                                      b.params[1].obj,
                                      b.params[2].obj,
                                      b.indices)
            
            block_strings.append( indir_b.cpp_str() )
        
    print ",\n".join(block_strings)
        
    return
    
    for obj_name in sorted( tree.obj_dict.keys() ):
        
        print obj_name
        
        obj = tree.obj_dict[obj_name]
        
        root_block = obj.block
        
        local_context = {
                            k : node(obj.params[k], k)
                            for k in obj.params.keys()
                        }
        
        tree.print_obj(obj, root_block, local_context)
    
    return
    
# ==============================================================================

if __name__ == '__main__':
    main()
    