# -*- coding: utf-8 -*-
#This file is inspired by/copies qgraphicsview.h and qgraphicsscene.h 
# from the Qt project. It is provided under the terms of the LGPL
# independently from whatever license the rest of the project uses.

import sys
import math
from PySide.QtGui import *
from PySide.QtCore import *


#This class mimics a lot of the functionality in QGraphicsView, but
#  is more restrictive in that it forces components to be all the same
#  size and also constrained to a grid. This allows much faster
#  item changing, insertion, and removal.
#The total area starts out at some arbitrary size and expands to 
#  fit the total number of tiles. It never shrinks (individual
#  tiles can be reclaimed by setting them to None), but can be
#  bound by a rectangle which "hides" any tiles outside of it.
#Hopefully we can implement some kind of "offset()" logic to allow 
#  for some padding on the left/right side of the screen (for 
#  showing metadata, etc., haven't figured this out yet).
#Note that a lot of these notes are for FixedGridGraphicsScene, but since
#  I haven't implemented that yet I'm leaving them here. The View will
#  mostly just mimic QGraphicsView; I'd subclass it, but it doesn't seem
#  to have everything wrapped exactly right (e.g., "drawItems()" or whatever)
#Note that the three functions scrollContentsBy(), paintEvent() and viewportEvent() handle most of the
#  complexity of this class. I've left in a lot of the default options for now
#  (might as well port them all if we can), but we can remove things we don't need/want
#  or troublesome code later. In particular, I'm not 100% sure that all references and
#  pointers are being passed correctly.
class FixedGridGraphicsView(QGraphicsView):
  def __init__(self, parent=None):
    super(FixedGridGraphicsView, self).__init__(parent)

  def event(self, event):
    #print 'FixedGridGraphicsView::' , stack()[0][3] , '::' , event.type()
    #if event.type() == QEvent.Type.Paint:
    #  print '  ' , event.rect()
    return super(FixedGridGraphicsView, self).event(event)

  def paintEvent(self, event):
    #If there's no scene, delegate this call to the parent (which will in turn 
    #  delegate it to the abstract scroll area
    if not self.scene():
      super(FixedGridGraphicsView, self).paintEvent(event)
      return

    #Determine the exposed region in scene coordinates
    exposedRegion = event.region()
    exposedSceneRect = self.mapToScene(exposedRegion.boundingRect()).boundingRect()

    #Set up the painter around the current viewport.
    #QGraphicsView will request a viewport upon its creation, and will
    #  will not request a new one for its lifetime. The viewport is a QWidget.
    painter = QPainter(self.viewport())

    #Save the current painter state. 
    # (This is unnecessary (but harmless) if rubberbanding is disabled)
    painter.save()

    #Set up render hints
    #NOTE: By default we turn OFF text antialiasing.
    painter.setRenderHint(QPainter.TextAntialiasing, False)

    #Set up viewport transform
    viewTransformed = self.isTransformed()
    if viewTransformed:
      painter.setWorldTransform(self.viewportTransform())
    viewTransform = painter.worldTransform()

    #Now that the painter is set up, dispatch all drawing.
    self.performDrawing_Background(painter, exposedSceneRect)
    self.performDrawing_Items(painter, exposedRegion, exposedSceneRect, viewTransformed, viewTransform)
    self.performDrawing_Foreground(painter, exposedSceneRect)

    #Restore the painter state.
    # (This is unnecessary (but harmless) if rubberbanding is disabled)
    painter.restore()

    #Painting is DONE
    painter.end()


  #Step 1: Background. Complicated by the fact that we may not have to repaint ALL of it, and
  #        parts of it may be cached
  def performDrawing_Background(self, painter, exposedSceneRect):
    cacheMode = self.cacheMode()
    if cacheMode==QGraphicsView.CacheNone:
        #The simplest option: just repaint it. Assume painter state is preserved 
        #  (may be slower, but is generally more stable)
        painter.save();
        self.drawBackground(painter, exposedSceneRect)
        painter.restore();
    elif cacheMode==QGraphicsView.CacheBackground:
      raise Exception("Drawing with CacheBackground is currently disabled.")
    else: raise Exception("Unexpected cacheMode: " + str(cacheMode))


  #Step 2: Items. May be painted indirectly. Determining which items to paint is 
  #        one of the key speedups provided by this class.
  def performDrawing_Items(self, painter, exposedRegion, exposedSceneRect, viewTransformed, viewTransform):
    if not (self.optimizationFlags() & QGraphicsView.IndirectPainting):
      #Call our internal drawing function
      rectAdjust = 1 if (self.optimizationFlags() & QGraphicsView.DontAdjustForAntialiasing) else 2
      self.scene().drawItemsPrivate(painter, rectAdjust, (viewTransform if viewTransformed else None), exposedRegion, self.viewport())

      #Ensure that the painter's state has been restored. This is simple, so just do it here; 
      #  no need for save() and restor() since only a few things changed
      painter.setOpacity(1.0)
      painter.setWorldTransform(viewTransform)
    else:
      raise Exception("Drawing with IndirectPainting is currently disabled.")

  #Step 3: Foreground. Very, very simple.
  def performDrawing_Foreground(self, painter, exposedSceneRect):
    self.drawForeground(painter, exposedSceneRect)




#And now...
#  tileSize is the size of each QPixmap tile
#  bufferSize is the size of each tile's border
#  numLayers is the number of tile layers. 
#  initArraySize is the initial size of the array of tiles
class FixedGridGraphicsScene(QGraphicsScene):
  def __init__(self, tileSize, bufferSize, numLayers=1, initArraySize=(32,32), parent=None):
    super(FixedGridGraphicsScene, self).__init__(parent)
    self.tileSize = tileSize
    self.bufferSize = bufferSize

    #3D arrays ftw. layers[layerID][yPos][xPos]
    self.layers = [[[None]*initArraySize[0] for x in xrange(initArraySize[1])] for x in xrange(numLayers)]  
    self.layerVisibility = [True] * numLayers
    self.visibleTiles = None

  def getTileSize(self):
    return self.tileSize

  def createItemGroup(self, *args):
    raise Exception("createItemGroup() not supported. Use layers instead")

  def setSceneRect(self, *args):
    raise Exception("setSceneRect() not supported. Use setVisibleTiles() instead")

  def addItem(self, *args):
    raise Exception("addItem() not supported. Use setTile()/newTile() instead")

  def setVisibleTiles(self, offsetX, offsetY, widthInTiles, heightInTiles):
    self.visibleTiles = (offsetX, offsetY, widthInTiles, heightInTiles)
    innerLimits = self.getPixelRectFromTiles(offsetX, offsetY) #A little messy
    outerLimits = self.getPixelRectFromTiles(widthInTiles, heightInTiles)
    super(FixedGridGraphicsScene, self).setSceneRect(innerLimits[0]-self.bufferSize, innerLimits[1]-self.bufferSize, outerLimits[0], outerLimits[1])

  def setLayerVisibility(self, layerID, isVisible):
    if layerID<0 or layerID>=len(self.layers):
      raise Exception('Layer ID out of range in addItem()')
    self.layerVisibility[layerID] = isVisible
    

  #Set the tile at the given location. (Allows shared tiles)
  #This expands the list of tiles as needed. 
  def setTile(self, item, tileX, tileY, layerID):
    self.tileAccessHelper(layerID, tileX, tileY, item, True)
    self.layers[layerID][tileY][tileX] = item

  #Retrieve the tile at the given location. If you modify this
  #  tile, be SURE you're not modifying a shared tile. 
  #  This class will not track it for you.
  #Creates a new tile if the requested tile does not exist.
  #TODO: If we store pairs of tiles (Ref, Real), then we can use "get"
  #      to _always_ return the Real version. Have to look into this.
  def getTile(self, tileX, tileY, layerID, expandTilesAsNeeded):
    self.tileAccessHelper(layerID, tileX, tileY, None, expandTilesAsNeeded)
    if not self.layers[layerID][tileY][tileX]:
      self.layers[layerID][tileY][tileX] = QPixmap(self.tileSize, self.tileSize)
    return self.layers[layerID][tileY][tileX]

  #Create a new tile at the given location and return it. (For non-shared tiles)
  #This expands the list of tiles as needed. 
  def newTile(self, tileX, tileY, layerID):
    self.tileAccessHelper(layerID, tileX, tileY, None, True)
    self.layers[layerID][tileY][tileX] = QPixmap(self.tileSize, self.tileSize)
    return self.layers[layerID][tileY][tileX]


  #Internal helper function: perform expansion and check the tile type
  def tileAccessHelper(self, layerID, tileX, tileY, item=None, expand=False):
    #Sanity checks: Item
    if item:
      if not isinstance(item, QPixmap):
        raise Exception('addItem() must be performed on a QPixmap')
      if (not item.width()==self.tileSize) or (not item.height()==self.tileSize):
        raise Exception('addItem() must fit with the requested tile size')

    #Sanity check: layers
    if layerID<0 or layerID>=len(self.layers):
      raise Exception('Layer ID out of range in addItem()')

    #Sanity check: x/y
    if not expand:
      if (tileY >= len(self.layers[layerID])) or (tileX >= len(self.layers[layerID][0])):
        raise Exception('Tile ID out of range in addItem(), and expand is set to False')

    #Expand _all_ layers in parallel as required. (Layers never decrease in size.)
    if expand:
      for layer in self.layers:
        self.expandLayer(layer, tileX, tileY)

  #Clear whichever tile was at the given location. 
  #Does NOT expand the list of tiles
  def clearTile(self, tileX, tileY, layerID):
    self.tileAccessHelper(layerID, tileX, tileY, None, False)
    self.layers[layerID][tileY][tileX] = None


  def getPixelRectFromTiles(self, tileX, tileY):
    tsP1 = self.tileSize + self.bufferSize
    return (tileX*tsP1+self.bufferSize, tileY*tsP1+self.bufferSize, self.tileSize, self.tileSize)

  def getTileAxisFromPixel(self, pixMin, pixMax):
    #Math is escaping me at the moment.
    pixMin -= self.bufferSize
    pixMax -= self.bufferSize
    tsP1 = float(self.tileSize + self.bufferSize)
    pixMin = int(math.floor(pixMin / tsP1))
    pixMax = int(math.ceil(pixMax / tsP1))
    return (pixMin, pixMax)
    

  def getTileRectFromPixels(self, pixelRect):
    #Each "axis" is returned as (min, max) including overflow
    xAxis = self.getTileAxisFromPixel(pixelRect.x(), pixelRect.x()+pixelRect.width())
    yAxis = self.getTileAxisFromPixel(pixelRect.y(), pixelRect.y()+pixelRect.height())

    #Bound by the actual number of tiles (and zero!)
    return [max(0, xAxis[0]), max(0, yAxis[0]), min(xAxis[1], -1+len(self.layers[0][0])), min(yAxis[1], -1+len(self.layers[0]))]

 
  #x,y = indices of items we want to add. E.g., layer[y][x] = item
  #expand with None.
  #Note: Let's just double it.
  def expandLayer(self, layer, x, y):
    #Expand X
    if x>=len(layer[0]):
      for row in layer:
        row += [None]*(2*len(row))

    #Expand Y, and account for the already expanded X
    if y>=len(layer):
      layer += [[None]*len(layer[0]) for i in xrange(2*len(layer))]



  #Private item drawing code. viewTransform may be null
  def drawItemsPrivate(self, painter, rectAdjust, viewTransform, exposedRegion, viewport):
    #Create a rectangle to represent the region which is exposed.
    exposedSceneRect = QRectF()
    if exposedRegion:  #Likely always true
        #Take an educated guess. Apply the current view transformation if one exists.
        exposedSceneRect = exposedRegion.boundingRect().adjusted(-1, -1, 1, 1)
        if viewTransform:
            inverted = viewTransform.inverted()  #(Transform, bool)
            if inverted[1]:
              inverted = viewTransform.inverted()[0]
              exposedSceneRect = inverted.mapRect(exposedSceneRect)

    #Finally, retrieve and paint all items which are visible. This is slightly complicated by the fact that
    #  a world transform might be active (but that should be fairly transparent)
    #  1) exposedSceneRect contains the "scene" coordinates that need repainting, +/- 1 pixel
    #  2) We then compute the "minTile" and "maxTile" for each component. Make sure to include tiles partway on-screen.
    #  3) Then it's just a matter of iterating through these and printing them (layer by layer)
    tileRect = self.getTileRectFromPixels(exposedSceneRect)  #Note: it's a "rectangle" in the Win32 sense, not the width/height sense
    tileRect[2] += 1
    tileRect[3] += 1

    #Bound the ranges if requested
    #TODO: Avoid duplicating code...
    #TODO: Wow this is a mess.
    if self.visibleTiles:
      tileRect[2] = max(min(tileRect[2], self.visibleTiles[0]+self.visibleTiles[2]), self.visibleTiles[0])
      tileRect[3] = max(min(tileRect[3], self.visibleTiles[1]+self.visibleTiles[3]), self.visibleTiles[1])

    #Note: Why doesn't it work to use this inside the loop itself???
    xRange = range(tileRect[0], tileRect[2])
    yRange = range(tileRect[1], tileRect[3])

    layerRange = xrange(0, len(self.layers))
    for l in layerRange:
      if not self.layerVisibility[l]:
        continue #Skip transparent layers
      layer = self.layers[l]
      for y in yRange:
        for x in xRange:
          tile = layer[y][x]
          if tile:
            tileRect = self.getPixelRectFromTiles(x,y) #Note: Not the fastest, but oh well
            painter.drawPixmap(tileRect[0], tileRect[1], tile)



