# -*- coding: utf-8 -*-

import sys
from PySide.QtGui import *
from PySide.QtCore import *
import blackmagic as bm
import random

from customqt import FixedGridGraphicsView, FixedGridGraphicsScene



#Simple custom drawing class
class MapView(FixedGridGraphicsView):
  def __init__(self, parent=None):
    super(MapView, self).__init__(parent)
    self.game = None
    self.mapLayerVis = True
    self.annotLayerVis = True
    self.validMap = False

    #Border area size
    self.BorderSize = (8, 7)
    self.hasLoadedSideMaps = False

    #Share this
    self.QuestionTile = None

    #Purely for fun
    self.triCount = 1

    #Offset by 8x7; this allows us to draw some nice border tiles as needed.
    self.offset = (self.BorderSize[0], self.BorderSize[1])
	
    #Create a Scene which rendrs into a fixed grid.
    self.setScene(FixedGridGraphicsScene(16, 1, 2))
    self.scene().setBackgroundBrush(QBrush(Qt.black))

  #Translate from input (x,y) values to offset values (raw values used by the underlying scene)
  def getTranslated(self, point):
    return (point[0]+self.offset[0], point[1]+self.offset[1])

  def setGame(self, game):
    self.game = game

  def newMap(self, mapID):
    self.validMap = True
    self.hasLoadedSideMaps = False
    self.rebuildScene(mapID)

  def triCountAdd(self, amount):
    #Might as well do something with those null button presses.
    self.triCount += amount
    self.triCount = min(max(self.triCount, 1), 4) #It really looks bad after 4
    self.viewport().update()

  def renderAnnotation(self, tileQuad, tileX, tileY):
    tileSize = self.scene().getTileSize()

    #Render our ??? tile *once*
    if not self.QuestionTile:
      self.QuestionTile = QPixmap(tileSize, tileSize)
      self.QuestionTile.fill(QColor(0x33, 0x33, 0x33))
      gcOver = QPainter(self.QuestionTile)
      gcOver.setFont(QFont("Arial", 8))
      gcOver.setPen(QColor(0xAA, 0xAA, 0xAA))
      gcOver.drawText(QRectF(0, 0, tileSize, tileSize), Qt.AlignCenter, "??")
      gcOver.end()

    #If this tile slot is empty, just set the question tile
    if not tileQuad:
      self.scene().setTile(self.QuestionTile, tileX, tileY, 1)
      return

    #Otherwise, draw something (flipped tile markings, for now)
    #NOTE: There's only a few hundred combinations; we could cache them if we wanted to.
    annot = self.scene().newTile(tileX, tileY, 1) #Has to be "new"; the tile may have been "set" before.
    annot.fill(QColor(0x00, 0x00, 0x00, 0x00))
    sq = tileSize/2

    gcOver = QPainter(annot)
    unHinted = QFont("Arial", 6)
    unHinted.setStyleStrategy(QFont.NoAntialias)
    gcOver.setFont(unHinted)
    gcOver.setPen(QColor(0xFF, 0x00, 0x00))

    tileOffsets = ((0,0), (1,0), (0,1), (1,1)) #(UpLeft, UpRight, LowLeft, LowRight)

    for i in xrange(4):
      tileEntry = tileQuad.get(i)
      if tileEntry.h_flip or tileEntry.v_flip:
        letter = 'B'
        if tileEntry.h_flip and not tileEntry.v_flip:
          letter = 'H'
        elif tileEntry.v_flip and not tileEntry.h_flip:
          letter = 'V'
        r = QRectF(tileOffsets[i][0]*sq, tileOffsets[i][1]*sq, sq, sq)
        gcOver.drawText(r, Qt.AlignCenter, letter)
    gcOver.end()


  def renderTile(self, tileQuad, tileX, tileY, bgColor):
    #Nothing to draw?
    if not tileQuad:
      self.scene().clearTile(tileX, tileY, 0)
      return

    #Draw each quadrant
    tileSize = self.scene().getTileSize()
    tileOffsets = ((0,0), (1,0), (0,1), (1,1)) #(UpLeft, UpRight, LowLeft, LowRight)
    sq = tileSize/2

    tile = self.scene().getTile(tileX, tileY, 0, True)
    tile.fill(QColor(0x00, 0x00, 0x00)) #Shouldn't matter, but let's be thorough
    gc = QPainter(tile)

    for i in xrange(4):
      #Draw the tile
      tileEntry = tileQuad.get(i)
      quad = self.game.getMap8Tile(tileEntry, bgColor)
      r = QRectF(tileOffsets[i][0]*sq, tileOffsets[i][1]*sq, sq, sq)
      gc.drawImage(tileOffsets[i][0]*sq, tileOffsets[i][1]*sq, quad)
    gc.end()



  def rebuildScene(self, mapID):
    #Failsafe
    if (not self.game) or (not self.game.getCurrMap()):
      return
    currMap = self.game.getCurrMap()

    #The QGraphics scene makes it easy to handle lots of objects at different locations
    scene = self.scene()
    bgColor = self.game.getAreaBgColor(mapID)
    for y in xrange(currMap.area_map.height()):
      for x in xrange(currMap.area_map.width()):
        #Translate
        pos = self.getTranslated((x, y))

        #Render it
        tileSlot = currMap.area_map.get_slot(x,y)
        tileQuad = None if tileSlot.IsEmpty() else self.game.GetMap16FromMap8(tileSlot.get())
        self.renderTile(tileQuad, pos[0], pos[1], bgColor)
        self.renderAnnotation(tileQuad, pos[0], pos[1])

    #Establish some ordering
    scene.setLayerVisibility(0, True)
    scene.setLayerVisibility(1, False)

    #Now hook it up to a QGraphicsView
    scene.setVisibleTiles(self.offset[0], self.offset[1], currMap.area_map.width(), currMap.area_map.height())
    self.resetTransform()
    self.viewport().update()
    self.setFocus()


  def clearSurrounding(self, mapWidth, mapHeight, ClearLeft, ClearRight, ClearTop, ClearBottom):
    #Clear top and bottom
    if ClearTop or ClearBottom:
      for x in xrange(mapWidth):
        for row in xrange(self.offset[1]):
          if ClearTop:
            pos = self.getTranslated((x, -1-row))
            self.scene().clearTile(pos[0], pos[1], 0)
          if ClearBottom:
            pos = self.getTranslated((x, mapHeight+row))
            self.scene().clearTile(pos[0], pos[1], 0)
    #Clear left and right
    if ClearLeft or ClearRight:
      for y in xrange(mapHeight):
        for col in xrange(self.offset[0]):
          if ClearLeft:
            pos = self.getTranslated((-1-col, y))
            self.scene().clearTile(pos[0], pos[1], 0)
          if ClearRight:
            pos = self.getTranslated((mapWidth+col, y))
            self.scene().clearTile(pos[0], pos[1], 0)


  def moveToEnd(self, lookup, startX, startY, xPlus, yPlus):
    x = startX
    y = startY
    while True:
      if x+xPlus<0 or x+xPlus>=len(lookup[y]): break
      if y+yPlus<0 or y+yPlus>=len(lookup): break
      if lookup[y+yPlus][x+xPlus] != lookup[y][x]: break
      x += xPlus
      y += yPlus
    return (x,y)



  #Returns a list of maps to use for looking up tiles.
  #Also includes the starting (tile) offset for the first map,
  #  in case a map spans multiple rows/columns
  #TODO: This doesn't work right for large maps (because they map to the LAST tile written)
  def getMapList(self, lookup, x, y, numTiles, horiz):
    #Step 1: If this map extends through several grids, set the offset
    #TODO: It's not that hard to generalize this, but it makes things complicated    
    offset = 0
    orig = lookup[y][x]
    if horiz and x>0 and lookup[y][x-1]==orig:
      offset = 16
    if (not horiz) and y>0 and lookup[y-1][x]==orig:
      offset = 14

    #Step 2: Prepare the mapList
    mapList = [orig]
    numTiles -= (16 if horiz else 14)
    if numTiles>0:
      if horiz and x+1<len(lookup[y]):
        if lookup[y][x+1] != lookup[y][x]:
          mapList.append(lookup[y][x+1])
      if not horiz and y+1<len(lookup):
        if lookup[y+1][x] != lookup[y][x]:
          mapList.append(lookup[y+1][x])

    #Done
    return (mapList, offset)


  #Using an offset and mapList populate a 'side' of the current viewing area. 
  #If you run out of tiles, clear the remaining ones.
  #TODO: We can combine fillHoriz() and fillVert()... if we really want a headache, that is....
  def fillVertTiles(self, srcY, mapList, onRight):
    #
    currMap = self.game.getCurrMap()
    otherMap = self.game.getMap(mapList[0])
    currMapID = 1
    for destY in xrange(currMap.area_map.height()):
      for col in xrange(self.offset[0]):
        #Convert to destX, translate
        destX = currMap.area_map.width()+col if onRight else -1-col
        pos = self.getTranslated((destX, destY))

        #Out of maps?
        if not otherMap: 
          self.scene().clearTile(pos[0], pos[1], 0)          
          continue

        #Determine srcX
        srcX = otherMap.area_map.width()-1-col if not onRight else col

        #Copy
        bgColor = self.game.getAreaBgColor(mapList[currMapID-1])  #TEMP
        tileSlot = otherMap.area_map.get_slot(srcX,srcY)
        tileQuad = None if tileSlot.IsEmpty() else self.game.GetMap16FromMap8(tileSlot.get())
        self.renderTile(tileQuad, pos[0], pos[1], bgColor)

      #Increment, possibly change maps or null out
      srcY += 1
      if otherMap and (srcY >= otherMap.area_map.height()):
        otherMap = self.game.getMap(mapList[currMapID]) if currMapID<len(mapList) else None
        currMapID += 1
        srcY = 0

        

  def fillHorizTiles(self, srcX, mapList, onBottom):
    #
    currMap = self.game.getCurrMap()
    otherMap = self.game.getMap(mapList[0])
    currMapID = 1
    for destX in xrange(currMap.area_map.width()):
      for row in xrange(self.offset[1]):
        #Convert to destY, translate
        destY = currMap.area_map.height()+row if onBottom else -1-row
        pos = self.getTranslated((destX, destY))

        #Out of maps?
        if not otherMap: 
          self.scene().clearTile(pos[0], pos[1], 0)          
          continue

        #Determine srcY
        srcY = otherMap.area_map.height()-1-row if not onBottom else row

        #Copy
        bgColor = self.game.getAreaBgColor(mapList[currMapID-1])  #TEMP
        tileSlot = otherMap.area_map.get_slot(srcX,srcY)
        tileQuad = None if tileSlot.IsEmpty() else self.game.GetMap16FromMap8(tileSlot.get())
        self.renderTile(tileQuad, pos[0], pos[1], bgColor)

      #Increment, possibly change maps or null out
      srcX += 1
      if otherMap and (srcX >= otherMap.area_map.width()):
        otherMap = self.game.getMap(mapList[currMapID]) if currMapID<len(mapList) else None
        currMapID += 1
        srcX = 0



  def loadSideMaps(self):
    self.hasLoadedSideMaps = True

    #Failsafe (put before "hasLoaded"?)
    if (not self.game) or (not self.game.getCurrMap()): return
    currMap = self.game.getCurrMap()

    #Step 1: Zero out all tiles in the "diagonal" buffer area
    for i in xrange(4):  #(UL, UR, LL, LR)
      for row in xrange(self.offset[1]):
        for col in xrange(self.offset[0]):
          x = -1-col
          y = -1-row
          if (i==1) or (i==3):
            x = currMap.area_map.width()+col
          if (i==2) or (i==3):
            y = currMap.area_map.height()+row
          pos = self.getTranslated((x, y))
          self.scene().clearTile(pos[0], pos[1], 0)

    #Step 2: For each remaining quadrant, load the adjacent map and
    #        fill in its tiles. This is complicated by a few points:
    # (1) there might not be an adjacent map
    # (2) there might be multiple adjacent maps
    # (3) there may be various alignment issues with adjacent maps
    # (4) we may not have any connectivity information at all
    lookup = self.game.annotations.areaLookup
    connectivity = self.game.annotations.areaConnections
    mapSize = (currMap.area_map.width(), currMap.area_map.height())
    if not lookup.has_key(self.game.getCurrMapID()):
      self.clearSurrounding(mapSize[0], mapSize[1], True, True, True, True)
    else:
      areaPos = lookup[self.game.getCurrMapID()]

      #Left
      topLeft = self.moveToEnd(connectivity, areaPos[0], areaPos[1], 0, -1)
      topLeft = self.moveToEnd(connectivity, topLeft[0], topLeft[1], -1, 0)
      if (topLeft[0]==0):
        self.clearSurrounding(mapSize[0], mapSize[1], True, False, False, False)
      else:
        (mapList, offset) = self.getMapList(connectivity, topLeft[0]-1, topLeft[1], currMap.area_map.height(), False)
        self.fillVertTiles(offset, mapList, False)

      #Right
      topRight = self.moveToEnd(connectivity, areaPos[0], areaPos[1], 0, -1)
      topRight = self.moveToEnd(connectivity, topRight[0], topRight[1], 1, 0)
      if (topRight[0]==len(connectivity[topRight[1]])-1):
        self.clearSurrounding(mapSize[0], mapSize[1], False, True, False, False)
      else:
        (mapList, offset) = self.getMapList(connectivity, topRight[0]+1, topRight[1], currMap.area_map.height(), False)
        self.fillVertTiles(offset, mapList, True)

      #Top
      if (topLeft[1]==0):
        self.clearSurrounding(mapSize[0], mapSize[1], False, False, True, False)
      else:
        (mapList, offset) = self.getMapList(connectivity, topLeft[0], topLeft[1]-1, currMap.area_map.width(), True)
        self.fillHorizTiles(offset, mapList, False)

      #Bottom
      lowerLeft = self.moveToEnd(connectivity, areaPos[0], areaPos[1], 0, 1)
      lowerLeft = self.moveToEnd(connectivity, lowerLeft[0], lowerLeft[1], -1, 0)
      if (lowerLeft[1]==len(connectivity)-1):
        self.clearSurrounding(mapSize[0], mapSize[1], False, False, False, True)
      else:
        (mapList, offset) = self.getMapList(connectivity, lowerLeft[0], lowerLeft[1]+1, currMap.area_map.width(), True)
        self.fillHorizTiles(offset, mapList, True)
    



  def paintEvent(self, event):
    #Common case
    if self.validMap:
      super(MapView, self).paintEvent(event)
      return

    #Fill
    painter = QPainter(self.viewport())
    bgColor = QColor(0x22, 0x22, 0x22)
    painter.fillRect(QRect(0, 0, self.viewport().width(), self.viewport().height()), bgColor)
    painter.setRenderHint(QPainter.TextAntialiasing, True)

    #Math!
    areaSize = min(self.viewport().width()/4.0, self.viewport().height()/2.0)
    viewCtr = QPointF(self.viewport().width()/2.0, self.viewport().height()/2.0)

    #Draw the first (and only) foreground triangle
    painter.setBrush(QBrush(QColor(0x59, 0x59, 0x1a)))
    painter.drawPolygon(QPolygonF([  QPointF(viewCtr.x()-areaSize/2.0, viewCtr.y()+areaSize/2.0),
                                     QPointF(viewCtr.x(),              viewCtr.y()-areaSize/2.0),
                                     QPointF(viewCtr.x()+areaSize/2.0, viewCtr.y()+areaSize/2.0)]))

    #Recursively eliminate it to "count"
    painter.setBrush(QBrush(bgColor))
    self.eraseTriangles(painter, viewCtr, areaSize, self.triCount)

    #Signal painter is done.
    painter.end()


  def eraseTriangles(self, painter, center, size, count):
    #Big erase
    painter.drawPolygon(QPolygonF([  QPointF(center.x()-size/4.0, center.y()),
                                     QPointF(center.x()+size/4.0, center.y()),
                                     QPointF(center.x(),          center.y()+size/2.0)]))

    #Continue? (NOTE: This is depth-first. We could make it breadth-first... but does it matter?)
    if count>1:
      count -= 1
      size /= 2
      #Avoid looping forever
      if size <= 1.0:
        return
      self.eraseTriangles(painter, QPointF(center.x()-size/2.0, center.y()+size/2.0), size, count)
      self.eraseTriangles(painter, QPointF(center.x(),          center.y()-size/2.0), size, count)
      self.eraseTriangles(painter, QPointF(center.x()+size/2.0, center.y()+size/2.0), size, count)


  def showHideMapLayer(self, show):
    self.mapLayerVis = show
    self.scene().setLayerVisibility(0, self.mapLayerVis)
    self.viewport().update()

  def showHideAnnotLayer(self, show):
    self.annotLayerVis = show
    self.scene().setLayerVisibility(1, self.annotLayerVis)
    self.viewport().update()

  #If true, it expands the current map's visibility by 
  # 8x7 tiles on each side, loaded from the previous map.
  def showBorderAreas(self, show):
    if not self.game or not self.game.getCurrMap(): return
    currMap = self.game.getCurrMap()
    currSize = (currMap.area_map.width(), currMap.area_map.height())

    if show:
      if not self.hasLoadedSideMaps:
        self.loadSideMaps()
      self.scene().setVisibleTiles(0, 0, currSize[0]+self.BorderSize[0]*2, currSize[1]+self.BorderSize[1]*2)
    else:
      self.scene().setVisibleTiles(self.offset[0], self.offset[1], currSize[0], currSize[1])
    self.viewport().update()



  #Handle some common key presses
  def keyPressEvent(self, event):
    #Zoom: Ctrl+, Ctrl-, Ctrl0 (reset)
    if event.modifiers()==Qt.ControlModifier:
      if event.key()==Qt.Key_Equal:  #Because Key_Plus means shift
        self.scale(1.20, 1.20)
      elif event.key()==Qt.Key_Minus:
        self.scale(0.80, 0.80)
      elif event.key()==Qt.Key_0:
        self.resetTransform()
        #self.currRotation = 0

    #Test
    if event.key()==Qt.Key_T:
      self.scene().setLayerVisibility(0, (False if self.mapLayerVis else True))
    if event.key()==Qt.Key_Y:
      self.scene().setLayerVisibility(1, (False if self.annotLayerVis else True))

    #Rotate! (Ctrl,< or >)
    if event.key()==44 and event.modifiers()==Qt.ControlModifier:
      self.rotate(2.0)
    elif event.key()==46 and event.modifiers()==Qt.ControlModifier:
      self.rotate(-2.0)








