# -*- coding: utf-8 -*-

from PySide import QtGui

import blackmagic as bm
import random

#Annotations class
class Annotations:
  def __init__(self):
    self.areaNames = {}
    self.areaBgColors = {}   #ARGB
    self.areaConnections = []  #2D array of area IDs
    self.areaLookup = {} #<int> -> (x,y) in areaConnections


#Helper wrapper for ZeldaGame that handles some translation issues 
# (e.g., map8, palette, etc.)
class ZGame:
  def __init__(self, game, annotations):
    self.game = game
    self.currMap = None
    self.currMapID = -1
    self.tileCache = {}
    self.annotations = annotations

  def GetMap16FromMap8(self, tileID):
    return bm.GetMap16FromMap8(self.game, tileID)

  def getAreaName(self, areaID):
    return str(areaID) + ": " +  self.annotations.areaNames.get(areaID, "")

  def getAreaBgColor(self, areaID):
    return self.annotations.areaBgColors.get(areaID, 0xFF000000)

  #Returns true on success
  #NOTE: This still doesn't handle map 0... unless I'm counting wrong
  def setCurrMap(self, currMapID):
    if currMapID<0 or currMapID>=self.game.areas.size():
      return False

    self.currMapID = currMapID
    res = self.game.areas[currMapID]
    self.tileCache.clear()
    self.currMap = res
    self.recalcMap()
    return True

  def getCurrMap(self):
    return self.currMap

  def getCurrMapID(self):
    return self.currMapID

  def getMap(self, mapID):
    if mapID<0 or mapID>=self.game.areas.size(): return None
    return self.game.areas[mapID]

  def getNumberOfMaps(self):
    return self.game.areas.size()


  #Convert a map8 tile into an 8x8 array of RGB pixel values
  def getMap8Tile(self, tileEntry, bgColor):
    #Some indexing properties
    tChr = tileEntry.chr
    tPal = tileEntry.pal
    half = self.halfPalSelect[tChr];

    #Make a key
    key = str(tChr) + ":" + str(tPal) + ":" + str(tileEntry.h_flip) + ":" + str(tileEntry.v_flip)

    #Add to the cache if needed
    if key not in self.tileCache:
      #Build up a ColorTile for reference, then fill all 8x8 pixels
      res = QtGui.QImage(8, 8, QtGui.QImage.Format_RGB32)
      res.fill(bgColor)
      cTile = bm.ColorTile(self.bgTiles[tChr], self.bgPals[2*tPal + half], 0)
      for y in xrange(8):
        for x in xrange(8):
          #Qt is pretty bad about converting strings, so we'll have to cache the actual QImage here.
          color = cTile.get(y, x, tileEntry.h_flip, tileEntry.v_flip).rgb()
          if (color&0xFF000000): #Opaque
            res.setPixel(x, y, color)

      #Update the cache
      if len(self.tileCache)>1000:
        self.tileCache.pop(random.choice(self.tileCache.keys()))
      self.tileCache[key] = res

    #Done
    return self.tileCache[key]


  #Perform any bookkeeping needed later (like bgTiles)
  def recalcMap(self):
    #SPRITES
    #print 'Test: ', bm.sprites(self.currMap, 1).size()
    #for spr in bm.sprites(self.currMap, 1):
    #  print 'Sprite: ' , spr.spriteNum



    #Build up our background palettes
    bgT = self.currMap.bg_pal_theme.Redeem()
    bgSec2 = bgT.secondary2.Redeem() if bgT.secondary2 else None
    self.buildBgPalettes(bgT.primary.Redeem(), bgT.secondary1.Redeem(), bgSec2, bgT.tertiary.Redeem())

    #Build up our set of background tiles & half-palettes
    self.buildBgTiles(self.currMap.bg_tile_theme_main.Redeem(), self.currMap.bg_tile_theme_aux.Redeem(), 8)


  def buildBgPalettes(self, bg_prim, bg_sec1, bg_sec2, bg_tert):
    #Reset
    self.bgPals = bm.Group_8_16()

    #Build 1:
    for i in xrange(5):
      for j in xrange(bg_prim.PaletteSize()):
        self.bgPals[(i+2)*2][j+1] = bg_prim[i][j]

    #Build 2:
    for i in xrange(3):
      for j in xrange(bg_sec1.PaletteSize()):
        self.bgPals[(i+2)*2+1][j+1] = bg_sec1[i][j]

    #Build 3
    if bg_sec2:
      for i in xrange(3):
        for j in xrange(bg_sec2.PaletteSize()):
          self.bgPals[(i+5)*2+1][j+1] = bg_sec2[i][j]

    #Build 4
    for j in xrange(bg_tert.PaletteSize()):
      self.bgPals[7*2][j+1] = bg_tert[0][j]


  
  def buildBgTiles(self, bgMain, bgAux, Count):
    #Reset
    self.bgTiles = bm.BgTileSet()
    self.halfPalSelect = [0] * 0x200

    #Merge the "main" and "aux" background layers
    for i in xrange(0,Count):
      #Get the ticket
      ticket = bgMain.get(i)

      #Under certain circumstances, get an auxiliary ticket
      j = Count - i - 1
      if (j==1 or j==2 or j==3 or j==4):
        if bgAux.get(i-3):
          ticket = bgAux.get(i-3)

      #Now use this ticket
      tileSet = ticket.Redeem()
      for t in xrange(tileSet.NumTiles()):
        #Update bgTiles
        offset = t + 0x40*i
        self.bgTiles[offset] = tileSet.get(t)

        #Update halfPalSelect
        if (j==2 or j==3 or j==4 or j==7):
          self.halfPalSelect[offset] = 1
        else:
          self.halfPalSelect[offset] = 0


    #Add in animated tiles
    #(todo)



