//Compile with:
//   g++ sample.cpp -o sample  
//Then run:
//   ./sample

#include <string>
#include <vector>
#include <iostream>
#include <boost/function.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>

using std::cout;
using std::endl;
using std::string;
using std::vector;
using boost::function;
using boost::shared_ptr;
using boost::weak_ptr;


//Use a typedef for our callback functions
class Subscribable;
typedef   function<void (Subscribable* item)>   Callback;


//Base class. 
//Allows subscription (I don't think this is necessary, but here's how you might do it)
class Subscribable {
public:
  explicit Subscribable(Callback on_delete) : on_delete(on_delete) {}
  virtual ~Subscribable() {
    //Inform anyone listening that we're about to be deleted.
    on_delete(this);
  }

private:
  Callback on_delete;
};


//Here's a subclass. I'm accomplishing this with inheritance, but there's other ways to do this
// if you don't like having a common base class.
class DialogueEntry : public Subscribable {
public:
  DialogueEntry(Callback on_delete, const string& text) : Subscribable(on_delete), text(text) {}
  const std::string& getText() { return text; }

private:
  string text;
};


//Print the memory address of the item deleted.
//If it's a message, also print its text.
void print_out_item(Subscribable* item) {
  cout <<"Deleted: " <<item;

  //Please ignore this abuse of OOP for the moment:
  DialogueEntry* msg = dynamic_cast<DialogueEntry*>(item);
  if (msg) {
    cout <<" (" <<msg->getText() <<")";
  }

  cout <<endl;
}

//NOTE: There's a weird bug in GCC that occurs if you don't save the functional
// pointer to a temporary. 
Callback callback_function = &print_out_item;



//Here we go!
string hrule(20, '-');
int main(int argc, char *argv[])
{
  { //First, a very simple demonstration of our RAII class. 
    Subscribable item1(callback_function);
    std::cout <<"Created: " <<&item1 <<endl;
    DialogueEntry item2(callback_function, "Zelda is your ...");
    std::cout <<"Created: " <<&item2 <<" (" <<item2.getText() <<")" <<endl;
  }
  cout <<"Out of scope." <<endl;
  std::cout <<hrule <<endl;

  { //Now a very simple demonstration of shared_ptr.
    shared_ptr<Subscribable> item(new Subscribable(callback_function));
    std::cout <<"Created: " <<&(*item) <<endl;

    //You can put shared_ptr into a vector just fine
    {
      vector< shared_ptr<Subscribable> > uselessCopies;
      cout <<"Vector created." <<endl;
      for (int i=0; i<10; i++) {
        uselessCopies.push_back(item);
      }
      cout <<"Vector filled with 10 copies of the pointer." <<endl;
    }
    cout <<"Vector out of scope." <<endl; //The useless copies have been deleted, but 1 reference remains.
  } //Here is where item should be destroyed
  cout <<"Out of scope." <<endl;
  std::cout <<hrule <<endl;

  //Now, the real fun begins. Assume that the main GUI loads a text message from
  // the ROM, and makes a new DialogueEntry object. The main GUI will perform all 
  // of its interactions through a "shared_ptr", which maintains a reference count.
  shared_ptr<DialogueEntry>* myMsg = 0; 
  {
    //I am using a pointer to a shared pointer so that I can explicitly call "delete" on it. 
    //Generally speaking, you should use shared_ptr by value, like so:
    //   shared_ptr<DialogueEntry> myMsg(new DialogueEntry(/*whatever*/));
    DialogueEntry* realMsg = new DialogueEntry(callback_function, "It's dangerous to go alone.");
    myMsg = new shared_ptr<DialogueEntry>(realMsg);
    std::cout <<"Authoritative object created: " <<realMsg <<endl;
  }

  //Our variable myMsg now holds the only reference to our original message. 
  //Now let's assume several other GUI windows want a copy. We give them a "weak" pointer which 
  //does not increment the reference count.
  {
    weak_ptr<DialogueEntry> gui2MsgCopy(*myMsg);   //Reference count is still 1
    std::cout <<"Weak copy created" <<endl;
    weak_ptr<DialogueEntry> gui3MsgCopy(*myMsg);   //Reference count is still 1
    std::cout <<"Weak copy created" <<endl;
    {
      //This goes immediately out of scope, but has no effect (it's just a weak reference)
      weak_ptr<DialogueEntry> gui4MsgCopy(*myMsg);  //Reference count is still 1
      std::cout <<"Weak copy created" <<endl;
    } //Reference count is still 1

    //Now, gui 2 wants to access the value. Use the "lock" function, which creates a 
    // temporary shared_ptr (this ensures the object is not deleted halfway through its use).
    //Do this inside the if conditional so that it goes out of scope after the "if" body.
    if (shared_ptr<DialogueEntry> msg = gui2MsgCopy.lock()) {
      std::cout <<"GUI 2 read the message: " <<msg->getText() <<endl;       //We end up here.
    } else {
      std::cout <<"GUI 2 couldn't read the object; it's been deleted.\n";   //...not here.
    }

    //Now, let's assume the GUI has decided that it must delete the message. It can accomplish this
    // by deleting the shared_ptr. At this point, the message will be deleted UNLESS there is a lock() 
    // out by another GUI. Since those locks are supposed to be used only temporarily, the main GUI
    // can just go on and pretend that the message has been deleted (it will be as soon as all the other locks
    // come back). There is no way this will corrupt memory (exceptions complicate this slightly.)
    std::cout <<"Am about to delete the authoritative pointer." <<endl;
    delete myMsg;  //The on_delete function should be called here, or soon after if a lock is outstanding.
    std::cout <<"Just deleted the authoritative pointer." <<endl;


    //So, what if GUI 3 wants to access the object? Well, try locking it...
    if (shared_ptr<DialogueEntry> msg = gui3MsgCopy.lock()) {
      std::cout <<"GUI 3 read the message: " <<msg->getText() <<endl;       //This time, we don't end up here
    } else {
      std::cout <<"GUI 3 couldn't read the object; it's been deleted.\n";   //...but here.
    }

    //End of scope. weak references will be deleted, but who cares?
  }
  cout <<"Out of scope. Todo: buy colgate." <<endl;


  //By the way, I'm not 100% sure why the on_delete function doesn't understand that this is a DialogueEntry. 
  //But it's irrelevant for the point I'm trying to make; the function definitely gets called, and you can always
  //fiddle with the dynamic casting if it's not working. (I'd guess that shared_ptr<DialogueEntry> casts it to a 
  //value type and then some weird truncation nonsense happens.)

}






















