#include "MainWindow.hpp"
#include "ui_main.h"

#include <fstream>
#include <sstream>
#include <vector>
#include <map>
#include <QtCore>
#include <QFileDialog>
//#include <boost/thread.hpp>
#include <jsoncpp/json/json.h>

#include "expandinfo.h"
#include "loaders.h"
#include "zeldagame.h"
#include "graphics.h"

//Keep track of all this Zelda stuff.
struct bm_stuff {
	std::string fn;
	ExpandInfo* ei;
	Loaders::RomLoader* rl;
	ZeldaGame* game;
	QThread* currThread;
} my_stuff;

//Keep track of annotated stuff
struct annot_data {
    bool empty;

    std::map<std::string, uint32_t> bgcolors;
    std::map<uint32_t, std::string> areaNames;
    std::map<uint32_t, std::string> areaBgs;

    //Later
    //std::vector< std::vector<uint32_t> > lightConnect;
    //std::vector< std::vector<uint32_t> > darkConnect;
} my_annot_data;

//I love these templates so much:
template <typename T>
void safe_delete_item(T& item) {}
template <typename T>
void safe_delete_item(T*& item) {
	if (item) {
		delete item;
		item = nullptr;
	}
}

template <size_t N, size_t M>
void PopulateList(typename Graphics::Group<N, M>::ShowDeque& show, QComboBox* combo) {
	for (auto item : show) {
		combo->addItem(QString("Show ") + QString::number(combo->count()));
	}
}

template <size_t N, size_t M>
void PopulateList2(typename Graphics::Group<N, M>::ShowDeque& show, int showID, QComboBox* combo) {
	for (auto item : show) {
		if (showID-- == 0) {
			for (int i=0; i<item.Get().PaletteCount(); i++) {
				combo->addItem(QString("Palette ") + QString::number(combo->count()));
			}
			break;
		}
	}
}

template <size_t N, size_t M>
void PopulateColors(typename Graphics::Group<N, M>::ShowDeque& show, int showID, int groupID, ColorLister* colorBoxes) {
	for (auto item : show) {
		if (showID-- == 0) {
                        auto colors = item.Get()[groupID];
			for (int currIndex=0; currIndex<colors.Length(); currIndex++) {
                                //if (currIndex>=colorBtns.size()) { break; }
				Graphics::Color c = colors[currIndex];
				uint32_t rgba = uint32_t(c);

                                colorBoxes->addColor(QColor::fromRgba(rgba));

                                /*QPalette p = colorBtns.at(currIndex)->palette();
				p.setBrush(QPalette::Button, QColor::fromRgba(rgba));
                                colorBtns.at(currIndex)->setPalette(p);*/
			}
			break;
		}
	}
}



MainWindow::MainWindow(QWidget* parent) : QMainWindow(parent), ui(new Ui::MainWindow)
{
	//Connect our UI to this window:
	ui->setupUi(this);

	//Init our struct
	my_stuff.ei = nullptr;
	my_stuff.rl = nullptr;
	my_stuff.game = nullptr;
	my_stuff.currThread = nullptr;
    my_annot_data.empty = true;

	//Typically this is where you would set up signals and slots.
	QObject::connect(ui->palMetaGroup, SIGNAL(currentIndexChanged(int)), this, SLOT(setGroupFromMetaGroup(int)));
	QObject::connect(ui->palGroup, SIGNAL(currentIndexChanged(int)), this, SLOT(setGroupFromGroup(int)));
	QObject::connect(ui->palCount, SIGNAL(currentIndexChanged(int)), this, SLOT(setPalFromCount(int)));

    QObject::connect(ui->overworldCombo, SIGNAL(currentIndexChanged(int)), this, SLOT(setMapFromId(int)));

}


MainWindow::~MainWindow()
{
	//Is this safe?
    //delete ui;
}


void MainWindow::on_actionNew_triggered()
{

}


void MainWindow::on_actionOpen_triggered()
{
	//Prompt for a file to open.
	QString fileName = QFileDialog::getOpenFileName(this, "Open Zelda3 ROM", "zelda3.sfc", "Super Famicon (*.sfc)");
	if (fileName.isEmpty()) { return; }

	//Join the old thread if we're still waiting for it
	if (my_stuff.currThread) {
		my_stuff.currThread->blockSignals(true); //Don't respond to completion.
		my_stuff.currThread->wait();
	}

	//Clear some old data structures (may need to unhook them later).
	safe_delete_item(my_stuff.ei);
	safe_delete_item(my_stuff.rl);
	safe_delete_item(my_stuff.game);
	safe_delete_item(my_stuff.currThread);

	//Make a new thread to process loading.
	my_stuff.fn = fileName.toStdString();
	my_stuff.currThread = new LoadRomThread(my_stuff);
	QObject::connect(my_stuff.currThread, SIGNAL(finished()), this, SLOT(romLoaded()));
	my_stuff.currThread->start();
}


void LoadRomThread::run()
{
	//Fire up the ol' BM library.
	std::ifstream tempFile(stuff->fn); //TODO: It is NOT clear that this reference is never saved.
	stuff->ei = new ExpandInfo(tempFile);
	stuff->game = new ZeldaGame();
	stuff->rl = new Loaders::RomLoader(*stuff->game, *stuff->ei);
	stuff->rl->LoadGame();
}



namespace {
unsigned int str_parse_hex(const std::string& str) {
    unsigned int res;
    std::stringstream ss;
    ss <<std::hex <<str;
    ss >>res;
    return res;
}
unsigned int str_parse_dec(const std::string& str) {
    unsigned int res;
    std::stringstream ss;
    ss <<str;
    ss >>res;
    return res;
}

void parseAnnotData(Json::Value& root, annot_data& res)
{
    //Reset
    res.areaBgs.clear();
    res.areaNames.clear();
    res.bgcolors.clear();
    res.empty = false;

    //Load area backgrounds
    Json::Value bgcolors = root["bgcolors"];
    Json::Value::Members bgColorNames = bgcolors.getMemberNames();
    for (Json::Value::Members::const_iterator it=bgColorNames.begin(); it!=bgColorNames.end(); it++) {
        std::string bgColorVal = bgcolors[*it].asString();
        if (bgColorVal.size()>2 && bgColorVal[0]=='0' && bgColorVal[1]=='x') {
            res.bgcolors[*it] = str_parse_hex(bgColorVal.substr(2));
        }
    }

    //Load "overworld" mappings.
    Json::Value overworlds = root["overworld"];
    Json::Value::Members overworldNames = overworlds.getMemberNames();
    for (Json::Value::Members::const_iterator it=overworldNames.begin(); it!=overworldNames.end(); it++) {
        uint32_t overId = str_parse_dec(*it);
        Json::Value overData = overworlds[*it];
        if (overData.isMember("name")) {
            res.areaNames[overId] = overData["name"].asString();
        }
        if (overData.isMember("bgcolor")) {
            res.areaBgs[overId] = overData["bgcolor"].asString();
        }
    }

    //Load connectivities
    //TODO
}
} //End anon namespace


void MainWindow::romLoaded()
{
	std::cout <<"Rom loaded" <<std::endl;

    //Temporarily cancel events.
    BlockAllQtSignals block({ui->palMetaGroup, ui->palGroup, ui->palCount, ui->overworldCombo});

    //TODO: Do this locally (how to include local files when using Qt?)
    const std::string path = "/home/sethhetu/mathonnapkins/Black_Magic/ZQGuiTest/data/overworld.json.txt";
    std::ifstream file(path);

    //Pretend we are loading annotation data from some local source.
    Json::Value root;   // will contains the root value after parsing.
    Json::Reader reader;
    if (!reader.parse(file, root)) {
        std::cout <<"Couldn't find annotation data.\n  Not all editor functions will work to their full potential." <<std::endl;
    } else {
        parseAnnotData(root, my_annot_data);
    }


	//Populate our first combo box.
	ui->palMetaGroup->clear();
	ui->palMetaGroup->addItem("H.U.D.");
	ui->palMetaGroup->addItem("Sword");
	ui->palMetaGroup->addItem("Shield");
	ui->palMetaGroup->addItem("Armor");
	ui->palMetaGroup->addItem("Overworld/Primary");
	ui->palMetaGroup->addItem("Overworld/Secondary");
	ui->palMetaGroup->addItem("Overworld/Tertiary");
	ui->palMetaGroup->addItem("Dungeon/Primary");
	ui->palMetaGroup->addItem("Sprite/Primary");
	ui->palMetaGroup->addItem("Sprite/Secondary");
	ui->palMetaGroup->addItem("Sprite/Tertiary");
	ui->palMetaGroup->addItem("Sprite/Misc");
	ui->palMetaGroup->addItem("PalaceMap/Bg");
	ui->palMetaGroup->addItem("PalaceMap/Sprite");
	ui->palMetaGroup->setCurrentIndex(-1);

	//Clear/disable our other combo boxes.
	ui->palGroup->clear();
	ui->palCount->clear();

    //Populate our overworld combo boxes.
    ui->overworldCombo->clear();
    for (int i=0; i<my_stuff.game->areas.size(); i++) {
        std::stringstream label;
        label <<i <<" : ";
        if (my_annot_data.areaNames.count(i)>0) {
            label <<my_annot_data.areaNames[i];
        }

        ui->overworldCombo->addItem(label.str().c_str());
        ui->overworldCombo->setCurrentIndex(-1);
    }
}


void MainWindow::setGroupFromMetaGroup(int index)
{
	if (!my_stuff.game || index==-1) { return; }
	std::cout <<"Loading MetaGroup: " <<index <<std::endl;

	//Clear existing items
	ui->palGroup->clear();

	//Populate:
	switch (index) {
		case 0:
			PopulateList<4,8>(my_stuff.game->palData.hud, ui->palGroup); break;
		case 1:
			PopulateList<3,1>(my_stuff.game->palData.sword, ui->palGroup); break;
		case 2:
			PopulateList<4,1>(my_stuff.game->palData.shield, ui->palGroup); break;
		case 3:
			PopulateList<15,1>(my_stuff.game->palData.armor, ui->palGroup); break;
		case 4:
			PopulateList<7,5>(my_stuff.game->palData.overworld.primary, ui->palGroup); break;
		case 5:
			PopulateList<7,3>(my_stuff.game->palData.overworld.secondary, ui->palGroup); break;
		case 6:
			PopulateList<7,1>(my_stuff.game->palData.overworld.tertiary, ui->palGroup); break;
		case 7:
			PopulateList<15,6>(my_stuff.game->palData.dungeon.primary, ui->palGroup); break;
		case 8:
			PopulateList<15,4>(my_stuff.game->palData.sprite.primary, ui->palGroup); break;
		case 9:
			PopulateList<7,1>(my_stuff.game->palData.sprite.secondary, ui->palGroup); break;
		case 10:
			PopulateList<7,1>(my_stuff.game->palData.sprite.tertiary, ui->palGroup); break;
		case 11:
			PopulateList<7,1>(my_stuff.game->palData.sprite.misc, ui->palGroup); break;
		case 12:
			PopulateList<16,6>(my_stuff.game->palData.palaceMap.bg, ui->palGroup); break;
		case 13:
			PopulateList<7,3>(my_stuff.game->palData.palaceMap.sprite, ui->palGroup); break;
	}
}


void MainWindow::setGroupFromGroup(int index)
{
	if (!my_stuff.game || index==-1 || ui->palMetaGroup->currentIndex()==-1) { return; }
	std::cout <<"Loading Group: " <<index <<std::endl;

	//Clear existing items
	ui->palCount->clear();

	//Populate:
	switch (ui->palMetaGroup->currentIndex()) {
		case 0:
			PopulateList2<4,8>(my_stuff.game->palData.hud, index, ui->palCount); break;
		case 1:
			PopulateList2<3,1>(my_stuff.game->palData.sword, index, ui->palCount); break;
		case 2:
			PopulateList2<4,1>(my_stuff.game->palData.shield, index, ui->palCount); break;
		case 3:
			PopulateList2<15,1>(my_stuff.game->palData.armor, index, ui->palCount); break;
		case 4:
			PopulateList2<7,5>(my_stuff.game->palData.overworld.primary, index, ui->palCount); break;
		case 5:
			PopulateList2<7,3>(my_stuff.game->palData.overworld.secondary, index, ui->palCount); break;
		case 6:
			PopulateList2<7,1>(my_stuff.game->palData.overworld.tertiary, index, ui->palCount); break;
		case 7:
			PopulateList2<15,6>(my_stuff.game->palData.dungeon.primary, index, ui->palCount); break;
		case 8:
			PopulateList2<15,4>(my_stuff.game->palData.sprite.primary, index, ui->palCount); break;
		case 9:
			PopulateList2<7,1>(my_stuff.game->palData.sprite.secondary, index, ui->palCount); break;
		case 10:
			PopulateList2<7,1>(my_stuff.game->palData.sprite.tertiary, index, ui->palCount); break;
		case 11:
			PopulateList2<7,1>(my_stuff.game->palData.sprite.misc, index, ui->palCount); break;
		case 12:
			PopulateList2<16,6>(my_stuff.game->palData.palaceMap.bg, index, ui->palCount); break;
		case 13:
			PopulateList2<7,3>(my_stuff.game->palData.palaceMap.sprite, index, ui->palCount); break;
	}
}



void MainWindow::setPalFromCount(int index)
{
	if (!my_stuff.game || index==-1 || ui->palMetaGroup->currentIndex()==-1 || ui->palGroup->currentIndex()==-1) { return; }
	std::cout <<"Loading Count: " <<index <<std::endl;

	//Build our lookup
        /*if (colors.empty()) {
		colors.push_back(ui->color1);
		colors.push_back(ui->color2);
		colors.push_back(ui->color3);
		colors.push_back(ui->color4);
		colors.push_back(ui->color5);
		colors.push_back(ui->color6);
		colors.push_back(ui->color7);
		colors.push_back(ui->color8);
		colors.push_back(ui->color9);
		colors.push_back(ui->color10);
        }*/

	//Clear existing items
        ui->colorBoxes->clearColors();
        /*for (auto button : colors) {
		button->setFlat(true);
		button->setAutoFillBackground(true);
		QPalette p = button->palette();
		p.setBrush(QPalette::Button, Qt::white);
		button->setPalette(p);
        }*/

	//Populate:
	switch (ui->palMetaGroup->currentIndex()) {
		case 0:
                        PopulateColors<4,8>(my_stuff.game->palData.hud, ui->palGroup->currentIndex(), index, ui->colorBoxes); break;
		case 1:
                        PopulateColors<3,1>(my_stuff.game->palData.sword, ui->palGroup->currentIndex(), index, ui->colorBoxes); break;
		case 2:
                        PopulateColors<4,1>(my_stuff.game->palData.shield, ui->palGroup->currentIndex(), index, ui->colorBoxes); break;
		case 3:
                        PopulateColors<15,1>(my_stuff.game->palData.armor, ui->palGroup->currentIndex(), index, ui->colorBoxes); break;
		case 4:
                        PopulateColors<7,5>(my_stuff.game->palData.overworld.primary, ui->palGroup->currentIndex(), index, ui->colorBoxes); break;
		case 5:
                        PopulateColors<7,3>(my_stuff.game->palData.overworld.secondary, ui->palGroup->currentIndex(), index, ui->colorBoxes); break;
		case 6:
                        PopulateColors<7,1>(my_stuff.game->palData.overworld.tertiary, ui->palGroup->currentIndex(), index, ui->colorBoxes); break;
		case 7:
                        PopulateColors<15,6>(my_stuff.game->palData.dungeon.primary, ui->palGroup->currentIndex(), index, ui->colorBoxes); break;
		case 8:
                        PopulateColors<15,4>(my_stuff.game->palData.sprite.primary, ui->palGroup->currentIndex(), index, ui->colorBoxes); break;
		case 9:
                        PopulateColors<7,1>(my_stuff.game->palData.sprite.secondary, ui->palGroup->currentIndex(), index, ui->colorBoxes); break;
		case 10:
                        PopulateColors<7,1>(my_stuff.game->palData.sprite.tertiary, ui->palGroup->currentIndex(), index, ui->colorBoxes); break;
		case 11:
                        PopulateColors<7,1>(my_stuff.game->palData.sprite.misc, ui->palGroup->currentIndex(), index, ui->colorBoxes); break;
		case 12:
                        PopulateColors<16,6>(my_stuff.game->palData.palaceMap.bg, ui->palGroup->currentIndex(), index, ui->colorBoxes); break;
		case 13:
                        PopulateColors<7,3>(my_stuff.game->palData.palaceMap.sprite, ui->palGroup->currentIndex(), index, ui->colorBoxes); break;
	}
}



void MainWindow::setMapFromId(int index)
{
    if (!my_stuff.game || index==-1) { return; }

    //TODO
}








