#include "FixedGridGraphicsScene.hpp"

#include <QPainter>

#include <limits>
#include <cmath>

using std::pair;

FixedGridGraphicsScene::FixedGridGraphicsScene(int tileSize, int bufferSize, int numLayers, pair<int,int> initArraySize, QObject* parent)
    : QGraphicsScene(parent), tileSize(tileSize), bufferSize(bufferSize)
{
    //3D arrays ftw. layers[layerID][yPos][xPos]
    for (size_t l=0; l<numLayers; l++) {
        layers.push_back(std::vector< std::vector<QPixmap*> >());
        for (size_t y=0; y<initArraySize.second; y++) {
            layers.back().push_back(std::vector<QPixmap*>(initArraySize.first, nullptr));
        }

        layerVisibility.push_back(true);
    }

    visibleTiles = SimpleQuad(0, 0, 0, 0);
}


FixedGridGraphicsScene::SimpleQuad FixedGridGraphicsScene::getPixelRectFromTiles(int tileX, int tileY)
{
    int tsP1 = tileSize + bufferSize;
    return SimpleQuad(tileX*tsP1+bufferSize, tileY*tsP1+bufferSize, tileSize, tileSize);
}


void FixedGridGraphicsScene::setVisibleTiles(int offsetX, int offsetY, int widthInTiles, int heightInTiles)
{
    visibleTiles = SimpleQuad(offsetX, offsetY, widthInTiles, heightInTiles);
    SimpleQuad innerLimits = getPixelRectFromTiles(offsetX, offsetY); //A little messy
    SimpleQuad outerLimits = getPixelRectFromTiles(widthInTiles, heightInTiles);

    //Dispatch to parent
    QGraphicsScene::setSceneRect(innerLimits.x-bufferSize, innerLimits.y-bufferSize, outerLimits.x, outerLimits.y);
}

void FixedGridGraphicsScene::setLayerVisibility(int layerID, bool isVisible)
{
    if (layerID<0 or layerID>=layers.size()) {
      throw std::runtime_error("Layer ID out of range in addItem()");
    }
    layerVisibility[layerID] = isVisible;
}

void FixedGridGraphicsScene::setTile(QPixmap* item, int tileX, int tileY, int layerID)
{
    tileAccessHelper(layerID, tileX, tileY, item, true);
    layers[layerID][tileY][tileX] = item;
}



namespace {
//x,y = indices of items we want to add.
//      E.g., layer[y][x] = item expand with None.
//Note: Let's just double it.
void expandLayer(std::vector< std::vector<QPixmap*> >& layer, int x, int y)
{
    //Expand X for existing rows
    if (x>=layer[0].size()) {
        for (auto& row : layer) {
            row.resize(2*row.size(), nullptr);
        }
    }

    //Expand Y, accounting for the already-expanded X
    if (y>=layer.size()) {
        layer.resize(2*layer.size(), std::vector<QPixmap*>(layer[0].size(), nullptr));
    }
}
} //End anon namespace


void FixedGridGraphicsScene::tileAccessHelper(int layerID, int tileX, int tileY, QPixmap* item, bool expand)
{
    //Sanity checks: Item
    if (item && (item->width()!=tileSize || item->height()!=tileSize)) {
        throw std::runtime_error("addItem() must fit with the requested tile size");
    }

    //Sanity check: layers
    if (layerID<0 || layerID>=layers.size()) {
        throw std::runtime_error("Layer ID out of range in addItem()");
    }

    //Sanity check: x/y
    if (!expand && (tileY>=layers[layerID].size() || tileX>=layers[layerID][0].size())) {
        throw std::runtime_error("Tile ID out of range in addItem(), and expand is set to False");
    }

    //Expand *all* layers in parallel as required. (Layers never decrease in size).
    if (expand) {
        for (auto& layer : layers) {
            expandLayer(layer, tileX, tileY);
        }
    }
}


QPixmap* FixedGridGraphicsScene::getTile(int tileX, int tileY, int layerID, bool expandTilesAsNeeded)
{
    tileAccessHelper(layerID, tileX, tileY, nullptr, expandTilesAsNeeded);
    if (!layers[layerID][tileY][tileX]) {
      layers[layerID][tileY][tileX] = new QPixmap(tileSize, tileSize);
    }
    return layers[layerID][tileY][tileX];
}



QPixmap* FixedGridGraphicsScene::newTile(int tileX, int tileY, int layerID)
{
    tileAccessHelper(layerID, tileX, tileY, nullptr, true);
    layers[layerID][tileY][tileX] = new QPixmap(tileSize, tileSize);
    return layers[layerID][tileY][tileX];
}


void FixedGridGraphicsScene::clearTile(int tileX, int tileY, int layerID)
{
    tileAccessHelper(layerID, tileX, tileY, nullptr, false);
    delete layers[layerID][tileY][tileX];
    layers[layerID][tileY][tileX] = nullptr;
}


std::pair<int, int> FixedGridGraphicsScene::getTileAxisFromPixel(int pixMin, int pixMax)
{
    pixMin -= bufferSize;
    pixMax -= bufferSize;
    int tsP1 = tileSize + bufferSize;
    std::pair<int, int> result;
    result.first = pixMin / tsP1;
    result.second = (int)ceil(pixMax / float(tsP1));
    return result;
}

FixedGridGraphicsScene::SimpleQuad FixedGridGraphicsScene::getTileRectFromPixels(QRectF pixelRect)
{
    //Each "axis" is returned as (min, max) including overflow
    std::pair<int, int> xAxis = getTileAxisFromPixel(pixelRect.x(), pixelRect.x()+pixelRect.width());
    std::pair<int, int> yAxis = getTileAxisFromPixel(pixelRect.y(), pixelRect.y()+pixelRect.height());

    //Bound by the actual number of tiles (and zero!)
    return SimpleQuad(std::max(0, xAxis.first), std::max(0, yAxis.first),
                      std::min(xAxis.second, -1+(int)layers[0][0].size()), std::min(yAxis.first, -1+(int)layers[0].size()));
}




void FixedGridGraphicsScene::drawItemsPrivate(QPainter* painter, int rectAdjust, const QTransform* viewTransform, const QRegion* exposedRegion, QWidget* viewport)
{
    //Create a rectangle to represent the region which is exposed.
    QRectF exposedSceneRect;
    if (exposedRegion) { //Likely always true
        //Take an educated guess. Apply the current view transformation if one exists.
        exposedSceneRect = exposedRegion->boundingRect().adjusted(-1, -1, 1, 1);
        if (viewTransform) {
          QTransform inverted = viewTransform->inverted();  //Used to be: (Transform, bool)
          if (!inverted.isIdentity()) {  //Not 100% sure this is right...
              //inverted = viewTransform.inverted()[0];
              exposedSceneRect = inverted.mapRect(exposedSceneRect);
          }
        }
    }

    //Finally, retrieve and paint all items which are visible. This is slightly complicated by the fact that
    // a world transform might be active (but that should be fairly transparent)
    //   1) exposedSceneRect contains the "scene" coordinates that need repainting, +/- 1 pixel
    //   2) We then compute the "minTile" and "maxTile" for each component. Make sure to include tiles partway on-screen.
    //   3) Then it's just a matter of iterating through these and printing them (layer by layer)
    SimpleQuad tileRect = getTileRectFromPixels(exposedSceneRect);  //Note: it's a "rectangle" in the Win32 sense, not the width/height sense
    tileRect.width += 1;
    tileRect.height += 1;

    //Bound the ranges if requested
    //TODO: Avoid duplicating code...
    //TODO: Wow this is a mess.
    if (!visibleTiles.isNull()) {
        tileRect.width = std::max(std::min(tileRect.width, visibleTiles.x+visibleTiles.width), visibleTiles.x);
        tileRect.height = std::max(std::min(tileRect.height, visibleTiles.y+visibleTiles.height), visibleTiles.y);
    }

    //Loop
    for (int l=0; l<layers.size(); l++) {
        if (!layerVisibility[l]) { continue; }
        for (int y=0; y<layers[l].size(); y++) {
            for (int x=0; y<layers[l][y].size(); x++) {
                QPixmap* tile = layers[l][y][x];
                if (tile) {
                    tileRect = getPixelRectFromTiles(x,y); //Note: Not the fastest, but oh well
                    painter->drawPixmap(tileRect.x, tileRect.y, *tile);
                }
            }
        }
    }

}







