#pragma once

#include <QGraphicsScene>

#include <stdexcept>
#include <vector>
#include <map>


/**
 * A GraphicsScene which is made for storing only tiles.
 *
 * \param tileSize  the size of each QPixmap tile
 * \param bufferSize  the size of each tile's border
 * \param numLayers  the number of tile layers
 * \param initArraySize  the initial size of the array of tiles
 */
class FixedGridGraphicsScene : public QGraphicsScene {
public:
    FixedGridGraphicsScene(int tileSize, int bufferSize, int numLayers=1, std::pair<int,int> initArraySize=std::pair<int,int>(32,32), QObject* parent=nullptr);

    //Normal functionality
    void setVisibleTiles(int offsetX, int offsetY, int widthInTiles, int heightInTiles);
    void setLayerVisibility(int layerID, bool isVisible);

    ///Set the tile at the given location. (Allows shared tiles)
    ///This expands the list of tiles as needed.
    void setTile(QPixmap* item, int tileX, int tileY, int layerID);

    ///Retrieve the tile at the given location. If you modify this tile, be SURE you're not modifying a shared tile.
    //  This class will not track it for you.
    //Creates a new tile if the requested tile does not exist.
    //TODO: If we store pairs of tiles (Ref, Real), then we can use "get" to _always_ return the Real version. Have to look into this.
    QPixmap* getTile(int tileX, int tileY, int layerID, bool expandTilesAsNeeded);

    ///Create a new tile at the given location and return it. (For non-shared tiles)
    //This expands the list of tiles as needed.
    QPixmap* newTile(int tileX, int tileY, int layerID);

    ///Clear whichever tile was at the given location.
    ///Does NOT expand the list of tiles
    void clearTile(int tileX, int tileY, int layerID);

    //Kill some QGraphicsScene-related functionality.
    QGraphicsItemGroup * createItemGroup ( const QList<QGraphicsItem *> & items ) { throw std::runtime_error("createItemGroup() not supported. Use layers instead"); }
    void setSceneRect ( const QRectF & rect ) { throw std::runtime_error("setSceneRect() not supported. Use setVisibleTiles() instead"); }
    void setSceneRect ( qreal x, qreal y, qreal w, qreal h ) { throw std::runtime_error("setSceneRect() not supported. Use setVisibleTiles() instead"); }
    void addItem ( QGraphicsItem * item ) { throw std::runtime_error("addItem() not supported. Use setTile()/newTile() instead"); }

    //Simple getter/setters
    int getTileSize() { return tileSize; }

    //Semi-private item drawing code. viewTransform may be null
    void drawItemsPrivate(QPainter* painter, int rectAdjust, const QTransform* viewTransform, const QRegion* exposedRegion, QWidget* viewport);

private:
    //Helper class
    struct SimpleQuad {
        SimpleQuad(int x=0, int y=0, int width=0, int height=0) : x(x), y(y), width(width), height(height) {}
        bool isNull(){ return width==0 || height==0; }
        int x, y, width, height;
    };

    //Helper functions
    SimpleQuad getPixelRectFromTiles(int tileX, int tileY);

    ///Internal helper function: perform expansion and check the tile type
    void tileAccessHelper(int layerID, int tileX, int tileY, QPixmap* item=nullptr, bool expand=false);

    ///Math is escaping me at the moment
    std::pair<int, int> getTileAxisFromPixel(int pixMin, int pixMax);
    SimpleQuad getTileRectFromPixels(QRectF pixelRect);



    //Internal data
    int tileSize;
    int bufferSize;
    std::vector< std::vector< std::vector<QPixmap*> > > layers;
    std::vector<bool> layerVisibility;
    SimpleQuad visibleTiles;
};




