#include "MapView.h"

#include <sstream>
#include <array>
#include <QPoint>
#include <QKeyEvent>

#include "controls/FixedGridGraphicsScene.hpp"


std::pair<int, int> MapView::BorderSize(8, 7);

QPixmap* MapView::QuestionTile(nullptr);

std::map<std::string, QImage*> MapView::Map8TileCache;

std::vector<int> MapView::HalfPalSelect;

Graphics::TileSet<Graphics::Pixel3bpp, 0x200> MapView::BgTiles;

Graphics::Group<8, 16> MapView::BgPals;

Overworld::Area* MapView::CurrArea;

uint32_t MapView::CurrAreaID;

std::vector< std::vector<uint32_t> > MapView::AreaConnections;

std::map<uint32_t, QPoint> MapView::AreaLookup;


QImage* MapView::GetMap8Tile(ZeldaGame* game, Graphics::TileEntry* tileEntry, uint32_t bgColor)
{
    //Some indexing properties
    auto tChr = tileEntry->GetChr();
    auto tPal = tileEntry->m_pal;
    auto half = HalfPalSelect[tChr];  //TODO: Need to port this (and a whole bunch of other stuff) into our model.

    //Make a key
    std::stringstream key_ss;
    key_ss <<tChr <<":" <<tPal <<":" <<tileEntry->m_hFlip <<":" <<tileEntry->m_vFlip;
    std::string key = key_ss.str();

    //Add to the cache if needed
    if (Map8TileCache.count(key)==0) {
        //Build up a ColorTile for reference, then fill all 8x8 pixels
        QImage* res = new QImage(8, 8, QImage::Format_RGB32);
        res->fill(bgColor);
        Graphics::ColorTile cTile(BgTiles(tChr), BgPals[2*tPal + half], 0);
        for (int y=0; y<8; y++) {
            for (int x=0; x<8; x++) {
                //Qt is pretty bad about converting strings, so we'll have to cache the actual QImage here.
                auto color = (uint32_t)cTile(y, x, tileEntry->m_hFlip, tileEntry->m_vFlip);
                if (color&0xFF000000) { //Opaque
                    res->setPixel(x, y, color);
                }
            }
        }

        //Update the cache
        if (Map8TileCache.size()>1000) {
            //Normally we'd remove a random element from the map, but for now let's just take the first one.
            Map8TileCache.erase(Map8TileCache.begin());
            //self.tileCache.pop(random.choice(self.tileCache.keys()))
        }
        Map8TileCache[key] = res;
    }

    //Done
    return Map8TileCache[key];
}


void MapView::BuildBgTiles(Graphics::BgTileMain bgMain, Graphics::BgTileAux bgAux, int count)
{
    //Reset
    BgTiles = Graphics::TileSet<Graphics::Pixel3bpp, 0x200>();
    HalfPalSelect.resize(0x200, 0);

    //Merge the "main" and "aux" background layers
    for (int i=0; i<count; i++) {
        //Get the ticket
        Containers::Ticket<Graphics::TileSet3bpp> ticket = bgMain[i];

        //Under certain circumstances, get an auxiliary ticket
        int j = count - i - 1;
        if (j==1 || j==2 || j==3 || j==4) {
            //if (bgAux[i-3]) {
            if (i-3>0) { //TODO: Not sure if this is right!
                ticket = bgAux[i-3];
            }
        }

        //Now use this ticket
        auto tileSet = ticket.Redeem();
        for (int t=0; t<tileSet.NumTiles(); t++) {
            //Update bgTiles
            int offset = t + 0x40*i;
            BgTiles(offset) = tileSet(t);

            //Update halfPalSelect
            if (j==2 || j==3 || j==4 || j==7) {
                HalfPalSelect[offset] = 1;
            } else {
                HalfPalSelect[offset] = 0;
            }
        }
    }

    //Add in animated tiles
    //(todo)
}



void MapView::BuildBgPalettes(Graphics::Group<7, 5> bg_prim, Graphics::Group<7, 3> bg_sec1, Graphics::Group<7, 3>* bg_sec2, Graphics::Group<7, 1> bg_tert)
{
    //Reset
    BgPals = Graphics::Group<8, 16>();

    //Build 1:
    for (int i=0; i<5; i++) {
        for (int j=0; j<bg_prim.PaletteSize(); j++) {
            BgPals[(i+2)*2][j+1] = bg_prim[i][j];
        }
    }

    //Build 2:
    for (int i=0; i<3; i++) {
        for (int j=0; j<bg_sec1.PaletteSize(); j++) {
            BgPals[(i+2)*2+1][j+1] = bg_sec1[i][j];
        }
    }

    //Build 3
    if (bg_sec2) {
        for (int i=0; i<3; i++) {
            for (int j=0; j<bg_sec2->PaletteSize(); j++) {
                BgPals[(i+5)*2+1][j+1] = (*bg_sec2)[i][j];
            }
        }
    }

    //Build 4
    for (int j=0; j<bg_tert.PaletteSize(); j++) {
        BgPals[7*2][j+1] = bg_tert[0][j];
    }
}


MapView::MapView(QWidget* parent) : FixedGridGraphicsView(parent),
    game(nullptr), mapLayerVis(true), annotLayerVis(true), validMap(false), hasLoadedSideMaps(false), triCount(1)
{
    //Offset by 8x7; this allows us to draw some nice border tiles as needed.
    offset = std::make_pair(BorderSize.first, BorderSize.second);

    //Create a Scene which rendrs into a fixed grid.
    setScene(new FixedGridGraphicsScene(16, 1, 2));
    scene()->setBackgroundBrush(QBrush(Qt::black));
}


std::pair<int, int> MapView::getTranslated(const std::pair<int, int>& point)
{
    return std::make_pair(point.first+offset.first, point.second+offset.second);
}


void MapView::newMap(int mapID)
{
    validMap = true;
    hasLoadedSideMaps = false;
    rebuildScene(mapID);
}


void MapView::triCountAdd(int amount)
{
    triCount += amount;
    triCount = std::min(std::max(triCount, 1), 4); //It really looks bad after 4
    viewport()->update();
}


void MapView::renderAnnotation(Overworld::Map16Tile* tileQuad, int tileX, int tileY)
{
    int tileSize = fg_scene()->getTileSize();

    //Render our ??? tile *once*
    if (!QuestionTile) {
      QuestionTile = new QPixmap(tileSize, tileSize);
      QuestionTile->fill(QColor(0x33, 0x33, 0x33));
      QPainter gcOver(QuestionTile);
      gcOver.setFont(QFont("Arial", 8));
      gcOver.setPen(QColor(0xAA, 0xAA, 0xAA));
      gcOver.drawText(QRectF(0, 0, tileSize, tileSize), Qt::AlignCenter, "??");
      gcOver.end();
    }

    //If this tile slot is empty, just set the question tile
    if (!tileQuad) {
      fg_scene()->setTile(QuestionTile, tileX, tileY, 1);
      return;
    }

    //Otherwise, draw something (flipped tile markings, for now)
    //NOTE: There's only a few hundred combinations; we could cache them if we wanted to.
    QPixmap* annot = fg_scene()->newTile(tileX, tileY, 1); //Has to be "new"; the tile may have been "set" before.
    annot->fill(QColor(0x00, 0x00, 0x00, 0x00));
    int sq = tileSize/2;

    QPainter gcOver(annot);
    QFont unHinted("Arial", 6);
    unHinted.setStyleStrategy(QFont::NoAntialias);
    gcOver.setFont(unHinted);
    gcOver.setPen(QColor(0xFF, 0x00, 0x00));

    //(UpLeft, UpRight, LowLeft, LowRight)
    std::array<QPoint, 4> tileOffsets { {{0,0}, {1,0}, {0,1}, {1,1}} };
    for (int i=0; i<4; i++) {
      Graphics::TileEntry tileEntry = (*tileQuad)[i];
      if (tileEntry.m_hFlip || tileEntry.m_vFlip) {
        QString letter = "B";
        if (tileEntry.m_hFlip && !tileEntry.m_vFlip) {
          letter = "H";
        } else if (tileEntry.m_vFlip && !tileEntry.m_hFlip) {
          letter = "V";
        }
        QRectF r(tileOffsets.at(i).x()*sq, tileOffsets.at(i).y()*sq, sq, sq);
        gcOver.drawText(r, Qt::AlignCenter, letter);
      }
    }
    gcOver.end();
}


void MapView::renderTile(Overworld::Map16Tile* tileQuad, int tileX, int tileY, uint32_t bgColor)
{
    //Nothing to draw?
    if (!tileQuad) {
        fg_scene()->clearTile(tileX, tileY, 0);
        return;
    }

    //Draw each quadrant
    int tileSize = fg_scene()->getTileSize();
    std::array<QPoint, 4> tileOffsets { {{0,0}, {1,0}, {0,1}, {1,1}} };
    int sq = tileSize/2;

    QPixmap* tile = fg_scene()->getTile(tileX, tileY, 0, true);
    tile->fill(QColor(0x00, 0x00, 0x00));  //Shouldn't matter, but let's be thorough
    QPainter gc(tile);

    for (int i=0; i<4; i++) {
        //Draw the tile
        QPoint pt = tileOffsets.at(i);
        Graphics::TileEntry tileEntry = (*tileQuad)[i];
        auto quad = GetMap8Tile(game, &tileEntry, bgColor);
        QRectF r (pt.x()*sq, pt.y()*sq, sq, sq);
        gc.drawImage(pt.x()*sq, pt.y()*sq, *quad);
    }
    gc.end();
}

void MapView::rebuildScene(int mapID)
{
    //Failsafe
    if (!game || !CurrArea) {  //TODO: CurrMap needs to be maintained statically (for now)
        return;
    }

    //This is the point where I will start abusing "auto" shamelessly.
    auto currMap = *CurrArea;

    //The QGraphics scene makes it easy to handle lots of objects at different locations
    uint32_t bgColor = 0xFF0000; //TODO: game->getAreaBgColor(mapID);
    for (int y=0; y<currMap.areaMap.Height(); y++) {
        for (int x=0; x<currMap.areaMap.Width(); x++) {
            //Translate
            auto pos = getTranslated(std::make_pair(x, y));

            //Render it
            auto tileSlot = currMap.areaMap(x,y);
            auto* tileQuad = tileSlot.IsEmpty() ? nullptr : &game->map16To8.at(tileSlot.GetRef()); //NOTE: The "&" is probably dangerous.
            renderTile(tileQuad, pos.first, pos.second, bgColor);
            renderAnnotation(tileQuad, pos.first, pos.second);
        }
    }

    //Establish some ordering
    fg_scene()->setLayerVisibility(0, true);
    fg_scene()->setLayerVisibility(1, false);

    //Now hook it up to a QGraphicsView
    fg_scene()->setVisibleTiles(offset.first, offset.second, currMap.areaMap.Width(), currMap.areaMap.Height());
    resetTransform();
    viewport()->update();
    setFocus();
}


void MapView::clearSurrounding(int mapWidth, int mapHeight, bool ClearLeft, bool ClearRight, bool ClearTop, bool ClearBottom)
{
    //Clear top and bottom
    if (ClearTop || ClearBottom) {
        for (int x=0; x<mapWidth; x++) {
            for (int row=0; row<offset.second; row++) {
                if (ClearTop) {
                    auto pos = getTranslated(std::make_pair(x, -1-row));
                    fg_scene()->clearTile(pos.first, pos.second, 0);
                }
                if (ClearBottom) {
                    auto pos = getTranslated(std::make_pair(x, mapHeight+row));
                    fg_scene()->clearTile(pos.first, pos.second, 0);
                }
            }
        }
    }

    //Clear left and right
    if (ClearLeft || ClearRight) {
        for (int y=0; y<mapHeight; y++) {
            for (int col=0; col<offset.first; col++) {
                if (ClearLeft) {
                    auto pos = getTranslated(std::make_pair(-1-col, y));
                    fg_scene()->clearTile(pos.first, pos.second, 0);
                }
                if (ClearRight) {
                    auto pos = getTranslated(std::make_pair(mapWidth+col, y));
                    fg_scene()->clearTile(pos.first, pos.second, 0);
                }
            }
        }
    }
}



std::pair<int, int> MapView::moveToEnd(const std::vector< std::vector<uint32_t> >& lookup, int startX, int startY, int xPlus, int yPlus)
{
    int x = startX;
    int y = startY;
    for (;;) {
        if (x+xPlus<0 || x+xPlus>=lookup[y].size()) { break; }
        if (y+yPlus<0 || y+yPlus>=lookup.size()) { break; }
        if (lookup[y+yPlus][x+xPlus] != lookup[y][x]) { break; }
        x += xPlus;
        y += yPlus;
        return std::pair<int, int>(x,y);
    }
}



std::pair<std::vector<uint32_t>, int> MapView::getMapList(const std::vector< std::vector<uint32_t> >& lookup, int x, int y, int numTiles, bool horiz)
{
    //Step 1: If this map extends through several grids, set the offset
    //TODO: It's not that hard to generalize this, but it makes things complicated
    int offset = 0;
    auto orig = lookup[y][x];
    if (horiz && x>0 && lookup[y][x-1]==orig) {
        offset = 16;
    }
    if (!horiz && y>0 && lookup[y-1][x]==orig) {
        offset = 14;
    }

    //Step 2: Prepare the mapList
    std::vector<uint32_t> mapList;
    mapList.push_back(orig);
    numTiles -= (horiz ? 16 : 14);
    if (numTiles>0) {
        if (horiz && x+1<lookup[y].size()) {
            if (lookup[y][x+1] != lookup[y][x]) {
                mapList.push_back(lookup[y][x+1]);
            }
        }
        if (!horiz && y+1<lookup.size()) {
            if (lookup[y+1][x] != lookup[y][x]) {
                mapList.push_back(lookup[y+1][x]);
            }
        }
    }

    //Done
    return std::make_pair(mapList, offset);
}


void MapView::fillVertTiles(int srcY, const std::vector<uint32_t>& mapList, bool onRight)
{
    //
    auto currMap = *CurrArea;
    auto otherMap = (mapList[0]<0 || mapList[0]>=game->areas.size()) ? nullptr : &game->areas[mapList[0]];
    int currMapID = 1;
    for (int destY=0; destY<currMap.areaMap.Height(); destY++) {
        for (int col=0; col<offset.first; col++) {
            //Convert to destX, translate
            int destX = onRight ? currMap.areaMap.Width()+col : -1-col;
            auto pos = getTranslated(std::make_pair(destX, destY));

            //Out of maps?
            if (!otherMap) {
                fg_scene()->clearTile(pos.first, pos.second, 0);
                continue;
            }

            //Determine srcX
            int srcX = onRight ? col : otherMap->areaMap.Width()-1-col;

            //Copy
            uint32_t bgColor = 0xFF0000; //TODO: game->getAreaBgColor(mapList[currMapID-1]);  //TEMP
            auto tileSlot = otherMap->areaMap(srcX,srcY);
            auto* tileQuad = tileSlot.IsEmpty() ? nullptr : &game->map16To8.at(tileSlot.GetRef()); //NOTE: The "&" is probably dangerous.
            renderTile(tileQuad, pos.first, pos.second, bgColor);
        }
        //Increment, possibly change maps or null out
        srcY += 1;
        if (otherMap && (srcY >= otherMap->areaMap.Height())) {
            otherMap = currMapID<mapList.size() ? &game->areas[currMapID] : nullptr;
            currMapID += 1;
            srcY = 0;
        }
    }
}


void MapView::fillHorizTiles(int srcX, const std::vector<uint32_t>& mapList, bool onBottom)
{
    //
    auto currMap = *CurrArea;
    auto otherMap = (mapList[0]<0 || mapList[0]>=game->areas.size()) ? nullptr : &game->areas[mapList[0]];
    int currMapID = 1;
    for (int destX=0; destX<currMap.areaMap.Width(); destX++) {
        for (int row=0; row<offset.second; row++) {
            //Convert to destY, translate
            int destY = onBottom ? currMap.areaMap.Height()+row : -1-row;
            auto pos = getTranslated(std::make_pair(destX, destY));

            //Out of maps?
            if (!otherMap) {
                fg_scene()->clearTile(pos.first, pos.second, 0);
                continue;
            }

            //Determine srcY
            int srcY = onBottom ? row : otherMap->areaMap.Height()-1-row;

            //Copy
            uint32_t bgColor = 0xFF0000; // game.getAreaBgColor(mapList[currMapID-1]);  //TEMP
            auto tileSlot = otherMap->areaMap(srcX,srcY);
            auto tileQuad = tileSlot.IsEmpty() ? nullptr : &game->map16To8.at(tileSlot.GetRef()); //NOTE: May be dangerous
            renderTile(tileQuad, pos.first, pos.second, bgColor);
        }

        //Increment, possibly change maps or null out
        srcX += 1;
        if (otherMap && (srcX >= otherMap->areaMap.Width())) {
            otherMap = currMapID<mapList.size() ? &game->areas[mapList[currMapID]] : nullptr;
            currMapID += 1;
            srcX = 0;
        }
    }
}



void MapView::loadSideMaps()
{
    hasLoadedSideMaps = true;

    //Failsafe (put before "hasLoaded"?)
    if (!game || !CurrArea) { return; }
    auto currMap = CurrArea;

    //Step 1: Zero out all tiles in the "diagonal" buffer area
    for (int i=0; i<4; i++) {  //(UL, UR, LL, LR)
      for (int row=0; row<offset.second; row++) {
        for (int col=0; col<offset.first; col++) {
          int x = -1-col;
          int y = -1-row;
          if (i==1 || i==3) {
            x = currMap->areaMap.Width()+col;
          }
          if (i==2 || i==3) {
            y = currMap->areaMap.Height()+row;
          }
          auto pos = getTranslated(std::make_pair(x, y));
          fg_scene()->clearTile(pos.first, pos.second, 0);
        }
      }
    }

    //Step 2: For each remaining quadrant, load the adjacent map and
    //        fill in its tiles. This is complicated by a few points:
    // (1) there might not be an adjacent map
    // (2) there might be multiple adjacent maps
    // (3) there may be various alignment issues with adjacent maps
    // (4) we may not have any connectivity information at all
    auto& lookup = AreaLookup;
    auto& connectivity = AreaConnections;
    auto mapSize = std::pair<int, int>(currMap->areaMap.Width(), currMap->areaMap.Height());
    if (lookup.count(CurrAreaID)==0) {
        clearSurrounding(mapSize.first, mapSize.second, true, true, true, true);
    } else {
        auto areaPos = lookup[CurrAreaID];

        //Left
        auto topLeft = moveToEnd(connectivity, areaPos.x(), areaPos.y(), 0, -1);
        topLeft = moveToEnd(connectivity, topLeft.first, topLeft.second, -1, 0);
        if ((topLeft.first==0)) {
            clearSurrounding(mapSize.first, mapSize.second, true, false, false, false);
        } else {
            auto res = getMapList(connectivity, topLeft.first-1, topLeft.second, currMap->areaMap.Height(), false);
            fillVertTiles(res.second, res.first, false);
        }

        //Right
        auto topRight = moveToEnd(connectivity, areaPos.x(), areaPos.y(), 0, -1);
        topRight = moveToEnd(connectivity, topRight.first, topRight.second, 1, 0);
        if (topRight.first==connectivity[topRight.second].size()-1) {
            clearSurrounding(mapSize.first, mapSize.second, false, true, false, false);
        } else {
            auto res = getMapList(connectivity, topRight.first+1, topRight.second, currMap->areaMap.Height(), false);
            fillVertTiles(res.second, res.first, true);
        }

        //Top
        if (topLeft.second==0) {
            clearSurrounding(mapSize.first, mapSize.second, false, false, true, false);
        } else {
            auto res = getMapList(connectivity, topLeft.first, topLeft.second-1, currMap->areaMap.Width(), true);
            fillHorizTiles(res.second, res.first, false);
        }

        //Bottom
        auto lowerLeft = moveToEnd(connectivity, areaPos.x(), areaPos.y(), 0, 1);
        lowerLeft = moveToEnd(connectivity, lowerLeft.first, lowerLeft.second, -1, 0);
        if (lowerLeft.second==connectivity.size()-1) {
            clearSurrounding(mapSize.first, mapSize.second, false, false, false, true);
        } else {
            auto res = getMapList(connectivity, lowerLeft.first, lowerLeft.second+1, currMap->areaMap.Width(), true);
            fillHorizTiles(res.second, res.first, true);
        }
    }
}


namespace {
void eraseTriangles(QPainter& painter, QPointF center, float size, int count) {
  //Big erase
  painter.drawPolygon(QPolygonF({  QPointF(center.x()-size/4.0, center.y()),
                                   QPointF(center.x()+size/4.0, center.y()),
                                   QPointF(center.x(),          center.y()+size/2.0)}));

  //Continue? (NOTE: This is depth-first. We could make it breadth-first... but does it matter?)
  if (count>1) {
    count -= 1;
    size /= 2;
    if (size <= 1.0) { //Avoid looping forever
      return;
    }
    eraseTriangles(painter, QPointF(center.x()-size/2.0, center.y()+size/2.0), size, count);
    eraseTriangles(painter, QPointF(center.x(),          center.y()-size/2.0), size, count);
    eraseTriangles(painter, QPointF(center.x()+size/2.0, center.y()+size/2.0), size, count);
  }
}
}//End anon namespace
void MapView::paintEvent (QPaintEvent* event)
{
    //Common case
    if (validMap) {
        FixedGridGraphicsView::paintEvent(event);
        return;
    }

    //Fill
    QPainter painter(viewport());
    QColor bgColor(0x22, 0x22, 0x22);
    painter.fillRect(QRect(0, 0, viewport()->width(), viewport()->height()), bgColor);
    painter.setRenderHint(QPainter::TextAntialiasing, true);

    //Math!
    float areaSize = std::min(viewport()->width()/4.0, viewport()->height()/2.0);
    auto viewCtr = QPointF(viewport()->width()/2.0, viewport()->height()/2.0);

    //Draw the first (and only) foreground triangle
    painter.setBrush(QBrush(QColor(0x59, 0x59, 0x1a)));
    painter.drawPolygon(QPolygonF({  QPointF(viewCtr.x()-areaSize/2.0, viewCtr.y()+areaSize/2.0),
                                 QPointF(viewCtr.x(),              viewCtr.y()-areaSize/2.0),
                                 QPointF(viewCtr.x()+areaSize/2.0, viewCtr.y()+areaSize/2.0)}));

    //Recursively eliminate it to "count"
    painter.setBrush(QBrush(bgColor));
    eraseTriangles(painter, viewCtr, areaSize, triCount);

    //Signal painter is done.
    painter.end();
}


void MapView::showHideMapLayer(bool show)
{
    mapLayerVis = show;
    fg_scene()->setLayerVisibility(0, mapLayerVis);
    viewport()->update();
}


void MapView::showHideAnnotLayer(bool show)
{
    annotLayerVis = show;
    fg_scene()->setLayerVisibility(1, annotLayerVis);
    viewport()->update();
}


void MapView::showBorderAreas(bool show)
{
    if (!game || !CurrArea) { return; }
    auto currMap = *CurrArea;
    auto currSize = std::pair<int, int>(currMap.areaMap.Width(), currMap.areaMap.Height());

    if (show) {
        if (!hasLoadedSideMaps) {
            loadSideMaps();
        }
        fg_scene()->setVisibleTiles(0, 0, currSize.first+BorderSize.first*2, currSize.second+BorderSize.second*2);
    } else {
        fg_scene()->setVisibleTiles(offset.first, offset.second, currSize.first, currSize.second);
    }
    viewport()->update();
}


void MapView::keyPressEvent (QKeyEvent* event)
{//Handle some common key presses
    //Zoom: Ctrl+, Ctrl-, Ctrl0 (reset)
    if (event->modifiers()==Qt::ControlModifier) {
        if (event->key()==Qt::Key_Equal) {  //Because Key_Plus means shift
            scale(1.20, 1.20);
        } else if (event->key()==Qt::Key_Minus) {
            scale(0.80, 0.80);
        } else if (event->key()==Qt::Key_0) {
            resetTransform();
            //self.currRotation = 0;
        }
    }

    //Test
    if (event->key()==Qt::Key_T) {
        fg_scene()->setLayerVisibility(0, (mapLayerVis ? false : true));
    }
    if (event->key()==Qt::Key_Y) {
        fg_scene()->setLayerVisibility(1, (annotLayerVis ? false : true));
    }

    //Completely unnecessary rotations! (Ctrl,< or >)
    if (event->key()==44 and event->modifiers()==Qt::ControlModifier) {
        rotate(2.0);
    } else if (event->key()==46 and event->modifiers()==Qt::ControlModifier) {
        rotate(-2.0);
    }
}


